@@ -17,7 +17,7 @@ main = do
     prefix <- System.Directory.makeAbsolute pwd
 
     Test.Mockery.Directory.inTempDirectory $ do
-        writeFile "makeBools" "λ(n : Bool) → [ n && True, n && False, n || True, n || False ]"
+        writeFile "makeBools.dhall" "λ(n : Bool) → [ n && True, n && False, n || True, n || False ]"
         writeFile "bool1" "True"
         writeFile "bool2" "False"
         writeFile "both" "./bool1 && ./bool2"
@@ -27,9 +27,17 @@ main = do
         Test.DocTest.doctest
             [ "-DWITH_HTTP"
             , "--fast"
+            , prefix </> "ghc-src"
+
+            -- Unfortunately we cannot target the entire @src@ directory.
+            -- The reason is that src/Dhall/Version.hs depends on
+            -- the generated Paths_dhall module which is "out-of-scope"
+            -- when running the testsuite with cabal v1-test.
+            -- Instead, we target a selection of modules whose combined module
+            -- dependency tree covers all modules that contain doctests.
+
+            -- , prefix </> "src"
             , "-i" <> (prefix </> "src")
-            , "-i" <> (prefix </> "ghc-src")
-            , prefix </> "src/Dhall.hs"
-            , prefix </> "src/Dhall/Import.hs"
+            , prefix </> "src/Dhall/Tags.hs"
             , prefix </> "src/Dhall/Tutorial.hs"
             ]
