@@ -25,8 +25,7 @@ import Data.String (IsString(..))
 import Data.Text (Text)
 import Data.Text.Prettyprint.Doc (Doc, Pretty)
 import Data.Void (Void)
-import Dhall.Core (Binding(..), Chunks (..), Const(..), Expr(..), Var(..))
-import Dhall.Binary (ToTerm)
+import Dhall.Syntax (Binding(..), Chunks (..), Const(..), DhallDouble(..), Expr(..), Var(..))
 import Dhall.Map (Map)
 import Dhall.Set (Set)
 import Dhall.Pretty.Internal (Ann)
@@ -37,11 +36,14 @@ import qualified Data.List.NonEmpty
 import qualified Data.Set
 import qualified Data.Text
 import qualified Data.Text.Prettyprint.Doc  as Pretty
-import qualified Dhall.Core
+import qualified Dhall.Normalize
 import qualified Dhall.Map
 import qualified Dhall.Set
 import qualified Dhall.Pretty.Internal      as Internal
 
+{-| This type is a `Doc` enriched with a `same` flag to efficiently track if
+    any difference was detected
+-}
 data Diff =
     Diff
         { same :: Bool
@@ -153,11 +155,11 @@ rparen :: Diff
 rparen = token Internal.rparen
 
 -- | Render the difference between the normal form of two expressions
-diffNormalized :: (Eq a, Pretty a, ToTerm a) => Expr s a -> Expr s a -> Diff
+diffNormalized :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
 diffNormalized l0 r0 = Dhall.Diff.diff l1 r1
   where
-    l1 = Dhall.Core.alphaNormalize (Dhall.Core.normalize l0)
-    r1 = Dhall.Core.alphaNormalize (Dhall.Core.normalize r0)
+    l1 = Dhall.Normalize.alphaNormalize (Dhall.Normalize.normalize l0)
+    r1 = Dhall.Normalize.alphaNormalize (Dhall.Normalize.normalize r0)
 
 diffPrimitive :: Eq a => (a -> Diff) -> a -> a -> Diff
 diffPrimitive f l r
@@ -185,8 +187,8 @@ diffLabels ksL ksR =
 diffNatural :: Natural -> Natural -> Diff
 diffNatural = diffPrimitive (token . Internal.prettyNatural)
 
-diffDouble :: Double -> Double -> Diff
-diffDouble = diffPrimitive (token . Internal.prettyDouble)
+diffDouble :: DhallDouble -> DhallDouble -> Diff
+diffDouble = diffPrimitive (token . Internal.prettyDouble . getDhallDouble)
 
 diffConst :: Const -> Const -> Diff
 diffConst = diffPrimitive (token . Internal.prettyConst)
@@ -359,15 +361,23 @@ diffChunks cL cR
 diffList
     :: (Eq a, Pretty a)
     => Seq (Expr Void a) -> Seq (Expr Void a) -> Diff
-diffList l r = bracketed (foldMap diffPart parts)
+diffList l r = bracketed (loop parts₀)
   where
     -- Sections of the list that are only in left, only in right, or in both
-    parts =
-        Algo.Diff.getGroupedDiffBy ((same .) . diff) (toList l) (toList r)
+    parts₀ = Algo.Diff.getGroupedDiffBy equal (toList l) (toList r)
+
+    equal a b = same (diff a b)
 
     -- Render each element of a list using an extra rendering function f
     prettyElems f = map (f . token . Internal.prettyExpr)
 
+    loop [] =
+        mempty
+    loop (Algo.Diff.First as : Algo.Diff.Second bs : parts)
+        | length as == length bs = zipWith diff as bs <> loop parts
+    loop (part : parts) =
+        diffPart part <> loop parts
+
     diffPart part =
       case part of
         -- Only present in left
@@ -583,6 +593,10 @@ skeleton (Prefer {}) =
     <>  operator "⫽"
     <>  " "
     <>  ignore
+skeleton (RecordCompletion {}) =
+        ignore
+    <>  operator "::"
+    <>  ignore
 skeleton (Merge {}) =
         keyword "merge"
     <>  " "
@@ -868,7 +882,7 @@ diffCombineExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffCombineExpression l@(Combine {}) r@(Combine {}) =
     enclosed' "  " (operator "∧" <> " ") (docs l r)
   where
-    docs (Combine aL bL) (Combine aR bR) =
+    docs (Combine _ aL bL) (Combine _ aR bR) =
         Data.List.NonEmpty.cons (diffPreferExpression aL aR) (docs bL bR)
     docs aL aR =
         pure (diffPreferExpression aL aR)
@@ -1005,6 +1019,16 @@ diffImportExpression l@(Embed {}) r =
 diffImportExpression l r@(Embed {}) =
     mismatch l r
 diffImportExpression l r =
+    diffRecordCompletionExpression l r
+
+diffRecordCompletionExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
+diffRecordCompletionExpression (RecordCompletion aL bL) (RecordCompletion aR bR) =
+       diffSelectorExpression aL aR <> "::" <> diffSelectorExpression bL bR
+diffRecordCompletionExpression l@(RecordCompletion {}) r =
+    mismatch l r
+diffRecordCompletionExpression l r@(RecordCompletion {}) =
+    mismatch l r
+diffRecordCompletionExpression l r =
     diffSelectorExpression l r
 
 diffSelectorExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
@@ -1120,6 +1144,18 @@ diffPrimitiveExpression l@Integer r =
     mismatch l r
 diffPrimitiveExpression l r@Integer =
     mismatch l r
+diffPrimitiveExpression IntegerClamp IntegerClamp =
+    "…"
+diffPrimitiveExpression l@IntegerClamp r =
+    mismatch l r
+diffPrimitiveExpression l r@IntegerClamp =
+    mismatch l r
+diffPrimitiveExpression IntegerNegate IntegerNegate =
+    "…"
+diffPrimitiveExpression l@IntegerNegate r =
+    mismatch l r
+diffPrimitiveExpression l r@IntegerNegate =
+    mismatch l r
 diffPrimitiveExpression IntegerShow IntegerShow =
     "…"
 diffPrimitiveExpression l@IntegerShow r =
