@@ -1,12 +1,16 @@
+{-# LANGUAGE AllowAmbiguousTypes #-}
 {-# LANGUAGE BangPatterns        #-}
 {-# LANGUAGE CPP                 #-}
+{-# LANGUAGE FlexibleContexts    #-}
 {-# LANGUAGE LambdaCase          #-}
 {-# LANGUAGE OverloadedStrings   #-}
 {-# LANGUAGE PatternSynonyms     #-}
 {-# LANGUAGE RankNTypes          #-}
 {-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE StandaloneDeriving  #-}
 {-# LANGUAGE TupleSections       #-}
 {-# LANGUAGE ViewPatterns        #-}
+{-# LANGUAGE UndecidableInstances #-}
 
 {-# OPTIONS_GHC -O #-}
 
@@ -45,6 +49,7 @@ module Dhall.Eval (
   , Environment(..)
   , Val(..)
   , (~>)
+  , textShow
   ) where
 
 import Data.Foldable (foldr', toList)
@@ -53,12 +58,12 @@ import Data.Sequence (Seq, ViewL(..), ViewR(..))
 import Data.Text (Text)
 import Data.Void (Void)
 
-import Dhall.Core
+import Dhall.Syntax
   ( Binding(..)
   , Expr(..)
   , Chunks(..)
   , Const(..)
-  , Import
+  , DhallDouble(..)
   , Var(..)
   )
 
@@ -67,20 +72,22 @@ import Dhall.Set (Set)
 import GHC.Natural (Natural)
 import Prelude hiding (succ)
 
-import qualified Codec.Serialise as Serialise
+import qualified Data.Char
 import qualified Data.Sequence   as Sequence
 import qualified Data.Set
 import qualified Data.Text       as Text
-import qualified Dhall.Binary    as Binary
-import qualified Dhall.Core      as Core
+import qualified Dhall.Syntax    as Syntax
 import qualified Dhall.Map       as Map
 import qualified Dhall.Set
+import qualified Text.Printf
 
 data Environment a
     = Empty
     | Skip   !(Environment a) {-# UNPACK #-} !Text
     | Extend !(Environment a) {-# UNPACK #-} !Text (Val a)
 
+deriving instance (Show a, Show (Val a -> Val a)) => Show (Environment a)
+
 errorMsg :: String
 errorMsg = unlines
   [ _ERROR <> ": Compiler bug                                                        "
@@ -98,8 +105,13 @@ errorMsg = unlines
 
 
 data Closure a = Closure !Text !(Environment a) !(Expr Void a)
+
+deriving instance (Show a, Show (Val a -> Val a)) => Show (Closure a)
+
 data VChunks a = VChunks ![(Text, Val a)] !Text
 
+deriving instance (Show a, Show (Val a -> Val a)) => Show (VChunks a)
+
 instance Semigroup (VChunks a) where
   VChunks xys z <> VChunks [] z' = VChunks xys (z <> z')
   VChunks xys z <> VChunks ((x', y'):xys') z' = VChunks (xys ++ (z <> x', y'):xys') z'
@@ -121,20 +133,13 @@ data HLamInfo a
   -- ^ Don't store any information
   | Typed !Text (Val a)
   -- ^ Store the original name and type of the variable bound by the `Lam`
-  | NaturalFoldCl (Val a)
-  -- ^ The original function was a @Natural/fold@.  We need to preserve this
-  --   information in order to implement @Natural/{build,fold}@ fusion
-  | ListFoldCl (Val a)
-  -- ^ The original function was a @List/fold@.  We need to preserve this
-  --   information in order to implement @List/{build,fold}@ fusion
-  | OptionalFoldCl (Val a)
-  -- ^ The original function was an @Optional/fold@.  We need to preserve this
-  --   information in order to implement @Optional/{build,fold}@ fusion
   | NaturalSubtractZero
   -- ^ The original function was a @Natural/subtract 0@.  We need to preserve
   --   this information in case the @Natural/subtract@ ends up not being fully
   --   saturated, in which case we need to recover the unsaturated built-in
 
+deriving instance (Show a, Show (Val a -> Val a)) => Show (HLamInfo a)
+
 pattern VPrim :: (Val a -> Val a) -> Val a
 pattern VPrim f = VHLam Prim f
 
@@ -178,11 +183,13 @@ data Val a
 
     | VInteger
     | VIntegerLit !Integer
+    | VIntegerClamp !(Val a)
+    | VIntegerNegate !(Val a)
     | VIntegerShow !(Val a)
     | VIntegerToDouble !(Val a)
 
     | VDouble
-    | VDoubleLit !Double
+    | VDoubleLit !DhallDouble
     | VDoubleShow !(Val a)
 
     | VText
@@ -209,7 +216,7 @@ data Val a
     | VRecord !(Map Text (Val a))
     | VRecordLit !(Map Text (Val a))
     | VUnion !(Map Text (Maybe (Val a)))
-    | VCombine !(Val a) !(Val a)
+    | VCombine !(Maybe Text) !(Val a) !(Val a)
     | VCombineTypes !(Val a) !(Val a)
     | VPrefer !(Val a) !(Val a)
     | VMerge !(Val a) !(Val a) !(Maybe (Val a))
@@ -221,6 +228,9 @@ data Val a
     | VEquivalent !(Val a) !(Val a)
     | VEmbed a
 
+-- | For use with "Text.Show.Functions".
+deriving instance (Show a, Show (Val a -> Val a)) => Show (Val a)
+
 (~>) :: Val a -> Val a -> Val a
 (~>) a b = VHPi "_" a (\_ -> b)
 {-# INLINE (~>) #-}
@@ -278,17 +288,17 @@ vPrefer env t u =
             VPrefer t' u'
 {-# INLINE vPrefer #-}
 
-vCombine :: Val a -> Val a -> Val a
-vCombine t u =
+vCombine :: Maybe Text -> Val a -> Val a -> Val a
+vCombine mk t u =
     case (t, u) of
         (VRecordLit m, u') | null m ->
             u'
         (t', VRecordLit m) | null m ->
             t'
         (VRecordLit m, VRecordLit m') ->
-            VRecordLit (Map.unionWith vCombine m m')
+            VRecordLit (Map.unionWith (vCombine Nothing) m m')
         (t', u') ->
-            VCombine t' u'
+            VCombine mk t' u'
 
 vCombineTypes :: Val a -> Val a -> Val a
 vCombineTypes t u =
@@ -346,11 +356,11 @@ vField t0 k = go t0
         VPrefer l (VRecordLit m) -> case Map.lookup k m of
             Just v -> v
             Nothing -> go l
-        VCombine (VRecordLit m) r -> case Map.lookup k m of
-            Just v -> VField (VCombine (singletonVRecordLit v) r) k
+        VCombine mk (VRecordLit m) r -> case Map.lookup k m of
+            Just v -> VField (VCombine mk (singletonVRecordLit v) r) k
             Nothing -> go r
-        VCombine l (VRecordLit m) -> case Map.lookup k m of
-            Just v -> VField (VCombine l (singletonVRecordLit v)) k
+        VCombine mk l (VRecordLit m) -> case Map.lookup k m of
+            Just v -> VField (VCombine mk l (singletonVRecordLit v)) k
             Nothing -> go l
         t -> VField t k
 
@@ -451,14 +461,12 @@ eval !env t0 =
                         go  acc m = go (vApp succ acc) (m - 1)
                     in  go zero (fromIntegral n :: Integer)
                 n ->
-                    VHLam (NaturalFoldCl n) $ \natural ->
+                    VPrim $ \natural ->
                     VPrim $ \succ ->
                     VPrim $ \zero ->
                     VNaturalFold n natural succ zero
         NaturalBuild ->
             VPrim $ \case
-                VHLam (NaturalFoldCl x) _ ->
-                    x
                 VPrimVar ->
                     VNaturalBuild VPrimVar
                 t ->       t
@@ -509,6 +517,16 @@ eval !env t0 =
             VInteger
         IntegerLit n ->
             VIntegerLit n
+        IntegerClamp ->
+            VPrim $ \case
+                VIntegerLit n
+                    | 0 <= n    -> VNaturalLit (fromInteger n)
+                    | otherwise -> VNaturalLit 0
+                n -> VIntegerClamp n
+        IntegerNegate ->
+            VPrim $ \case
+                VIntegerLit n -> VIntegerLit (negate n)
+                n             -> VIntegerNegate n
         IntegerShow ->
             VPrim $ \case
                 VIntegerLit n
@@ -517,9 +535,10 @@ eval !env t0 =
                 n -> VIntegerShow n
         IntegerToDouble ->
             VPrim $ \case
-                VIntegerLit n -> VDoubleLit (read (show n))
+                VIntegerLit n -> VDoubleLit (DhallDouble (read (show n)))
                 -- `(read . show)` is used instead of `fromInteger`
-                -- because `read` uses the correct rounding rule
+                -- because `read` uses the correct rounding rule.
+                -- See https://gitlab.haskell.org/ghc/ghc/issues/17231.
                 n             -> VIntegerToDouble n
         Double ->
             VDouble
@@ -527,8 +546,8 @@ eval !env t0 =
             VDoubleLit n
         DoubleShow ->
             VPrim $ \case
-                VDoubleLit n -> VTextLit (VChunks [] (Text.pack (show n)))
-                n            -> VDoubleShow n
+                VDoubleLit (DhallDouble n) -> VTextLit (VChunks [] (Text.pack (show n)))
+                n                          -> VDoubleShow n
         Text ->
             VText
         TextLit cs ->
@@ -539,7 +558,7 @@ eval !env t0 =
             eval env (TextLit (Chunks [("", t), ("", u)] ""))
         TextShow ->
             VPrim $ \case
-                VTextLit (VChunks [] x) -> VTextLit (VChunks [] (Core.textShow x))
+                VTextLit (VChunks [] x) -> VTextLit (VChunks [] (textShow x))
                 t                       -> VTextShow t
         List ->
             VPrim VList
@@ -550,8 +569,6 @@ eval !env t0 =
         ListBuild ->
             VPrim $ \a ->
             VPrim $ \case
-                VHLam (ListFoldCl x) _ ->
-                    x
                 VPrimVar ->
                     VListBuild a VPrimVar
                 t ->       t
@@ -570,7 +587,7 @@ eval !env t0 =
                     VHLam (Typed "nil"  list) $ \nil ->
                     foldr' (\x b -> cons `vApp` x `vApp` b) nil as
                 as ->
-                    VHLam (ListFoldCl as) $ \t ->
+                    VPrim $ \t ->
                     VPrim $ \c ->
                     VPrim $ \n ->
                     VListFold a as t c n
@@ -602,14 +619,14 @@ eval !env t0 =
                 VListLit _ as ->
                     let a' =
                             if null as
-                            then Just (VList (VRecord (Map.fromList [("index", VNatural), ("value", a)])))
+                            then Just (VList (VRecord (Map.unorderedFromList [("index", VNatural), ("value", a)])))
                             else Nothing
 
                         as' =
                             Sequence.mapWithIndex
                                 (\i t ->
                                     VRecordLit
-                                        (Map.fromList
+                                        (Map.unorderedFromList
                                             [ ("index", VNaturalLit (fromIntegral i))
                                             , ("value", t)
                                             ]
@@ -624,7 +641,7 @@ eval !env t0 =
             VPrim $ \ ~a ->
             VPrim $ \case
                 VListLit t as | null as ->
-                    VListLit t (Sequence.reverse as)
+                    VListLit t as
                 VListLit _ as ->
                     VListLit Nothing (Sequence.reverse as)
                 t ->
@@ -649,14 +666,13 @@ eval !env t0 =
                     VHLam (Typed "none" optional) $ \_none ->
                     some `vApp` t
                 opt ->
-                    VHLam (OptionalFoldCl opt) $ \o ->
+                    VPrim $ \o ->
                     VPrim $ \s ->
                     VPrim $ \n ->
                     VOptionalFold a opt o s n
         OptionalBuild ->
             VPrim $ \ ~a ->
             VPrim $ \case
-                VHLam (OptionalFoldCl x) _ -> x
                 VPrimVar -> VOptionalBuild a VPrimVar
                 t ->       t
                     `vApp` VOptional a
@@ -668,32 +684,40 @@ eval !env t0 =
             VRecordLit (Map.sort (fmap (eval env) kts))
         Union kts ->
             VUnion (Map.sort (fmap (fmap (eval env)) kts))
-        Combine t u ->
-            vCombine (eval env t) (eval env u)
+        Combine mk t u ->
+            vCombine mk (eval env t) (eval env u)
         CombineTypes t u ->
             vCombineTypes (eval env t) (eval env u)
         Prefer t u ->
             vPrefer env (eval env t) (eval env u)
+        RecordCompletion t u ->
+            eval env (Annot (Prefer (Field t "default") u) (Field t "Type"))
         Merge x y ma ->
             case (eval env x, eval env y, fmap (eval env) ma) of
                 (VRecordLit m, VInject _ k mt, _)
                     | Just f <- Map.lookup k m -> maybe f (vApp f) mt
                     | otherwise                -> error errorMsg
+                (VRecordLit m, VSome t, _)
+                    | Just f <- Map.lookup "Some" m -> vApp f t
+                    | otherwise                     -> error errorMsg
+                (VRecordLit m, VNone _, _)
+                    | Just t <- Map.lookup "None" m -> t
+                    | otherwise                     -> error errorMsg
                 (x', y', ma') -> VMerge x' y' ma'
         ToMap x ma ->
             case (eval env x, fmap (eval env) ma) of
                 (VRecordLit m, ma'@(Just _)) | null m ->
-                    VListLit ma' (Sequence.empty)
+                    VListLit ma' Sequence.empty
                 (VRecordLit m, _) ->
                     let entry (k, v) =
                             VRecordLit
-                                (Map.fromList
+                                (Map.unorderedFromList
                                     [ ("mapKey", VTextLit $ VChunks [] k)
                                     , ("mapValue", v)
                                     ]
                                 )
 
-                        s = (Sequence.fromList . map entry . Map.toList) m
+                        s = (Sequence.fromList . map entry . Map.toAscList) m
 
                     in  VListLit Nothing s
                 (x', ma') ->
@@ -741,7 +765,7 @@ eqListBy f = go
 eqMapsBy :: Ord k => (v -> v -> Bool) -> Map k v -> Map k v -> Bool
 eqMapsBy f mL mR =
     Map.size mL == Map.size mR
-    && eqListBy eq (Map.toList mL) (Map.toList mR)
+    && eqListBy eq (Map.toAscList mL) (Map.toAscList mR)
   where
     eq (kL, vL) (kR, vR) = kL == kR && f vL vR
 {-# INLINE eqMapsBy #-}
@@ -754,6 +778,21 @@ eqMaybeBy f = go
     go _        _        = False
 {-# INLINE eqMaybeBy #-}
 
+-- | Utility that powers the @Text/show@ built-in
+textShow :: Text -> Text
+textShow text = "\"" <> Text.concatMap f text <> "\""
+  where
+    f '"'  = "\\\""
+    f '$'  = "\\u0024"
+    f '\\' = "\\\\"
+    f '\b' = "\\b"
+    f '\n' = "\\n"
+    f '\r' = "\\r"
+    f '\t' = "\\t"
+    f '\f' = "\\f"
+    f c | c <= '\x1F' = Text.pack (Text.Printf.printf "\\u%04x" (Data.Char.ord c))
+        | otherwise   = Text.singleton c
+
 conv :: forall a. Eq a => Environment a -> Val a -> Val a -> Bool
 conv !env t0 t0' =
     case (t0, t0') of
@@ -829,6 +868,10 @@ conv !env t0 t0' =
             True
         (VIntegerLit t, VIntegerLit t') ->
             t == t'
+        (VIntegerClamp t, VIntegerClamp t') ->
+            conv env t t'
+        (VIntegerNegate t, VIntegerNegate t') ->
+            conv env t t'
         (VIntegerShow t, VIntegerShow t') ->
             conv env t t'
         (VIntegerToDouble t, VIntegerToDouble t') ->
@@ -836,8 +879,7 @@ conv !env t0 t0' =
         (VDouble, VDouble) ->
             True
         (VDoubleLit n, VDoubleLit n') ->
-                Serialise.serialise (Binary.encode (DoubleLit n  :: Expr Void Import))
-            ==  Serialise.serialise (Binary.encode (DoubleLit n' :: Expr Void Import))
+            n == n'
         (VDoubleShow t, VDoubleShow t') ->
             conv env t t'
         (VText, VText) ->
@@ -882,7 +924,7 @@ conv !env t0 t0' =
             eqMapsBy (conv env) m m'
         (VUnion m, VUnion m') ->
             eqMapsBy (eqMaybeBy (conv env)) m m'
-        (VCombine t u, VCombine t' u') ->
+        (VCombine _ t u, VCombine _ t' u') ->
             conv env t t' && conv env u u'
         (VCombineTypes t u, VCombineTypes t' u') ->
             conv env t t' && conv env u u'
@@ -928,14 +970,8 @@ conv !env t0 t0' =
     convSkip x = conv (Skip env x)
     {-# INLINE convSkip #-}
 
-{-| Returns `True` if two expressions are α-equivalent and β-equivalent and
-    `False` otherwise
-
-    `judgmentallyEqual` can fail with an `error` if you compare ill-typed
-    expressions
--}
 judgmentallyEqual :: Eq a => Expr s a -> Expr t a -> Bool
-judgmentallyEqual (Core.denote -> t) (Core.denote -> u) =
+judgmentallyEqual (Syntax.denote -> t) (Syntax.denote -> u) =
     conv Empty (eval Empty t) (eval Empty u)
 
 data Names
@@ -961,7 +997,7 @@ quote !env !t0 =
         VConst k ->
             Const k
         VVar !x !i ->
-            Var (V x (fromIntegral (countNames x env - i - 1)))
+            Var (V x (countNames x env - i - 1))
         VApp t u ->
             quote env t `qApp` u
         VLam a (freshClosure -> (x, v, t)) ->
@@ -970,9 +1006,6 @@ quote !env !t0 =
             case i of
                 Typed (fresh -> (x, v)) a -> Lam x (quote env a) (quoteBind x (t v))
                 Prim                      -> quote env (t VPrimVar)
-                NaturalFoldCl{}           -> quote env (t VPrimVar)
-                ListFoldCl{}              -> quote env (t VPrimVar)
-                OptionalFoldCl{}          -> quote env (t VPrimVar)
                 NaturalSubtractZero       -> App NaturalSubtract (NaturalLit 0)
 
         VPi a (freshClosure -> (x, v, b)) ->
@@ -1021,6 +1054,10 @@ quote !env !t0 =
             Integer
         VIntegerLit n ->
             IntegerLit n
+        VIntegerClamp t ->
+            IntegerClamp `qApp` t
+        VIntegerNegate t ->
+            IntegerNegate `qApp` t
         VIntegerShow t ->
             IntegerShow `qApp` t
         VIntegerToDouble t ->
@@ -1075,8 +1112,8 @@ quote !env !t0 =
             RecordLit (fmap (quote env) m)
         VUnion m ->
             Union (fmap (fmap (quote env)) m)
-        VCombine t u ->
-            Combine (quote env t) (quote env u)
+        VCombine mk t u ->
+            Combine mk (quote env t) (quote env u)
         VCombineTypes t u ->
             CombineTypes (quote env t) (quote env u)
         VPrefer t u ->
@@ -1122,32 +1159,11 @@ quote !env !t0 =
 -- | Normalize an expression in an environment of values. Any variable pointing out of
 --   the environment is treated as opaque free variable.
 nf :: Eq a => Environment a -> Expr s a -> Expr t a
-nf !env = Core.renote . quote (envNames env) . eval env . Core.denote
+nf !env = Syntax.renote . quote (envNames env) . eval env . Syntax.denote
 
-{-| Reduce an expression to its normal form, performing beta reduction
-
-    `normalize` does not type-check the expression.  You may want to type-check
-    expressions before normalizing them since normalization can convert an
-    ill-typed expression into a well-typed expression.
-
-    `normalize` can also fail with `error` if you normalize an ill-typed
-    expression
--}
 normalize :: Eq a => Expr s a -> Expr t a
 normalize = nf Empty
 
-{-| α-normalize an expression by renaming all bound variables to @\"_\"@ and
-    using De Bruijn indices to distinguish them
-
->>> alphaNormalize (Lam "a" (Const Type) (Lam "b" (Const Type) (Lam "x" "a" (Lam "y" "b" "x"))))
-Lam "_" (Const Type) (Lam "_" (Const Type) (Lam "_" (Var (V "_" 1)) (Lam "_" (Var (V "_" 1)) (Var (V "_" 1)))))
-
-    α-normalization does not affect free variables:
-
->>> alphaNormalize "x"
-Var (V "x" 0)
-
--}
 alphaNormalize :: Expr s a -> Expr s a
 alphaNormalize = goEnv EmptyNames
   where
@@ -1218,6 +1234,10 @@ alphaNormalize = goEnv EmptyNames
                 Integer
             IntegerLit n ->
                 IntegerLit n
+            IntegerClamp ->
+                IntegerClamp
+            IntegerNegate ->
+                IntegerNegate
             IntegerShow ->
                 IntegerShow
             IntegerToDouble ->
@@ -1272,12 +1292,14 @@ alphaNormalize = goEnv EmptyNames
                 RecordLit (fmap go kts)
             Union kts ->
                 Union (fmap (fmap go) kts)
-            Combine t u ->
-                Combine (go t) (go u)
+            Combine m t u ->
+                Combine m (go t) (go u)
             CombineTypes t u ->
                 CombineTypes (go t) (go u)
             Prefer t u ->
                 Prefer (go t) (go u)
+            RecordCompletion t u ->
+                RecordCompletion (go t) (go u)
             Merge x y ma ->
                 Merge (go x) (go y) (fmap go ma)
             ToMap x ma ->
