@@ -1,5 +1,6 @@
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards   #-}
+{-# LANGUAGE ViewPatterns      #-}
 
 -- | This module contains the implementation of the @dhall freeze@ subcommand
 
@@ -14,26 +15,35 @@ module Dhall.Freeze
     , Intent(..)
     ) where
 
-import Control.Exception (SomeException)
+import Data.Bifunctor (first)
 import Data.Monoid ((<>))
 import Data.Text
-import Dhall.Core (Expr(..), Import(..), ImportHashed(..), ImportType(..))
 import Dhall.Parser (Src)
-import Dhall.Pretty (CharacterSet, annToAnsiStyle, layoutOpts, prettyCharacterSet)
-import Dhall.TypeCheck (X)
-import Dhall.Util (Censor, Input(..))
+import Dhall.Pretty (CharacterSet)
+import Dhall.Syntax (Expr(..), Import(..), ImportHashed(..), ImportType(..))
+import Dhall.Util
+    ( Censor
+    , CheckFailed(..)
+    , Header(..)
+    , Input(..)
+    , OutputMode(..)
+    )
 import System.Console.ANSI (hSupportsANSI)
 
-import qualified Control.Exception
+import qualified Control.Exception                         as Exception
 import qualified Control.Monad.Trans.State.Strict          as State
+import qualified Data.Text.IO                              as Text.IO
 import qualified Data.Text.Prettyprint.Doc                 as Pretty
 import qualified Data.Text.Prettyprint.Doc.Render.Terminal as Pretty
-import qualified Data.Text.IO
-import qualified Dhall.Core
+import qualified Data.Text.Prettyprint.Doc.Render.Text     as Pretty.Text
+import qualified Dhall.Core                                as Core
 import qualified Dhall.Import
 import qualified Dhall.Optics
+import qualified Dhall.Parser                              as Parser
+import qualified Dhall.Pretty
 import qualified Dhall.TypeCheck
-import qualified Dhall.Util
+import qualified Dhall.Util                                as Util
+import qualified System.AtomicWrite.Writer.LazyText        as AtomicWrite.LazyText
 import qualified System.FilePath
 import qualified System.IO
 
@@ -54,25 +64,16 @@ freezeImport directory import_ = do
 
     let status = Dhall.Import.emptyStatus directory
 
-    let download =
-            State.evalStateT (Dhall.Import.loadWith (Embed import_)) status
-
-    -- Try again without the semantic integrity check if decoding fails
-    let handler :: SomeException -> IO (Expr Src X)
-        handler _ = do
-            State.evalStateT (Dhall.Import.loadWith (Embed unprotectedImport)) status
-
-    expression <- Control.Exception.handle handler download
+    expression <- State.evalStateT (Dhall.Import.loadWith (Embed unprotectedImport)) status
 
     case Dhall.TypeCheck.typeOf expression of
-        Left  exception -> Control.Exception.throwIO exception
+        Left  exception -> Exception.throwIO exception
         Right _         -> return ()
 
-    let normalizedExpression =
-            Dhall.Core.alphaNormalize (Dhall.Core.normalize expression)
+    let normalizedExpression = Core.alphaNormalize (Core.normalize expression)
 
     -- make sure the frozen import is present in the semantic cache
-    Dhall.Import.writeExpressionToSemanticCache expression
+    Dhall.Import.writeExpressionToSemanticCache (Core.denote expression)
 
     let expressionHash = Dhall.Import.hashExpression normalizedExpression
 
@@ -94,23 +95,26 @@ freezeRemoteImport directory import_ = do
         _         -> return import_
 
 writeExpr :: Input -> (Text, Expr Src Import) -> CharacterSet -> IO ()
-writeExpr inplace (header, expr) characterSet = do
+writeExpr input (header, expr) characterSet = do
     let doc =  Pretty.pretty header
             <> Dhall.Pretty.prettyCharacterSet characterSet expr
+            <> "\n"
 
-    let unAnnotated = Pretty.layoutSmart layoutOpts (Pretty.unAnnotate doc)
+    let stream = Dhall.Pretty.layout doc
 
-    case inplace of
-        File f ->
-            System.IO.withFile f System.IO.WriteMode (\handle -> do
-                Pretty.renderIO handle unAnnotated
-                Data.Text.IO.hPutStrLn handle "" )
+    let unAnnotated = Pretty.unAnnotateS stream
+
+    case input of
+        InputFile file ->
+            AtomicWrite.LazyText.atomicWriteFile
+                file
+                (Pretty.Text.renderLazy unAnnotated)
 
         StandardInput -> do
             supportsANSI <- System.Console.ANSI.hSupportsANSI System.IO.stdout
             if supportsANSI
                then
-                 Pretty.renderIO System.IO.stdout (annToAnsiStyle <$> Pretty.layoutSmart layoutOpts doc)
+                 Pretty.renderIO System.IO.stdout (Dhall.Pretty.annToAnsiStyle <$> stream)
                else
                  Pretty.renderIO System.IO.stdout unAnnotated
 
@@ -134,18 +138,17 @@ data Intent
 
 -- | Implementation of the @dhall freeze@ subcommand
 freeze
-    :: Input
+    :: OutputMode
+    -> Input
     -> Scope
     -> Intent
     -> CharacterSet
     -> Censor
     -> IO ()
-freeze inplace scope intent characterSet censor = do
-    let directory = case inplace of
-            StandardInput -> "."
-            File file     -> System.FilePath.takeDirectory file
-
-    (header, parsedExpression) <- Dhall.Util.getExpressionAndHeader censor inplace
+freeze outputMode input scope intent characterSet censor = do
+    let directory = case input of
+            StandardInput  -> "."
+            InputFile file -> System.FilePath.takeDirectory file
 
     let freezeScope =
             case scope of
@@ -156,10 +159,10 @@ freeze inplace scope intent characterSet censor = do
 
     let cache
             (ImportAlt
-                (Embed
+                (Core.shallowDenote -> Embed
                     (Import { importHashed = ImportHashed { hash = Just _expectedHash } })
                 )
-                import_@(ImportAlt
+                import_@(Core.shallowDenote -> ImportAlt
                     (Embed
                         (Import { importHashed = ImportHashed { hash = Just _actualHash } })
                     )
@@ -194,10 +197,46 @@ freeze inplace scope intent characterSet censor = do
                     traverse freezeFunction expression
                 Cache  ->
                     Dhall.Optics.transformMOf
-                        Dhall.Core.subExpressions
+                        Core.subExpressions
                         cache
-                        (Dhall.Core.denote expression)
+                        expression
 
-    frozenExpression <- rewrite parsedExpression
+    case outputMode of
+        Write -> do
+            (Header header, parsedExpression) <- do
+                Util.getExpressionAndHeader censor input
 
-    writeExpr inplace (header, frozenExpression) characterSet
+            frozenExpression <- rewrite parsedExpression
+
+            writeExpr input (header, frozenExpression) characterSet
+
+        Check -> do
+            originalText <- case input of
+                InputFile file -> Text.IO.readFile file
+                StandardInput  -> Text.IO.getContents
+
+            let name = case input of
+                    InputFile file -> file
+                    StandardInput  -> "(stdin)"
+
+            (Header header, parsedExpression) <- do
+                Core.throws (first Parser.censor (Parser.exprAndHeaderFromText name originalText))
+
+            frozenExpression <- rewrite parsedExpression
+
+            let doc =  Pretty.pretty header
+                    <> Dhall.Pretty.prettyCharacterSet characterSet frozenExpression
+                    <> "\n"
+
+            let stream = Dhall.Pretty.layout doc
+
+            let modifiedText = Pretty.Text.renderStrict stream
+
+            if originalText == modifiedText
+                then return ()
+                else do
+                    let command = "freeze"
+
+                    let modified = "frozen"
+
+                    Exception.throwIO CheckFailed{..}
