@@ -122,6 +122,7 @@ module Dhall.Import (
     , graph
     , remote
     , toHeaders
+    , substitutions
     , normalizer
     , startingContext
     , chainImport
@@ -138,7 +139,6 @@ module Dhall.Import (
     ) where
 
 import Control.Applicative (Alternative(..))
-import Codec.CBOR.Term (Term(..))
 import Control.Exception (Exception, SomeException, IOException, toException)
 import Control.Monad (when)
 import Control.Monad.Catch (throwM, MonadCatch(catch), handle)
@@ -150,15 +150,11 @@ import Data.CaseInsensitive (CI)
 import Data.List.NonEmpty (NonEmpty(..))
 import Data.Semigroup (Semigroup(..))
 import Data.Text (Text)
-import Data.Void (absurd)
-#if MIN_VERSION_base(4,8,0)
-#else
-import Data.Traversable (traverse)
-#endif
+import Data.Void (Void, absurd)
 import Data.Typeable (Typeable)
 import System.FilePath ((</>))
 import Dhall.Binary (StandardVersion(..))
-import Dhall.Core
+import Dhall.Syntax
     ( Expr(..)
     , Chunks(..)
     , Directory(..)
@@ -173,25 +169,25 @@ import Dhall.Core
     , chunkExprs
     )
 #ifdef WITH_HTTP
-import Dhall.Import.HTTP hiding (HTTPHeader)
+import Dhall.Import.HTTP
 #endif
 import Dhall.Import.Types
 
 import Dhall.Parser (Parser(..), ParseError(..), Src(..), SourcedException(..))
-import Dhall.TypeCheck (X)
 import Lens.Family.State.Strict (zoom)
 
+import qualified Codec.CBOR.Encoding                         as Encoding
+import qualified Codec.CBOR.Write                            as Write
 import qualified Codec.Serialise
-import qualified Control.Monad.Trans.Maybe        as Maybe
-import qualified Control.Monad.Trans.State.Strict as State
+import qualified Control.Monad.Trans.Maybe                   as Maybe
+import qualified Control.Monad.Trans.State.Strict            as State
 import qualified Data.ByteString
 import qualified Data.ByteString.Lazy
 import qualified Data.CaseInsensitive
 import qualified Data.Foldable
-import qualified Data.List.NonEmpty               as NonEmpty
-import qualified Data.Map.Strict                  as Map
+import qualified Data.List.NonEmpty                          as NonEmpty
 import qualified Data.Text.Encoding
-import qualified Data.Text                        as Text
+import qualified Data.Text                                   as Text
 import qualified Data.Text.IO
 import qualified Dhall.Binary
 import qualified Dhall.Core
@@ -199,12 +195,14 @@ import qualified Dhall.Crypto
 import qualified Dhall.Map
 import qualified Dhall.Parser
 import qualified Dhall.Pretty.Internal
+import qualified Dhall.Substitution
 import qualified Dhall.TypeCheck
+import qualified System.AtomicWrite.Writer.ByteString.Binary as AtomicWrite.Binary
 import qualified System.Environment
 import qualified System.Info
 import qualified System.IO
-import qualified System.Directory                 as Directory
-import qualified System.FilePath                  as FilePath
+import qualified System.Directory                            as Directory
+import qualified System.FilePath                             as FilePath
 import qualified Text.Megaparsec
 import qualified Text.Parser.Combinators
 import qualified Text.Parser.Token
@@ -462,7 +460,7 @@ chainedChangeMode :: ImportMode -> Chained -> Chained
 chainedChangeMode mode (Chained (Import importHashed _)) =
     Chained (Import importHashed mode)
 
--- Chain imports, also typecheck and normalize headers if applicable.
+-- | Chain imports, also typecheck and normalize headers if applicable.
 chainImport :: Chained -> Import -> StateT Status IO Chained
 chainImport (Chained parent) child@(Import importHashed@(ImportHashed _ (Remote url)) _) = do
     url' <- normalizeHeaders url
@@ -479,11 +477,12 @@ chainImport (Chained parent) child =
 loadImport :: Chained -> StateT Status IO ImportSemantics
 loadImport import_ = do
     Status {..} <- State.get
-    case Map.lookup import_ _cache of
+
+    case Dhall.Map.lookup import_ _cache of
         Just importSemantics -> return importSemantics
         Nothing -> do
             importSemantics <- loadImportWithSemanticCache import_
-            zoom cache (State.modify (Map.insert import_ importSemantics))
+            zoom cache (State.modify (Dhall.Map.insert import_ importSemantics))
             return importSemantics
 
 -- | Load an import from the 'semantic cache'. Defers to
@@ -505,6 +504,7 @@ loadImportWithSemanticCache
     case mCached of
         Just bytesStrict -> do
             let actualHash = Dhall.Crypto.sha256Hash bytesStrict
+
             if semanticHash == actualHash
                 then return ()
                 else do
@@ -512,12 +512,10 @@ loadImportWithSemanticCache
                     throwMissingImport (Imported _stack (HashMismatch {expectedHash = semanticHash, ..}))
 
             let bytesLazy = Data.ByteString.Lazy.fromStrict bytesStrict
-            term <- case Codec.Serialise.deserialiseOrFail bytesLazy of
-                Left err -> throwMissingImport (Imported _stack err)
-                Right t -> return t
-            importSemantics <- case Dhall.Binary.decodeExpression term of
-                Left err -> throwMissingImport (Imported _stack err)
-                Right sem -> return sem
+
+            importSemantics <- case Dhall.Binary.decodeExpression bytesLazy of
+                Left  err -> throwMissingImport (Imported _stack err)
+                Right e   -> return e
 
             return (ImportSemantics {..})
 
@@ -545,7 +543,7 @@ fetchFromSemanticCache expectedHash = Maybe.runMaybeT $ do
 
 -- | Ensure that the given expression is present in the semantic cache. The
 --   given expression should be alpha-beta-normal.
-writeExpressionToSemanticCache :: Expr Src X -> IO ()
+writeExpressionToSemanticCache :: Expr Void Void -> IO ()
 writeExpressionToSemanticCache expression = writeToSemanticCache hash bytes
   where
     bytes = encodeExpression NoVersion expression
@@ -555,7 +553,7 @@ writeToSemanticCache :: Dhall.Crypto.SHA256Digest -> Data.ByteString.ByteString
 writeToSemanticCache hash bytes = do
     _ <- Maybe.runMaybeT $ do
         cacheFile <- getCacheFile "dhall" hash
-        liftIO (Data.ByteString.writeFile cacheFile bytes)
+        liftIO (AtomicWrite.Binary.atomicWriteFile cacheFile bytes)
     return ()
 
 -- Check the "semi-semantic" disk cache, otherwise typecheck and normalise from
@@ -572,7 +570,7 @@ loadImportWithSemisemanticCache (Chained (Import (ImportHashed _ importType) Cod
             absolutePath <- Directory.makeAbsolute path
             return absolutePath
         Remote url -> do
-            let urlText = Dhall.Pretty.Internal.pretty (url { headers = Nothing })
+            let urlText = Dhall.Core.pretty (url { headers = Nothing })
             return (Text.unpack urlText)
         Env env -> return $ Text.unpack env
         Missing -> throwM (MissingImports [])
@@ -593,27 +591,25 @@ loadImportWithSemisemanticCache (Chained (Import (ImportHashed _ importType) Cod
     -- Check the semi-semantic cache. See
     -- https://github.com/dhall-lang/dhall-haskell/issues/1098 for the reasoning
     -- behind semi-semantic caching.
-    let semisemanticHash = computeSemisemanticHash resolvedExpr
+    let semisemanticHash = computeSemisemanticHash (Dhall.Core.denote resolvedExpr)
     mCached <- lift $ fetchFromSemisemanticCache semisemanticHash
 
     importSemantics <- case mCached of
         Just bytesStrict -> do
             let bytesLazy = Data.ByteString.Lazy.fromStrict bytesStrict
 
-            term <- case Codec.Serialise.deserialiseOrFail bytesLazy of
-                Left err -> throwMissingImport (Imported _stack err)
-                Right t -> return t
-
-            importSemantics <- case Dhall.Binary.decodeExpression term of
+            importSemantics <- case Dhall.Binary.decodeExpression bytesLazy of
                 Left err -> throwMissingImport (Imported _stack err)
                 Right sem -> return sem
 
             return importSemantics
 
         Nothing -> do
-            betaNormal <- case Dhall.TypeCheck.typeWith _startingContext resolvedExpr of
-                Left  err -> throwMissingImport (Imported _stack err)
-                Right _ -> return (Dhall.Core.normalizeWith _normalizer resolvedExpr)
+            betaNormal <- do
+                let substitutedExpr = Dhall.Substitution.substitute resolvedExpr _substitutions
+                case Dhall.TypeCheck.typeWith _startingContext substitutedExpr of
+                    Left  err -> throwMissingImport (Imported _stack err)
+                    Right _ -> return (Dhall.Core.normalizeWith _normalizer resolvedExpr)
 
             let bytes = encodeExpression NoVersion betaNormal
             lift $ writeToSemisemanticCache semisemanticHash bytes
@@ -646,13 +642,13 @@ loadImportWithSemisemanticCache (Chained (Import (ImportHashed _ importType) Loc
             Missing -> Field locationType "Missing"
             local@(Local _ _) ->
                 App (Field locationType "Local")
-                  (TextLit (Chunks [] (Dhall.Pretty.Internal.pretty local)))
+                  (TextLit (Chunks [] (Dhall.Core.pretty local)))
             remote_@(Remote _) ->
                 App (Field locationType "Remote")
-                  (TextLit (Chunks [] (Dhall.Pretty.Internal.pretty remote_)))
+                  (TextLit (Chunks [] (Dhall.Core.pretty remote_)))
             Env env ->
                 App (Field locationType "Environment")
-                  (TextLit (Chunks [] (Dhall.Pretty.Internal.pretty env)))
+                  (TextLit (Chunks [] (Dhall.Core.pretty env)))
 
     return (ImportSemantics {..})
 
@@ -660,7 +656,7 @@ loadImportWithSemisemanticCache (Chained (Import (ImportHashed _ importType) Loc
 -- AST (without normalising or type-checking it first). See
 -- https://github.com/dhall-lang/dhall-haskell/issues/1098 for further
 -- discussion.
-computeSemisemanticHash :: Expr Src X -> Dhall.Crypto.SHA256Digest
+computeSemisemanticHash :: Expr Void Void -> Dhall.Crypto.SHA256Digest
 computeSemisemanticHash resolvedExpr = hashExpression resolvedExpr
 
 -- Fetch encoded normal form from "semi-semantic cache"
@@ -674,7 +670,7 @@ writeToSemisemanticCache :: Dhall.Crypto.SHA256Digest -> Data.ByteString.ByteStr
 writeToSemisemanticCache semisemanticHash bytes = do
     _ <- Maybe.runMaybeT $ do
         cacheFile <- getCacheFile "dhall-haskell" semisemanticHash
-        liftIO (Data.ByteString.writeFile cacheFile bytes)
+        liftIO (AtomicWrite.Binary.atomicWriteFile cacheFile bytes)
     return ()
 
 -- Fetch source code directly from disk/network
@@ -752,6 +748,7 @@ getCacheFile cacheName hash = do
 
     return cacheFile
 
+-- | Warn if no cache directory is available
 warnAboutMissingCaches :: (MonadCatch m, Alternative m, MonadIO m) => m ()
 warnAboutMissingCaches = warn <|> return ()
     where warn = Data.Foldable.traverse_ (getOrCreateCacheDirectory True) ["dhall", "dhall-haskell"]
@@ -926,7 +923,7 @@ normalizeHeaders url@URL { headers = Just headersExpression } = do
     loadedExpr <- loadWith headersExpression
 
     let go key₀ key₁ = do
-            let expected :: Expr Src X
+            let expected :: Expr Src Void
                 expected =
                     App List
                         ( Record
@@ -976,7 +973,7 @@ emptyStatus = emptyStatusWith fetchRemote
     You can configure the desired behavior through the initial `Status` that you
     supply
 -}
-loadWith :: Expr Src Import -> StateT Status IO (Expr Src X)
+loadWith :: Expr Src Import -> StateT Status IO (Expr Src Void)
 loadWith expr₀ = case expr₀ of
   Embed import₀ -> do
     Status {..} <- State.get
@@ -988,7 +985,7 @@ loadWith expr₀ = case expr₀ of
     let local (Chained (Import (ImportHashed _ (Remote  {})) _)) = False
         local (Chained (Import (ImportHashed _ (Local   {})) _)) = True
         local (Chained (Import (ImportHashed _ (Env     {})) _)) = True
-        local (Chained (Import (ImportHashed _ (Missing {})) _)) = True
+        local (Chained (Import (ImportHashed _ (Missing {})) _)) = False
 
     let referentiallySane = not (local child) || local parent
 
@@ -1012,7 +1009,7 @@ loadWith expr₀ = case expr₀ of
     ImportSemantics {..} <- loadImport child
     zoom stack (State.put _stack)
 
-    return importSemantics
+    return (Dhall.Core.renote importSemantics)
 
   ImportAlt a b -> loadWith a `catch` handler₀
     where
@@ -1052,6 +1049,8 @@ loadWith expr₀ = case expr₀ of
   NaturalTimes a b     -> NaturalTimes <$> loadWith a <*> loadWith b
   Integer              -> pure Integer
   IntegerLit a         -> pure (IntegerLit a)
+  IntegerClamp         -> pure IntegerClamp
+  IntegerNegate        -> pure IntegerNegate
   IntegerShow          -> pure IntegerShow
   IntegerToDouble      -> pure IntegerToDouble
   Double               -> pure Double
@@ -1079,9 +1078,10 @@ loadWith expr₀ = case expr₀ of
   Record a             -> Record <$> mapM loadWith a
   RecordLit a          -> RecordLit <$> mapM loadWith a
   Union a              -> Union <$> mapM (mapM loadWith) a
-  Combine a b          -> Combine <$> loadWith a <*> loadWith b
+  Combine m a b        -> Combine m <$> loadWith a <*> loadWith b
   CombineTypes a b     -> CombineTypes <$> loadWith a <*> loadWith b
   Prefer a b           -> Prefer <$> loadWith a <*> loadWith b
+  RecordCompletion a b -> RecordCompletion <$> loadWith a <*> loadWith b
   Merge a b c          -> Merge <$> loadWith a <*> loadWith b <*> mapM loadWith c
   ToMap a b            -> ToMap <$> loadWith a <*> mapM loadWith b
   Field a b            -> Field <$> loadWith a <*> pure b
@@ -1094,45 +1094,42 @@ loadWith expr₀ = case expr₀ of
       (Note <$> pure a <*> loadWith b) `catch` handler
 
 -- | Resolve all imports within an expression
-load :: Expr Src Import -> IO (Expr Src X)
+load :: Expr Src Import -> IO (Expr Src Void)
 load = loadRelativeTo "." UseSemanticCache
 
 -- | Resolve all imports within an expression, importing relative to the given
 -- directory.
-loadRelativeTo :: FilePath -> SemanticCacheMode -> Expr Src Import -> IO (Expr Src X)
+loadRelativeTo :: FilePath -> SemanticCacheMode -> Expr Src Import -> IO (Expr Src Void)
 loadRelativeTo rootDirectory semanticCacheMode expression =
     State.evalStateT
         (loadWith expression)
         (emptyStatus rootDirectory) { _semanticCacheMode = semanticCacheMode }
 
 encodeExpression
-    :: forall s
-    .  StandardVersion
+    :: StandardVersion
     -- ^ `NoVersion` means to encode without the version tag
-    -> Expr s X
+    -> Expr Void Void
     -> Data.ByteString.ByteString
 encodeExpression _standardVersion expression = bytesStrict
   where
-    intermediateExpression :: Expr s Import
+    intermediateExpression :: Expr Void Import
     intermediateExpression = fmap absurd expression
 
-    term :: Term
-    term = Dhall.Binary.encodeExpression intermediateExpression
-
-    taggedTerm :: Term
-    taggedTerm =
+    encoding =
         case _standardVersion of
-            NoVersion -> term
-            s         -> TList [ TString v, term ]
+            NoVersion ->
+                Codec.Serialise.encode intermediateExpression
+            s ->
+                    Encoding.encodeListLen 2
+                <>  Encoding.encodeString v
+                <>  Codec.Serialise.encode intermediateExpression
               where
                 v = Dhall.Binary.renderStandardVersion s
 
-    bytesLazy = Codec.Serialise.serialise taggedTerm
-
-    bytesStrict = Data.ByteString.Lazy.toStrict bytesLazy
+    bytesStrict = Write.toStrictByteString encoding
 
 -- | Hash a fully resolved expression
-hashExpression :: Expr s X -> Dhall.Crypto.SHA256Digest
+hashExpression :: Expr Void Void -> Dhall.Crypto.SHA256Digest
 hashExpression expression =
     Dhall.Crypto.sha256Hash (encodeExpression NoVersion expression)
 
@@ -1142,7 +1139,7 @@ hashExpression expression =
     In other words, the output of this function can be pasted into Dhall
     source code to add an integrity check to an import
 -}
-hashExpressionToCode :: Expr s X -> Text
+hashExpressionToCode :: Expr Void Void -> Text
 hashExpressionToCode expr =
     "sha256:" <> Text.pack (show (hashExpression expr))
 
@@ -1153,6 +1150,6 @@ instance Show ImportResolutionDisabled where
     show _ = "\nImport resolution is disabled"
 
 -- | Assert than an expression is import-free
-assertNoImports :: MonadIO io => Expr Src Import -> io (Expr Src X)
+assertNoImports :: MonadIO io => Expr Src Import -> io (Expr Src Void)
 assertNoImports expression =
     Dhall.Core.throws (traverse (\_ -> Left ImportResolutionDisabled) expression)
