@@ -16,65 +16,88 @@ module Dhall.Main
     , parserInfoOptions
 
       -- * Execution
-    , command
+    , Dhall.Main.command
     , main
     ) where
 
 import Control.Applicative (optional, (<|>))
 import Control.Exception (Handler(..), SomeException)
+import Control.Monad (when)
+import Data.Bifunctor (first)
 import Data.List.NonEmpty (NonEmpty(..))
 import Data.Monoid ((<>))
 import Data.Text (Text)
 import Data.Text.Prettyprint.Doc (Doc, Pretty)
-import Dhall.Core (Expr(Annot), Import, pretty)
+import Data.Void (Void)
 import Dhall.Freeze (Intent(..), Scope(..))
-import Dhall.Import (Imported(..), Depends(..), SemanticCacheMode(..))
+import Dhall.Import (Imported(..), Depends(..), SemanticCacheMode(..), _semanticCacheMode)
 import Dhall.Parser (Src)
-import Dhall.Pretty (Ann, CharacterSet(..), annToAnsiStyle, layoutOpts)
-import Dhall.TypeCheck (Censored(..), DetailedTypeError(..), TypeError, X)
-import Dhall.Util (Censor(..), Input(..))
+import Dhall.Pretty (Ann, CharacterSet(..), annToAnsiStyle)
+import Dhall.TypeCheck (Censored(..), DetailedTypeError(..), TypeError)
 import Dhall.Version (dhallVersionString)
 import Options.Applicative (Parser, ParserInfo)
 import System.Exit (ExitCode, exitFailure)
 import System.IO (Handle)
 import Text.Dot ((.->.))
 
+import Dhall.Core
+    ( Expr(Annot)
+    , Import(..)
+    , ImportHashed(..)
+    , ImportType(..)
+    , URL(..)
+    , pretty
+    )
+import Dhall.Util
+    ( Censor(..)
+    , CheckFailed(..)
+    , Header (..)
+    , Input(..)
+    , OutputMode(..)
+    , Output(..)
+    )
+
 import qualified Codec.CBOR.JSON
 import qualified Codec.CBOR.Read
 import qualified Codec.CBOR.Write
-import qualified Codec.Serialise
 import qualified Control.Exception
 import qualified Control.Monad.Trans.State.Strict          as State
 import qualified Data.Aeson
 import qualified Data.Aeson.Encode.Pretty
 import qualified Data.ByteString.Lazy
 import qualified Data.ByteString.Lazy.Char8
+import qualified Data.Map
 import qualified Data.Text
 import qualified Data.Text.IO
 import qualified Data.Text.Prettyprint.Doc                 as Pretty
 import qualified Data.Text.Prettyprint.Doc.Render.Terminal as Pretty
+import qualified Data.Text.Prettyprint.Doc.Render.Text     as Pretty.Text
 import qualified Dhall
 import qualified Dhall.Binary
 import qualified Dhall.Core
 import qualified Dhall.Diff
+import qualified Dhall.DirectoryTree                       as DirectoryTree
 import qualified Dhall.Format
 import qualified Dhall.Freeze
-import qualified Dhall.Hash
 import qualified Dhall.Import
 import qualified Dhall.Import.Types
 import qualified Dhall.Lint
+import qualified Dhall.Parser                              as Parser
+import qualified Dhall.Map
+import qualified Dhall.Tags
 import qualified Dhall.Pretty
 import qualified Dhall.Repl
 import qualified Dhall.TypeCheck
 import qualified Dhall.Util
 import qualified GHC.IO.Encoding
 import qualified Options.Applicative
+import qualified System.AtomicWrite.Writer.LazyText        as AtomicWrite.LazyText
 import qualified System.Console.ANSI
 import qualified System.Exit                               as Exit
 import qualified System.IO
 import qualified System.FilePath
 import qualified Text.Dot
-import qualified Data.Map
+import qualified Text.Pretty.Simple
 
 -- | Top-level program options
 data Options = Options
@@ -85,28 +108,51 @@ data Options = Options
     , censor  :: Censor
     }
 
+ignoreSemanticCache :: Mode -> Bool
+ignoreSemanticCache Default {..} = semanticCacheMode == IgnoreSemanticCache
+ignoreSemanticCache Resolve {..} = semanticCacheMode == IgnoreSemanticCache
+ignoreSemanticCache Type {..}    = semanticCacheMode == IgnoreSemanticCache
+ignoreSemanticCache _            = False
+
 -- | The subcommands for the @dhall@ executable
 data Mode
     = Default
           { file :: Input
+          , output :: Output
           , annotate :: Bool
           , alpha :: Bool
           , semanticCacheMode :: SemanticCacheMode
           , version :: Bool
           }
     | Version
-    | Resolve { file :: Input, resolveMode :: Maybe ResolveMode }
-    | Type { file :: Input, quiet :: Bool }
+    | Resolve
+          { file :: Input
+          , resolveMode :: Maybe ResolveMode
+          , semanticCacheMode :: SemanticCacheMode
+          }
+    | Type
+          { file :: Input
+          , quiet :: Bool
+          , semanticCacheMode :: SemanticCacheMode
+          }
     | Normalize { file :: Input , alpha :: Bool }
     | Repl
-    | Format { formatMode :: Dhall.Format.FormatMode }
-    | Freeze { inplace :: Input, all_ :: Bool, cache :: Bool }
-    | Hash
+    | Format { input :: Input, outputMode :: OutputMode }
+    | Freeze { input :: Input, all_ :: Bool, cache :: Bool, outputMode :: OutputMode }
+    | Hash { file :: Input }
     | Diff { expr1 :: Text, expr2 :: Text }
-    | Lint { inplace :: Input }
+    | Lint { input :: Input, outputMode :: OutputMode }
+    | Tags
+          { input :: Input
+          , output :: Output
+          , suffixes :: Maybe [Text]
+          , followSymlinks :: Bool
+          }
     | Encode { file :: Input, json :: Bool }
     | Decode { file :: Input, json :: Bool }
     | Text { file :: Input }
+    | DirectoryTree { file :: Input, path :: FilePath }
+    | SyntaxTree { file :: Input }
 
 data ResolveMode
     = Dot
@@ -135,11 +181,12 @@ parseOptions =
         f True  = Censor
         f False = NoCensor
 
-subcommand :: String -> String -> Parser a -> Parser a
-subcommand name description parser =
+subcommand' :: Bool -> String -> String -> Parser a -> Parser a
+subcommand' internal name description parser =
     Options.Applicative.hsubparser
         (   Options.Applicative.command name parserInfo
         <>  Options.Applicative.metavar name
+        <>  if internal then Options.Applicative.internal else mempty
         )
   where
     parserInfo =
@@ -148,6 +195,12 @@ subcommand name description parser =
             <>  Options.Applicative.progDesc description
             )
 
+subcommand :: String -> String -> Parser a -> Parser a
+subcommand = subcommand' False
+
+internalSubcommand :: String -> String -> Parser a -> Parser a
+internalSubcommand = subcommand' True
+
 parseMode :: Parser Mode
 parseMode =
         subcommand
@@ -157,11 +210,11 @@ parseMode =
     <|> subcommand
             "resolve"
             "Resolve an expression's imports"
-            (Resolve <$> parseFile <*> parseResolveMode)
+            (Resolve <$> parseFile <*> parseResolveMode <*> parseSemanticCacheMode)
     <|> subcommand
             "type"
             "Infer an expression's type"
-            (Type <$> parseFile <*> parseQuiet)
+            (Type <$> parseFile <*> parseQuiet <*> parseSemanticCacheMode)
     <|> subcommand
             "normalize"
             "Normalize an expression"
@@ -177,19 +230,23 @@ parseMode =
     <|> subcommand
             "hash"
             "Compute semantic hashes for Dhall expressions"
-            (pure Hash)
+            (Hash <$> parseFile)
     <|> subcommand
             "lint"
             "Improve Dhall code by using newer language features and removing dead code"
-            (Lint <$> parseInplace)
+            (Lint <$> parseInplace <*> parseCheck)
+    <|> subcommand
+            "tags"
+            "Generate etags file"
+            (Tags <$> parseInput <*> parseTagsOutput <*> parseSuffixes <*> parseFollowSymlinks)
     <|> subcommand
             "format"
             "Standard code formatter for the Dhall language"
-            (Format <$> parseFormatMode)
+            (Format <$> parseInplace <*> parseCheck)
     <|> subcommand
             "freeze"
             "Add integrity checks to remote import statements of an expression"
-            (Freeze <$> parseInplace <*> parseAllFlag <*> parseCacheFlag)
+            (Freeze <$> parseInplace <*> parseAllFlag <*> parseCacheFlag <*> parseCheck)
     <|> subcommand
             "encode"
             "Encode a Dhall expression to binary"
@@ -202,8 +259,17 @@ parseMode =
             "text"
             "Render a Dhall expression that evaluates to a Text literal"
             (Text <$> parseFile)
+    <|> subcommand
+            "to-directory-tree"
+            "Convert nested records of Text literals into a directory tree"
+            (DirectoryTree <$> parseFile <*> parseDirectoryTreeOutput)
+    <|> internalSubcommand
+            "haskell-syntax-tree"
+            "Output the parsed syntax tree (for debugging)"
+            (SyntaxTree <$> parseFile)
     <|> (   Default
         <$> parseFile
+        <*> parseOutput
         <*> parseAnnotate
         <*> parseAlpha
         <*> parseSemanticCacheMode
@@ -218,7 +284,7 @@ parseMode =
     parseFile = fmap f (optional p)
       where
         f  Nothing    = StandardInput
-        f (Just file) = File file
+        f (Just file) = InputFile file
 
         p = Options.Applicative.strOption
                 (   Options.Applicative.long "file"
@@ -226,6 +292,17 @@ parseMode =
                 <>  Options.Applicative.metavar "FILE"
                 )
 
+    parseOutput = fmap f (optional p)
+      where
+        f Nothing = StandardOutput
+        f (Just file) = OutputFile file
+
+        p = Options.Applicative.strOption
+                (   Options.Applicative.long "output"
+                <>  Options.Applicative.help "Write result to a file instead of standard output"
+                <>  Options.Applicative.metavar "FILE"
+                )
+
     parseAlpha =
         Options.Applicative.switch
             (   Options.Applicative.long "alpha"
@@ -269,7 +346,7 @@ parseMode =
           Options.Applicative.flag' (Just ListTransitiveDependencies)
               (   Options.Applicative.long "transitive-dependencies"
               <>  Options.Applicative.help
-                    "List transitive import dependencies"
+                    "List transitive import dependencies in post-order"
               )
         <|> pure Nothing
 
@@ -282,7 +359,7 @@ parseMode =
     parseInplace = fmap f (optional p)
       where
         f  Nothing    = StandardInput
-        f (Just file) = File file
+        f (Just file) = InputFile file
 
         p = Options.Applicative.strOption
             (   Options.Applicative.long "inplace"
@@ -290,6 +367,46 @@ parseMode =
             <>  Options.Applicative.metavar "FILE"
             )
 
+    parseInput = fmap f (optional p)
+      where
+        f  Nothing    = StandardInput
+        f (Just path) = InputFile path
+
+        p = Options.Applicative.strOption
+            (   Options.Applicative.long "path"
+            <>  Options.Applicative.help "Index all files in path recursively. Will get list of files from STDIN if omitted."
+            <>  Options.Applicative.metavar "PATH"
+            )
+
+    parseTagsOutput = fmap f (optional p)
+      where
+        f  Nothing    = OutputFile "tags"
+        f (Just file) = OutputFile file
+
+        p = Options.Applicative.strOption
+            (   Options.Applicative.long "output"
+            <>  Options.Applicative.help "The name of the file that the tags are written to. Defaults to \"tags\""
+            <>  Options.Applicative.metavar "FILENAME"
+            )
+
+    parseSuffixes = fmap f (optional p)
+      where
+        f  Nothing    = Just [".dhall"]
+        f (Just "")   = Nothing
+        f (Just line) = Just (Data.Text.splitOn " " line)
+
+        p = Options.Applicative.strOption
+            (   Options.Applicative.long "suffixes"
+            <>  Options.Applicative.help "Index only files with suffixes. \"\" to index all files."
+            <>  Options.Applicative.metavar "SUFFIXES"
+            )
+
+    parseFollowSymlinks =
+        Options.Applicative.switch
+        (   Options.Applicative.long "follow-symlinks"
+        <>  Options.Applicative.help "Follow symlinks when recursing directories"
+        )
+
     parseJSONFlag =
         Options.Applicative.switch
         (   Options.Applicative.long "json"
@@ -308,16 +425,23 @@ parseMode =
         <>  Options.Applicative.help "Add fallback unprotected imports when using integrity checks purely for caching purposes"
         )
 
-    parseCheck =
-        Options.Applicative.switch
-        (   Options.Applicative.long "check"
-        <>  Options.Applicative.help "Only check if the input is formatted"
-        )
-
-    parseFormatMode = adapt <$> parseCheck <*> parseInplace
+    parseCheck = fmap adapt switch
       where
-        adapt True  path    = Dhall.Format.Check {..}
-        adapt False inplace = Dhall.Format.Modify {..}
+        adapt True  = Check
+        adapt False = Write
+
+        switch =
+            Options.Applicative.switch
+            (   Options.Applicative.long "check"
+            <>  Options.Applicative.help "Only check if the input is formatted"
+            )
+
+    parseDirectoryTreeOutput =
+        Options.Applicative.strOption
+            (   Options.Applicative.long "output"
+            <>  Options.Applicative.help "The destination path to create"
+            <>  Options.Applicative.metavar "PATH"
+            )
 
 -- | `ParserInfo` for the `Options` type
 parserInfoOptions :: ParserInfo Options
@@ -328,6 +452,13 @@ parserInfoOptions =
         <>  Options.Applicative.fullDesc
         )
 
+noHeaders :: Import -> Import
+noHeaders
+    (Import { importHashed = ImportHashed { importType = Remote URL{ .. }, ..}, .. }) =
+    Import { importHashed = ImportHashed { importType = Remote URL{ headers = Nothing, .. }, .. }, .. }
+noHeaders i =
+    i
+
 -- | Run the command specified by the `Options` type
 command :: Options -> IO ()
 command (Options {..}) = do
@@ -338,7 +469,7 @@ command (Options {..}) = do
     GHC.IO.Encoding.setLocaleEncoding System.IO.utf8
 
     let rootDirectory = \case
-            File f        -> System.FilePath.takeDirectory f
+            InputFile f   -> System.FilePath.takeDirectory f
             StandardInput -> "."
 
     let toStatus = Dhall.Import.emptyStatus . rootDirectory
@@ -363,7 +494,7 @@ command (Options {..}) = do
                 System.Exit.exitFailure
 
             handleTypeError e = Control.Exception.handle handleAll $ do
-                let _ = e :: TypeError Src X
+                let _ = e :: TypeError Src Void
                 System.IO.hPutStrLn System.IO.stderr ""
                 if explain
                     then
@@ -378,7 +509,7 @@ command (Options {..}) = do
                             NoCensor -> Control.Exception.throwIO e
 
             handleImported (Imported ps e) = Control.Exception.handle handleAll $ do
-                let _ = e :: TypeError Src X
+                let _ = e :: TypeError Src Void
                 System.IO.hPutStrLn System.IO.stderr ""
                 if explain
                     then Control.Exception.throwIO (Imported ps (DetailedTypeError e))
@@ -391,7 +522,7 @@ command (Options {..}) = do
 
     let renderDoc :: Handle -> Doc Ann -> IO ()
         renderDoc h doc = do
-            let stream = Pretty.layoutSmart layoutOpts doc
+            let stream = Dhall.Pretty.layout doc
 
             supportsANSI <- System.Console.ANSI.hSupportsANSI h
             let ansiStream =
@@ -407,8 +538,14 @@ command (Options {..}) = do
             let doc = Dhall.Pretty.prettyCharacterSet characterSet expression
 
             renderDoc h doc
-    
-    Dhall.Import.warnAboutMissingCaches
+
+    let writeDocToFile :: FilePath -> Doc ann -> IO ()
+        writeDocToFile file doc = do
+            let stream = Dhall.Pretty.layout (doc <> "\n")
+
+            AtomicWrite.LazyText.atomicWriteFile file (Pretty.Text.renderLazy stream)
+
+    when (not $ ignoreSemanticCache mode) Dhall.Import.warnAboutMissingCaches
 
     handle $ case mode of
         Version -> do
@@ -440,13 +577,19 @@ command (Options {..}) = do
                         then Annot alphaNormalizedExpression inferredType
                         else alphaNormalizedExpression
 
-            render System.IO.stdout annotatedExpression
+            case output of
+                StandardOutput -> render System.IO.stdout annotatedExpression
+
+                OutputFile file_ ->
+                    writeDocToFile
+                        file_
+                        (Dhall.Pretty.prettyCharacterSet characterSet annotatedExpression)
 
         Resolve { resolveMode = Just Dot, ..} -> do
             expression <- getExpression file
 
             (Dhall.Import.Types.Status { _graph, _stack }) <-
-                State.execStateT (Dhall.Import.loadWith expression) (toStatus file)
+                State.execStateT (Dhall.Import.loadWith expression) (toStatus file) { _semanticCacheMode = semanticCacheMode }
 
             let (rootImport :| _) = _stack
                 imports = rootImport : map parent _graph ++ map child _graph
@@ -455,10 +598,12 @@ command (Options {..}) = do
             let dotNode (i, nodeId) =
                     Text.Dot.userNode
                         nodeId
-                        [ ("label", Data.Text.unpack $ pretty i)
+                        [ ("label", Data.Text.unpack $ pretty (convert i))
                         , ("shape", "box")
                         , ("style", "rounded")
                         ]
+                  where
+                    convert = noHeaders . Dhall.Import.chainedImport
 
             let dotEdge (Depends parent child) =
                     case (Data.Map.lookup parent importIds, Data.Map.lookup child importIds) of
@@ -474,29 +619,28 @@ command (Options {..}) = do
         Resolve { resolveMode = Just ListImmediateDependencies, ..} -> do
             expression <- getExpression file
 
-            mapM_ (print
-                        . Pretty.pretty
-                        . Dhall.Core.importHashed) expression
+            mapM_ (print . Pretty.pretty . noHeaders) expression
 
         Resolve { resolveMode = Just ListTransitiveDependencies, ..} -> do
             expression <- getExpression file
 
             (Dhall.Import.Types.Status { _cache }) <-
-                State.execStateT (Dhall.Import.loadWith expression) (toStatus file)
+                State.execStateT (Dhall.Import.loadWith expression) (toStatus file) { _semanticCacheMode = semanticCacheMode }
 
             mapM_ print
-                 .   fmap (   Pretty.pretty
-                          .   Dhall.Core.importType
-                          .   Dhall.Core.importHashed
-                          .   Dhall.Import.chainedImport )
-                 .   Data.Map.keys
+                 .   fmap ( Pretty.pretty
+                          . noHeaders
+                          . Dhall.Import.chainedImport
+                          )
+                 .   reverse
+                 .   Dhall.Map.keys
                  $   _cache
 
         Resolve { resolveMode = Nothing, ..} -> do
             expression <- getExpression file
 
             resolvedExpression <-
-                Dhall.Import.loadRelativeTo (rootDirectory file) UseSemanticCache expression
+                Dhall.Import.loadRelativeTo (rootDirectory file) semanticCacheMode expression
 
             render System.IO.stdout resolvedExpression
 
@@ -520,7 +664,7 @@ command (Options {..}) = do
             expression <- getExpression file
 
             resolvedExpression <-
-                Dhall.Import.loadRelativeTo (rootDirectory file) UseSemanticCache expression
+                Dhall.Import.loadRelativeTo (rootDirectory file) semanticCacheMode expression
 
             inferredType <- Dhall.Core.throws (Dhall.TypeCheck.typeOf resolvedExpression)
 
@@ -552,38 +696,69 @@ command (Options {..}) = do
 
             let intent = if cache then Cache else Secure
 
-            Dhall.Freeze.freeze inplace scope intent characterSet censor
+            Dhall.Freeze.freeze outputMode input scope intent characterSet censor
 
-        Hash -> do
-            Dhall.Hash.hash censor
+        Hash {..} -> do
+            expression <- getExpression file
+
+            resolvedExpression <-
+                Dhall.Import.loadRelativeTo (rootDirectory file) UseSemanticCache expression
+
+            _ <- Dhall.Core.throws (Dhall.TypeCheck.typeOf resolvedExpression)
+
+            let normalizedExpression =
+                    Dhall.Core.alphaNormalize (Dhall.Core.normalize resolvedExpression)
+
+            Data.Text.IO.putStrLn (Dhall.Import.hashExpressionToCode normalizedExpression)
 
         Lint {..} -> do
-            (header, expression) <- getExpressionAndHeader inplace
+            case outputMode of
+                Write -> do
+                    (Header header, expression) <- do
+                        getExpressionAndHeader input
 
-            case inplace of
-                File file -> do
                     let lintedExpression = Dhall.Lint.lint expression
 
                     let doc =   Pretty.pretty header
                             <>  Dhall.Pretty.prettyCharacterSet characterSet lintedExpression
 
-                    System.IO.withFile file System.IO.WriteMode (\h -> do
-                        renderDoc h doc )
+                    case input of
+                        InputFile file -> writeDocToFile file doc
+
+                        StandardInput -> renderDoc System.IO.stdout doc
+
+                Check -> do
+                    originalText <- case input of
+                        InputFile file -> Data.Text.IO.readFile file
+                        StandardInput  -> Data.Text.IO.getContents
+
+                    let name = case input of
+                            InputFile file -> file
+                            StandardInput  -> "(stdin)"
+
+                    (Header header, expression) <- do
+                        Dhall.Core.throws (first Parser.censor (Parser.exprAndHeaderFromText name originalText))
 
-                StandardInput -> do
                     let lintedExpression = Dhall.Lint.lint expression
 
                     let doc =   Pretty.pretty header
                             <>  Dhall.Pretty.prettyCharacterSet characterSet lintedExpression
 
-                    renderDoc System.IO.stdout doc
+                    let stream = Dhall.Pretty.layout doc
+
+                    let modifiedText = Pretty.Text.renderStrict stream <> "\n"
+
+                    if originalText == modifiedText
+                        then return ()
+                        else do
+                            let modified = "linted"
+
+                            Control.Exception.throwIO CheckFailed{ command = "lint", ..}
 
         Encode {..} -> do
             expression <- getExpression file
 
-            let term = Dhall.Binary.encodeExpression expression
-
-            let bytes = Codec.Serialise.serialise term
+            let bytes = Dhall.Binary.encodeExpression (Dhall.Core.denote expression)
 
             if json
                 then do
@@ -601,10 +776,10 @@ command (Options {..}) = do
         Decode {..} -> do
             bytes <- do
                 case file of
-                    File f        -> Data.ByteString.Lazy.readFile f
+                    InputFile f   -> Data.ByteString.Lazy.readFile f
                     StandardInput -> Data.ByteString.Lazy.getContents
 
-            term <- do
+            expression <- do
                 if json
                     then do
                         value <- case Data.Aeson.eitherDecode' bytes of
@@ -613,15 +788,14 @@ command (Options {..}) = do
 
                         let encoding = Codec.CBOR.JSON.encodeValue value
 
-                        let cborBytes = Codec.CBOR.Write.toLazyByteString encoding
-                        Dhall.Core.throws (Codec.Serialise.deserialiseOrFail cborBytes)
+                        let cborgBytes = Codec.CBOR.Write.toLazyByteString encoding
+
+                        Dhall.Core.throws (Dhall.Binary.decodeExpression cborgBytes)
                     else do
-                        Dhall.Core.throws (Codec.Serialise.deserialiseOrFail bytes)
+                        Dhall.Core.throws (Dhall.Binary.decodeExpression bytes)
 
-            expression <- Dhall.Core.throws (Dhall.Binary.decodeExpression term)
-                :: IO (Expr Src Import)
 
-            let doc = Dhall.Pretty.prettyCharacterSet characterSet expression
+            let doc = Dhall.Pretty.prettyCharacterSet characterSet (Dhall.Core.renote expression :: Expr Src Import)
 
             renderDoc System.IO.stdout doc
 
@@ -639,16 +813,46 @@ command (Options {..}) = do
                 Dhall.Core.TextLit (Dhall.Core.Chunks [] text) -> do
                     Data.Text.IO.putStr text
                 _ -> do
-                    let invalidTypeExpected :: Expr X X
-                        invalidTypeExpected = Dhall.Core.Text
+                    let invalidDecoderExpected :: Expr Void Void
+                        invalidDecoderExpected = Dhall.Core.Text
 
-                    let invalidTypeExpression :: Expr X X
-                        invalidTypeExpression = normalizedExpression
+                    let invalidDecoderExpression :: Expr Void Void
+                        invalidDecoderExpression = normalizedExpression
 
-                    Control.Exception.throwIO (Dhall.InvalidType {..})
+                    Control.Exception.throwIO (Dhall.InvalidDecoder {..})
+
+        Tags {..} -> do
+            tags <- Dhall.Tags.generate input suffixes followSymlinks
+
+            case output of
+                OutputFile file ->
+                    System.IO.withFile file System.IO.WriteMode (`Data.Text.IO.hPutStr` tags)
+
+                StandardOutput -> Data.Text.IO.putStrLn tags
+
+        DirectoryTree {..} -> do
+            expression <- getExpression file
+
+            resolvedExpression <-
+                Dhall.Import.loadRelativeTo (rootDirectory file) UseSemanticCache expression
+
+            _ <- Dhall.Core.throws (Dhall.TypeCheck.typeOf resolvedExpression)
+
+            let normalizedExpression = Dhall.Core.normalize resolvedExpression
+
+            DirectoryTree.toDirectoryTree path normalizedExpression
+
+        SyntaxTree {..} -> do
+            expression <- getExpression file
+
+            let denoted :: Expr Void Import
+                denoted = Dhall.Core.denote expression
+
+            Text.Pretty.Simple.pPrintNoColor denoted
 
 -- | Entry point for the @dhall@ executable
 main :: IO ()
 main = do
     options <- Options.Applicative.execParser parserInfoOptions
-    command options
+
+    Dhall.Main.command options
