@@ -13,6 +13,7 @@ module Dhall.Map
       Map
 
       -- * Construction
+    , empty
     , singleton
     , fromList
     , fromListWithKey
@@ -33,6 +34,7 @@ module Dhall.Map
     , delete
     , filter
     , restrictKeys
+    , withoutKeys
     , mapMaybe
 
       -- * Query
@@ -58,6 +60,7 @@ module Dhall.Map
 
       -- * Conversions
     , toList
+    , toAscList
     , toMap
     , keys
     , keysSet
@@ -77,6 +80,7 @@ import qualified Data.List
 import qualified Data.Map
 import qualified Data.Set
 import qualified GHC.Exts
+import qualified Language.Haskell.TH.Syntax as Syntax
 import qualified Prelude
 
 {-| A `Map` that remembers the original ordering of keys
@@ -89,14 +93,16 @@ import qualified Prelude
 data Map k v = Map (Data.Map.Map k v) (Keys k)
     deriving (Data, Generic, NFData)
 
-instance (Data k, Data v, Lift k, Lift v, Ord k) => Lift (Map k v)
+instance (Data k, Data v, Lift k, Lift v, Ord k) => Lift (Map k v) where
+    lift = Syntax.liftData
 
 data Keys a
     = Sorted
     | Original [a]
     deriving (Data, Generic, NFData)
 
-instance (Data a, Lift a) => Lift (Keys a)
+instance (Data a, Lift a) => Lift (Keys a) where
+    lift = Syntax.liftData
 
 instance (Ord k, Eq v) => Eq (Map k v) where
   m1 == m2 =
@@ -124,6 +130,9 @@ instance Ord k => Foldable (Map k) where
   length m = size m
   {-# INLINABLE length #-}
 
+  null (Map m _) = null m
+  {-# INLINABLE null #-}
+
 instance Ord k => Traversable (Map k) where
   traverse f m = traverseWithKey (\_ v -> f v) m
   {-# INLINABLE traverse #-}
@@ -161,6 +170,10 @@ instance Ord k => GHC.Exts.IsList (Map k v) where
 
     toList = Dhall.Map.toList
 
+-- | Create an empty `Map`
+empty :: Ord k => Map k v
+empty = mempty
+
 {-| Create a `Map` from a single key-value pair
 
 >>> singleton "A" 1
@@ -369,6 +382,23 @@ restrictKeys (Map m ks) s = Map m' ks'
     ks' = filterKeys (\k -> Data.Set.member k s) ks
 {-# INLINABLE restrictKeys #-}
 
+{-| Remove all keys in a @"Data.Set".'Set'@ from a 'Map'
+
+>>> withoutKeys (fromList [("A",1),("B",2)]) (Data.Set.fromList ["A"])
+fromList [("B",2)]
+-}
+withoutKeys :: Ord k => Map k a -> Data.Set.Set k -> Map k a
+withoutKeys (Map m ks) s = Map m' ks'
+  where
+#if MIN_VERSION_containers(0,5,8)
+    m' = Data.Map.withoutKeys m s
+#else
+    m' = Data.Map.filterWithKey (\k _ -> Data.Set.notMember k s) m
+#endif
+
+    ks' = filterKeys (\k -> Data.Set.notMember k s) ks
+{-# INLINABLE withoutKeys #-}
+
 {-| Transform all values in a `Map` using the supplied function, deleting the
     key if the function returns `Nothing`
 
@@ -631,6 +661,12 @@ toList (Map m Sorted)        = Data.Map.toList m
 toList (Map m (Original ks)) = fmap (\k -> (k, m Data.Map.! k)) ks
 {-# INLINABLE toList #-}
 
+{-| Convert a `Map` to a list of key-value pairs in ascending order of keys
+-}
+toAscList :: Map k v -> [(k, v)]
+toAscList (Map m _) = Data.Map.toAscList m
+{-# INLINABLE toAscList #-}
+
 {-| Convert a @"Dhall.Map".`Map`@ to a @"Data.Map".`Data.Map.Map`@
 
 >>> toMap (fromList [("B",1),("A",2)]) -- Order is lost upon conversion
