@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP             #-}
 {-# LANGUAGE RecordWildCards #-}
 
 -- | This module contains Dhall's parsing logic
@@ -8,11 +7,13 @@ module Dhall.Parser (
       exprFromText
     , exprAndHeaderFromText
     , censor
+    , createHeader
 
     -- * Parsers
     , expr, exprA
 
     -- * Types
+    , Header(..)
     , Src(..)
     , SourcedException(..)
     , ParseError(..)
@@ -23,7 +24,7 @@ import Control.Exception (Exception)
 import Data.Semigroup (Semigroup(..))
 import Data.Text (Text)
 import Data.Void (Void)
-import Dhall.Core
+import Dhall.Syntax
 import Dhall.Src (Src(..))
 import Prelude hiding (const, pi)
 import Text.Megaparsec (ParseErrorBundle(..), PosState(..))
@@ -34,7 +35,7 @@ import qualified Dhall.Core      as Core
 import qualified Text.Megaparsec
 
 import Dhall.Parser.Combinators
-import Dhall.Parser.Token
+import Dhall.Parser.Token hiding (text)
 import Dhall.Parser.Expression
 
 -- | Parser for a top-level Dhall expression
@@ -48,11 +49,7 @@ exprA = completeExpression
 
 -- | A parsing error
 data ParseError = ParseError {
-#if MIN_VERSION_megaparsec(7, 0, 0)
       unwrap :: Text.Megaparsec.ParseErrorBundle Text Void
-#else
-      unwrap :: Text.Megaparsec.ParseError Char Void
-#endif
     , input  :: Text
     }
 
@@ -76,11 +73,7 @@ censor parseError =
 
 instance Show ParseError where
     show (ParseError {..}) =
-#if MIN_VERSION_megaparsec(7, 0, 0)
       "\n\ESC[1;31mError\ESC[0m: Invalid input\n\n" <> Text.Megaparsec.errorBundlePretty unwrap
-#else
-      "\n\ESC[1;31mError\ESC[0m: Invalid input\n\n" <> Text.Megaparsec.parseErrorPretty unwrap
-#endif
 
 instance Exception ParseError
 
@@ -92,6 +85,17 @@ exprFromText
   -> Either ParseError (Expr Src Import)
 exprFromText delta text = fmap snd (exprAndHeaderFromText delta text)
 
+-- | A header corresponds to the leading comment at the top of a Dhall file.
+--
+-- The header includes comment characters but is stripped of leading spaces and
+-- trailing newlines
+newtype Header = Header Text deriving Show
+
+-- | Create a header with stripped leading spaces and trailing newlines
+createHeader :: Text -> Header
+createHeader =
+    Header . Data.Text.dropWhile Data.Char.isSpace . Data.Text.dropWhileEnd (/= '\n')
+
 {-| Like `exprFromText` but also returns the leading comments and whitespace
     (i.e. header) up to the last newline before the code begins
 
@@ -108,10 +112,10 @@ exprAndHeaderFromText
     :: String -- ^ User-friendly name describing the input expression,
               --   used in parsing error messages
     -> Text   -- ^ Input expression to parse
-    -> Either ParseError (Text, Expr Src Import)
+    -> Either ParseError (Header, Expr Src Import)
 exprAndHeaderFromText delta text = case result of
     Left errInfo   -> Left (ParseError { unwrap = errInfo, input = text })
-    Right (txt, r) -> Right (stripHeader txt, r)
+    Right (txt, r) -> Right (createHeader txt, r)
   where
     parser = do
         (bytes, _) <- Text.Megaparsec.match whitespace
@@ -120,5 +124,3 @@ exprAndHeaderFromText delta text = case result of
         return (bytes, r)
 
     result = Text.Megaparsec.parse (unParser parser) delta text
-
-    stripHeader = Data.Text.dropWhile Data.Char.isSpace . Data.Text.dropWhileEnd (/= '\n')
