@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP                 #-}
 {-# LANGUAGE NamedFieldPuns      #-}
 {-# LANGUAGE RecordWildCards     #-}
 {-# LANGUAGE OverloadedLists     #-}
@@ -8,65 +7,55 @@
 -- | Parsing Dhall expressions.
 module Dhall.Parser.Expression where
 
-import Control.Applicative (Alternative(..), optional)
+import Control.Applicative (liftA2, Alternative(..), optional)
 import Data.ByteArray.Encoding (Base(..))
+import Data.Foldable (foldl')
 import Data.Functor (void)
-import Data.List.NonEmpty (NonEmpty(..))
 import Data.Semigroup (Semigroup(..))
 import Data.Text (Text)
-import Dhall.Core
+import Dhall.Syntax
 import Dhall.Src (Src(..))
 import Prelude hiding (const, pi)
 import Text.Parser.Combinators (choice, try, (<?>))
 
 import qualified Control.Monad
+import qualified Control.Monad.Combinators.NonEmpty as Combinators.NonEmpty
 import qualified Data.ByteArray.Encoding
 import qualified Data.ByteString
-import qualified Data.Char               as Char
-import qualified Data.Foldable
+import qualified Data.Char                          as Char
 import qualified Data.List
-import qualified Data.List.NonEmpty
+import qualified Data.List.NonEmpty                 as NonEmpty
 import qualified Data.Sequence
 import qualified Data.Text
 import qualified Data.Text.Encoding
 import qualified Dhall.Crypto
 import qualified Text.Megaparsec
-#if !MIN_VERSION_megaparsec(7, 0, 0)
-import qualified Text.Megaparsec.Char    as Text.Megaparsec
-#endif
-import qualified Text.Parser.Char
 
 import Dhall.Parser.Combinators
 import Dhall.Parser.Token
 
+-- | Get the current source position
 getSourcePos :: Text.Megaparsec.MonadParsec e s m =>
                 m Text.Megaparsec.SourcePos
 getSourcePos =
-#if MIN_VERSION_megaparsec(7, 0, 0)
     Text.Megaparsec.getSourcePos
-#else
-    Text.Megaparsec.getPosition
-#endif
 {-# INLINE getSourcePos #-}
 
+-- | Get the current source offset (in tokens)
 getOffset :: Text.Megaparsec.MonadParsec e s m => m Int
-#if MIN_VERSION_megaparsec(7, 0, 0)
 getOffset = Text.Megaparsec.stateOffset <$> Text.Megaparsec.getParserState
-#else
-getOffset = Text.Megaparsec.stateTokensProcessed <$> Text.Megaparsec.getParserState
-#endif
 {-# INLINE getOffset #-}
 
+-- | Set the current source offset
 setOffset :: Text.Megaparsec.MonadParsec e s m => Int -> m ()
-#if MIN_VERSION_megaparsec(7, 0, 0)
-setOffset o = Text.Megaparsec.updateParserState $ \(Text.Megaparsec.State s _ pst) ->
-  Text.Megaparsec.State s o pst
-#else
-setOffset o = Text.Megaparsec.updateParserState $ \(Text.Megaparsec.State s p _ stw) ->
-  Text.Megaparsec.State s p o stw
-#endif
+setOffset o = Text.Megaparsec.updateParserState $ \state ->
+    state
+        { Text.Megaparsec.stateOffset = o }
 {-# INLINE setOffset #-}
 
+{-| Wrap a `Parser` to still match the same text but return only the `Src`
+    span
+-}
 src :: Parser a -> Parser Src
 src parser = do
     before      <- getSourcePos
@@ -74,6 +63,9 @@ src parser = do
     after       <- getSourcePos
     return (Src before after tokens)
 
+{-| Wrap a `Parser` to still match the same text, but to wrap the resulting
+    `Expr` in a `Note` constructor containing the `Src` span
+-}
 noted :: Parser (Expr Src a) -> Parser (Expr Src a)
 noted parser = do
     before      <- getSourcePos
@@ -84,27 +76,43 @@ noted parser = do
         Note src₁ _ | laxSrcEq src₀ src₁ -> return e
         _                                -> return (Note src₀ e)
 
+{-| Parse a complete expression (with leading and trailing whitespace)
+
+    This corresponds to the @complete-expression@ rule from the official
+    grammar
+-}
 completeExpression :: Parser a -> Parser (Expr Src a)
 completeExpression embedded = completeExpression_
   where
     Parsers {..} = parsers embedded
 
+{-| Parse an \"import expression\"
+
+    This is not the same thing as @`fmap` `Embed`@.  This parses any
+    expression of the same or higher precedence as an import expression (such
+    as a selector expression).  For example, this parses @(1)@
+
+    This corresponds to the @import-expression@ rule from the official grammar
+-}
 importExpression :: Parser a -> Parser (Expr Src a)
 importExpression embedded = importExpression_
   where
     Parsers {..} = parsers embedded
 
+{-| For efficiency (and simplicity) we only expose two parsers from the
+    result of the `parsers` function, since these are the only parsers needed
+    outside of this module
+-}
 data Parsers a = Parsers
     { completeExpression_ :: Parser (Expr Src a)
     , importExpression_   :: Parser (Expr Src a)
     }
 
+-- | Given a parser for imports, 
 parsers :: Parser a -> Parsers a
 parsers embedded = Parsers {..}
   where
-    completeExpression_ = do
-        whitespace
-        expression
+    completeExpression_ = whitespace *> expression <* whitespace
 
     expression =
         noted
@@ -120,53 +128,65 @@ parsers embedded = Parsers {..}
       where
         alternative0 = do
             _lambda
+            whitespace
             _openParens
+            whitespace
             a <- label
+            whitespace
             _colon
+            nonemptyWhitespace
             b <- expression
+            whitespace
             _closeParens
+            whitespace
             _arrow
+            whitespace
             c <- expression
             return (Lam a b c)
 
         alternative1 = do
-            _if
+            try (_if *> nonemptyWhitespace)
             a <- expression
-            _then
+            whitespace
+            try (_then *> nonemptyWhitespace)
             b <- expression
-            _else
+            whitespace
+            try (_else *> nonemptyWhitespace)
             c <- expression
             return (BoolIf a b c)
 
         alternative2 = do
             let binding = do
-                    _letOnly
+                    src0 <- try (_let *> src nonemptyWhitespace)
 
-                    src0 <- src nonemptyWhitespace
-
-                    c <- labelOnly
+                    c <- label
 
                     src1 <- src whitespace
 
                     d <- optional (do
-                        _colonOnly
+                        _colon
 
                         src2 <- src nonemptyWhitespace
 
                         e <- expression
+
+                        whitespace
+
                         return (Just src2, e) )
 
-                    _equalOnly
+                    _equal
 
                     src3 <- src whitespace
 
                     f <- expression
 
+                    whitespace
+
                     return (Binding (Just src0) c (Just src1) d (Just src3) f)
 
-            as <- Data.List.NonEmpty.some1 binding
+            as <- NonEmpty.some1 binding
 
-            _in
+            try (_in *> nonemptyWhitespace)
 
             b <- expression
 
@@ -187,22 +207,27 @@ parsers embedded = Parsers {..}
             --     (Let y …
             --       (Note …
             --         (Let z …
-            return (Dhall.Core.wrapInLets as b)
+            return (Dhall.Syntax.wrapInLets as b)
 
         alternative3 = do
-            _forall
-            _openParens
+            try (_forall *> whitespace *> _openParens)
+            whitespace
             a <- label
+            whitespace
             _colon
+            nonemptyWhitespace
             b <- expression
+            whitespace
             _closeParens
+            whitespace
             _arrow
+            whitespace
             c <- expression
             return (Pi a b c)
 
         alternative4 = do
-            _assert
-            _colon
+            try (_assert *> whitespace *> _colon)
+            nonemptyWhitespace
             a <- expression
             return (Assert a)
 
@@ -211,81 +236,78 @@ parsers embedded = Parsers {..}
 
             let alternative4A = do
                     _arrow
+                    whitespace
                     b <- expression
+                    whitespace
                     return (Pi "_" a b)
 
             let alternative4B = do
                     _colon
-
+                    nonemptyWhitespace
                     b <- expression
-
                     case shallowDenote a of
-                        ListLit _ [] ->
+                        ListLit Nothing [] ->
                             return (ListLit (Just b) [])
-                        Merge c d _ ->
+                        Merge c d Nothing ->
                             return (Merge c d (Just b))
-                        ToMap c _ ->
+                        ToMap c Nothing ->
                             return (ToMap c (Just b))
                         _ -> return (Annot a b)
 
             alternative4A <|> alternative4B <|> pure a
 
-    operatorExpression = precedence0Expression
+    operatorExpression =
+        foldr makeOperatorExpression applicationExpression operatorParsers
 
-    makeOperatorExpression subExpression operatorParser =
+    makeOperatorExpression operatorParser subExpression =
             noted (do
                 a <- subExpression
+
+                whitespace
+
                 b <- Text.Megaparsec.many $ do
                     op <- operatorParser
+
                     r  <- subExpression
 
+                    whitespace
+
                     return (\l -> l `op` r)
-                return (foldl (\x f -> f x) a b) )
+                return (foldl' (\x f -> f x) a b))
 
-    precedence0Operator =
-                ImportAlt   <$ _importAlt
-            <|> BoolOr      <$ _or
-            <|> TextAppend  <$ _textAppend
-            <|> NaturalPlus <$ _plus
-            <|> ListAppend  <$ _listAppend
-
-    precedence1Operator =
-                BoolAnd     <$ _and
-            <|> Combine     <$ _combine
-
-    precedence2Operator =
-                CombineTypes <$ _combineTypes
-            <|> Prefer       <$ _prefer
-            <|> NaturalTimes <$ _times
-            <|> BoolEQ       <$ _doubleEqual
-
-    precedence3Operator
-            =   BoolNE     <$ _notEqual
-            <|> Equivalent <$ _equivalent
-
-    precedence0Expression =
-            makeOperatorExpression precedence1Expression precedence0Operator
-
-    precedence1Expression =
-            makeOperatorExpression precedence2Expression precedence1Operator
-
-    precedence2Expression =
-            makeOperatorExpression precedence3Expression precedence2Operator
-
-    precedence3Expression =
-            makeOperatorExpression applicationExpression precedence3Operator
+    operatorParsers :: [Parser (Expr s a -> Expr s a -> Expr s a)]
+    operatorParsers =
+        [ ImportAlt       <$ _importAlt    <* nonemptyWhitespace
+        , BoolOr          <$ _or           <* whitespace
+        , NaturalPlus     <$ _plus         <* nonemptyWhitespace
+        , TextAppend      <$ _textAppend   <* whitespace
+        , ListAppend      <$ _listAppend   <* whitespace
+        , BoolAnd         <$ _and          <* whitespace
+        , Combine Nothing <$ _combine      <* whitespace
+        , Prefer          <$ _prefer       <* whitespace
+        , CombineTypes    <$ _combineTypes <* whitespace
+        , NaturalTimes    <$ _times        <* whitespace
+        , BoolEQ          <$ _doubleEqual  <* whitespace
+        , BoolNE          <$ _notEqual     <* whitespace
+        , Equivalent      <$ _equivalent   <* whitespace
+        ]
 
     applicationExpression = do
-            f <-    (do _Some; return Some)
+            f <-    (Some <$ try (_Some <* nonemptyWhitespace))
                 <|> return id
             a <- noted importExpression_
-            b <- Text.Megaparsec.many (noted importExpression_)
-            return (foldl app (f a) b)
+            bs <- Text.Megaparsec.many . try $ do
+                (sep, _) <- Text.Megaparsec.match nonemptyWhitespace
+                b <- importExpression_
+                return (sep, b)
+            return (foldl' app (f a) bs)
           where
-            app nL@(Note (Src before _ bytesL) _) nR@(Note (Src _ after bytesR) _) =
-                Note (Src before after (bytesL <> bytesR)) (App nL nR)
-            app nL nR =
-                App nL nR
+            app a (sep, b)
+                | Note (Src left _ bytesL) _ <- a
+                , Note (Src _ right bytesR) _ <- b
+                = Note (Src left right (bytesL <> sep <> bytesR)) (App a b)
+            app a (_, b) =
+                App a b
 
     importExpression_ = noted (choice [ alternative0, alternative1 ])
           where
@@ -293,12 +315,26 @@ parsers embedded = Parsers {..}
                 a <- embedded
                 return (Embed a)
 
-            alternative1 = selectorExpression
+            alternative1 = completionExpression
+
+    completionExpression = noted (do
+        a <- selectorExpression
+
+        mb <- optional (do
+            try (whitespace *> _doubleColon)
+
+            whitespace
+
+            selectorExpression )
+
+        case mb of
+            Nothing -> return a
+            Just b  -> return (RecordCompletion a b) )
 
     selectorExpression = noted (do
             a <- primitiveExpression
 
-            let recordType = _openParens *> expression <* _closeParens
+            let recordType = _openParens *> whitespace *> expression <* whitespace <* _closeParens
 
             let field               x  e = Field   e  x
             let projectBySet        xs e = Project e (Left  xs)
@@ -309,8 +345,8 @@ parsers embedded = Parsers {..}
                     <|> fmap projectBySet        labels
                     <|> fmap projectByExpression recordType
 
-            b <- Text.Megaparsec.many (try (do _dot; alternatives))
-            return (foldl (\e k -> k e) a b) )
+            b <- Text.Megaparsec.many (try (whitespace *> _dot *> whitespace *> alternatives))
+            return (foldl' (\e k -> k e) a b) )
 
     primitiveExpression =
             noted
@@ -326,8 +362,7 @@ parsers embedded = Parsers {..}
                     , alternative08
                     , alternative37
                     , alternative09
-
-                    , builtin <?> "built-in expression"
+                    , builtin
                     ]
                 )
             <|> alternative38
@@ -338,7 +373,7 @@ parsers embedded = Parsers {..}
                 b <- if isInfinite a
                        then setOffset n *> fail "double out of bounds"
                        else return a
-                return (DoubleLit b)
+                return (DoubleLit (DhallDouble b))
 
             alternative01 = do
                 a <- try naturalLiteral
@@ -352,28 +387,38 @@ parsers embedded = Parsers {..}
 
             alternative04 = (do
                 _openBrace
+
+                whitespace
+
+                _ <- optional (_comma *> whitespace)
+
                 a <- recordTypeOrLiteral
+
+                whitespace
+
                 _closeBrace
-                return a ) <?> "record type or literal"
+
+                return a ) <?> "literal"
 
             alternative05 = unionType
 
             alternative06 = listLiteral
 
             alternative07 = do
-                _merge
+                try (_merge *> nonemptyWhitespace)
                 a <- importExpression_
+                nonemptyWhitespace
                 b <- importExpression_ <?> "second argument to ❰merge❱"
                 return (Merge a b Nothing)
 
             alternative08 = do
-                _toMap
+                try (_toMap *> nonemptyWhitespace)
                 a <- importExpression_
                 return (ToMap a Nothing)
 
             alternative09 = do
                 a <- try doubleInfinity
-                return (DoubleLit a)
+                return (DoubleLit (DhallDouble a))
 
             builtin = do
                 let predicate c =
@@ -388,7 +433,7 @@ parsers embedded = Parsers {..}
                         ||  c == 'F'
                         ||  c == 'K'
 
-                let nan = (0.0/0.0)
+                let nan = DhallDouble (0.0/0.0)
 
                 c <- Text.Megaparsec.lookAhead (Text.Megaparsec.satisfy predicate)
 
@@ -409,7 +454,9 @@ parsers embedded = Parsers {..}
                             ]
                     'I' ->
                         choice
-                            [ IntegerShow      <$ _IntegerShow
+                            [ IntegerClamp     <$ _IntegerClamp
+                            , IntegerNegate    <$ _IntegerNegate
+                            , IntegerShow      <$ _IntegerShow
                             , IntegerToDouble  <$ _IntegerToDouble
                             , Integer          <$ _Integer
                             ]
@@ -455,7 +502,9 @@ parsers embedded = Parsers {..}
 
             alternative38 = do
                 _openParens
+                whitespace
                 a <- expression
+                whitespace
                 _closeParens
                 return a
 
@@ -468,9 +517,9 @@ parsers embedded = Parsers {..}
                 ]
           where
             interpolation = do
-                _ <- Text.Parser.Char.text "${"
+                _ <- text "${"
                 e <- completeExpression_
-                _ <- Text.Parser.Char.char '}'
+                _ <- char '}'
                 return (Chunks [(mempty, e)] mempty)
 
             unescapedCharacterFast = do
@@ -484,11 +533,11 @@ parsers embedded = Parsers {..}
                     ) && c /= '$'
 
             unescapedCharacterSlow = do
-                _ <- Text.Parser.Char.char '$'
+                _ <- char '$'
                 return (Chunks [] "$")
 
             escapedCharacter = do
-                _ <- Text.Parser.Char.char '\\'
+                _ <- char '\\'
                 c <- choice
                     [ quotationMark
                     , dollarSign
@@ -503,53 +552,60 @@ parsers embedded = Parsers {..}
                     ]
                 return (Chunks [] (Data.Text.singleton c))
               where
-                quotationMark = Text.Parser.Char.char '"'
+                quotationMark = char '"'
 
-                dollarSign = Text.Parser.Char.char '$'
+                dollarSign = char '$'
 
-                backSlash = Text.Parser.Char.char '\\'
+                backSlash = char '\\'
 
-                forwardSlash = Text.Parser.Char.char '/'
+                forwardSlash = char '/'
 
-                backSpace = do _ <- Text.Parser.Char.char 'b'; return '\b'
+                backSpace = do _ <- char 'b'; return '\b'
 
-                formFeed = do _ <- Text.Parser.Char.char 'f'; return '\f'
+                formFeed = do _ <- char 'f'; return '\f'
 
-                lineFeed = do _ <- Text.Parser.Char.char 'n'; return '\n'
+                lineFeed = do _ <- char 'n'; return '\n'
 
-                carriageReturn = do _ <- Text.Parser.Char.char 'r'; return '\r'
+                carriageReturn = do _ <- char 'r'; return '\r'
 
-                tab = do _ <- Text.Parser.Char.char 't'; return '\t'
+                tab = do _ <- char 't'; return '\t'
 
                 unicode = do
-                    _  <- Text.Parser.Char.char 'u';
+                    _  <- char 'u';
 
                     let toNumber = Data.List.foldl' (\x y -> x * 16 + y) 0
 
-                    let fourCharacterEscapeSequence =
-                            fmap toNumber (Control.Monad.replicateM 4 hexNumber)
+                    let fourCharacterEscapeSequence = do
+                            ns <- Control.Monad.replicateM 4 hexNumber
+                            
+                            let number = toNumber ns
+
+                            Control.Monad.guard (validCodepoint number)
+                                <|> fail "Invalid Unicode code point"
+
+                            return number
 
                     let bracedEscapeSequence = do
-                            _  <- Text.Parser.Char.char '{'
+                            _  <- char '{'
                             ns <- some hexNumber
 
                             let number = toNumber ns
 
-                            Control.Monad.guard (number <= 0x10FFFF && validCodepoint (Char.chr number))
+                            Control.Monad.guard (number <= 0x10FFFD && validCodepoint number)
                                 <|> fail "Invalid Unicode code point"
 
-                            _  <- Text.Parser.Char.char '}'
+                            _  <- char '}'
 
-                            return (toNumber ns)
+                            return number
 
                     n <- bracedEscapeSequence <|> fourCharacterEscapeSequence
 
                     return (Char.chr n)
 
     doubleQuotedLiteral = do
-            _      <- Text.Parser.Char.char '"'
+            _      <- char '"'
             chunks <- Text.Megaparsec.many doubleQuotedChunk
-            _      <- Text.Parser.Char.char '"'
+            _      <- char '"'
             return (mconcat chunks)
 
     singleQuoteContinue =
@@ -570,19 +626,19 @@ parsers embedded = Parsers {..}
                     return ("''" <> b)
 
                 interpolation = do
-                    _ <- Text.Parser.Char.text "${"
+                    _ <- text "${"
                     a <- completeExpression_
-                    _ <- Text.Parser.Char.char '}'
+                    _ <- char '}'
                     b <- singleQuoteContinue
                     return (Chunks [(mempty, a)] mempty <> b)
 
                 escapeInterpolation = do
-                    _ <- Text.Parser.Char.text "''${"
+                    _ <- text "''${"
                     b <- singleQuoteContinue
                     return ("${" <> b)
 
                 endLiteral = do
-                    _ <- Text.Parser.Char.text "''"
+                    _ <- text "''"
                     return mempty
 
                 unescapedCharacterFast = do
@@ -606,25 +662,22 @@ parsers embedded = Parsers {..}
                     return (Chunks [] a <> b)
 
                 tab = do
-                    _ <- Text.Parser.Char.char '\t'
+                    _ <- char '\t' <?> "tab"
                     b <- singleQuoteContinue
                     return ("\t" <> b)
 
     singleQuoteLiteral = do
-            _ <- Text.Parser.Char.text "''"
+            _ <- text "''"
             _ <- endOfLine
             a <- singleQuoteContinue
 
-            return (toDoubleQuoted a)
+            return (Dhall.Syntax.toDoubleQuoted a)
           where
-            endOfLine =
-                    void (Text.Parser.Char.char '\n'  )
-                <|> void (Text.Parser.Char.text "\r\n")
+            endOfLine = (void (char '\n') <|> void (text "\r\n")) <?> "newline"
 
     textLiteral = (do
             literal <- doubleQuotedLiteral <|> singleQuoteLiteral
-            whitespace
-            return (TextLit literal) ) <?> "text literal"
+            return (TextLit literal) ) <?> "literal"
 
     recordTypeOrLiteral =
             choice
@@ -642,30 +695,78 @@ parsers embedded = Parsers {..}
             alternative2 = return (Record mempty)
 
     nonEmptyRecordTypeOrLiteral = do
-            a <- anyLabel
-
             let nonEmptyRecordType = do
-                    _colon
+                    a <- try (anyLabel <* whitespace <* _colon)
+
+                    nonemptyWhitespace
+
                     b <- expression
+
+                    whitespace
+
                     e <- Text.Megaparsec.many (do
                         _comma
+
+                        whitespace
+
                         c <- anyLabel
+
+                        whitespace
+
                         _colon
+
+                        nonemptyWhitespace
+
                         d <- expression
+
+                        whitespace
+
                         return (c, d) )
+
                     m <- toMap ((a, b) : e)
+
                     return (Record m)
 
-            let nonEmptyRecordLiteral = do
+            let keysValue = do
+                    keys <- Combinators.NonEmpty.sepBy1 anyLabel (try (whitespace *> _dot) *> whitespace)
+
+                    whitespace
+
                     _equal
-                    b <- expression
+
+                    whitespace
+
+                    value <- expression
+
+                    whitespace
+
+                    let cons key (key', values) =
+                            (key, RecordLit [ (key', values) ])
+
+                    let nil = (NonEmpty.last keys, value)
+
+                    return (foldr cons nil (NonEmpty.init keys))
+
+            let nonEmptyRecordLiteral = do
+                    (a, b) <- keysValue
+
                     e <- Text.Megaparsec.many (do
                         _comma
-                        c <- anyLabel
-                        _equal
-                        d <- expression
+
+                        whitespace
+
+                        (c, d) <- keysValue
+
                         return (c, d) )
-                    m <- toMap ((a, b) : e)
+
+                    {- The `flip` is necessary because `toMapWith` is internally
+                       based on `Data.Map.fromListWithKey` which combines keys
+                       in reverse order
+                    -}
+                    let combine k = liftA2 (flip (Combine (Just k)))
+
+                    m <- toMapWith combine ((a, b) : e)
+
                     return (RecordLit m)
 
             nonEmptyRecordType <|> nonEmptyRecordLiteral
@@ -673,42 +774,58 @@ parsers embedded = Parsers {..}
     unionType = (do
             _openAngle
 
+            whitespace
+
+            _ <- optional (_bar *> whitespace)
+
             let unionTypeEntry = do
                     a <- anyLabel
-                    b <- optional (do _colon; expression)
+                    whitespace
+                    b <- optional (_colon *> nonemptyWhitespace *> expression <* whitespace)
                     return (a, b)
 
-            kvs <- Text.Megaparsec.sepBy unionTypeEntry _bar
+            kvs <- Text.Megaparsec.sepBy unionTypeEntry (_bar *> whitespace)
 
             m <- toMap kvs
 
             _closeAngle
 
-            return (Union m) ) <?> "union type"
+            return (Union m) ) <?> "literal"
 
     listLiteral = (do
             _openBracket
-            a <- Text.Megaparsec.sepBy expression _comma
-            _closeBracket
-            return (ListLit Nothing (Data.Sequence.fromList a)) ) <?> "list literal"
 
+            whitespace
+
+            _ <- optional (_comma *> whitespace)
+
+            a <- Text.Megaparsec.sepBy (expression <* whitespace) (_comma *> whitespace)
+
+            _closeBracket
+
+            return (ListLit Nothing (Data.Sequence.fromList a)) ) <?> "literal"
+
+{-| Parse an environment variable import
+
+    This corresponds to the @env@ rule from the official grammar
+-}
 env :: Parser ImportType
 env = do
-    _ <- Text.Parser.Char.text "env:"
+    _ <- text "env:"
     a <- (alternative0 <|> alternative1)
-    whitespace
     return (Env a)
   where
     alternative0 = bashEnvironmentVariable
 
     alternative1 = do
-        _ <- Text.Parser.Char.char '"'
+        _ <- char '"'
         a <- posixEnvironmentVariable
-        _ <- Text.Parser.Char.char '"'
+        _ <- char '"'
         return a
 
-localRaw :: Parser ImportType
-localRaw =
+-- | Parse a local import without trailing whitespace
+localOnly :: Parser ImportType
+localOnly =
     choice
         [ parentPath
         , herePath
@@ -739,26 +856,40 @@ localRaw =
 
         return (Local Absolute file)
 
+{-| Parse a local import
+
+    This corresponds to the @local@ rule from the official grammar
+-}
 local :: Parser ImportType
 local = do
-    a <- localRaw
-    whitespace
+    a <- localOnly
     return a
 
+{-| Parse an HTTP(S) import
+
+    This corresponds to the @http@ rule from the official grammar
+-}
 http :: Parser ImportType
 http = do
     url <- httpRaw
-    whitespace
     headers <- optional (do
-        _using
+        try (whitespace *> _using *> nonemptyWhitespace)
         importExpression import_ )
     return (Remote (url { headers }))
 
+{-| Parse a `Missing` import
+
+    This corresponds to the @missing@ rule from the official grammar
+-}
 missing :: Parser ImportType
 missing = do
   _missing
   return Missing
 
+{-| Parse an `ImportType`
+
+    This corresponds to the @import-type@ rule from the official grammar
+-}
 importType_ :: Parser ImportType
 importType_ = do
     let predicate c =
@@ -768,12 +899,15 @@ importType_ = do
 
     choice [ local, http, env, missing ]
 
+{-| Parse a `Dhall.Crypto.SHA256Digest`
+
+    This corresponds to the @hash@ rule from the official grammar
+-}
 importHash_ :: Parser Dhall.Crypto.SHA256Digest
 importHash_ = do
-    _ <- Text.Parser.Char.text "sha256:"
-    text <- count 64 (satisfy hexdig <?> "hex digit")
-    whitespace
-    let strictBytes16 = Data.Text.Encoding.encodeUtf8 text
+    _ <- text "sha256:"
+    t <- count 64 (satisfy hexdig <?> "hex digit")
+    let strictBytes16 = Data.Text.Encoding.encodeUtf8 t
     strictBytes <- case Data.ByteArray.Encoding.convertFromBase Base16 strictBytes16 of
         Left  string      -> fail string
         Right strictBytes -> return (strictBytes :: Data.ByteString.ByteString)
@@ -781,12 +915,20 @@ importHash_ = do
       Nothing -> fail "Invalid sha256 hash"
       Just h  -> pure h
 
+{-| Parse an `ImportHashed`
+
+    This corresponds to the @import-hashed@ rule from the official grammar
+-}
 importHashed_ :: Parser ImportHashed
 importHashed_ = do
     importType <- importType_
-    hash       <- optional importHash_
+    hash       <- optional (try (nonemptyWhitespace *> importHash_))
     return (ImportHashed {..})
 
+{-| Parse an `Import`
+
+    This corresponds to the @import@ rule from the official grammar
+-}
 import_ :: Parser Import
 import_ = (do
     importHashed <- importHashed_
@@ -794,95 +936,6 @@ import_ = (do
     return (Import {..}) ) <?> "import"
   where
     alternative = do
-      _as
+      try (whitespace *> _as *> nonemptyWhitespace)
+
       (_Text >> pure RawText) <|> (_Location >> pure Location)
-
--- | Similar to `Dhall.Core.renderChunks` except that this doesn't bother to
--- render interpolated expressions to avoid a `Buildable a` constraint.  The
--- interpolated contents are not necessary for computing how much to dedent a
--- multi-line string
---
--- This also doesn't include the surrounding quotes since they would interfere
--- with the whitespace detection
-renderChunks :: Chunks s a -> Text
-renderChunks (Chunks a b) = foldMap renderChunk a <> b
-  where
-    renderChunk :: (Text, Expr s a) -> Text
-    renderChunk (c, _) = c <> "${x}"
-
-splitOn :: Text -> Text -> NonEmpty Text
-splitOn needle haystack =
-    case Data.Text.splitOn needle haystack of
-        []     -> "" :| []
-        t : ts -> t  :| ts
-
-linesLiteral :: Chunks s a -> NonEmpty (Chunks s a)
-linesLiteral (Chunks [] suffix) =
-    fmap (Chunks []) (splitOn "\n" suffix)
-linesLiteral (Chunks ((prefix, interpolation) : pairs₀) suffix₀) =
-    foldr
-        Data.List.NonEmpty.cons
-        (Chunks ((lastLine, interpolation) : pairs₁) suffix₁ :| chunks)
-        (fmap (Chunks []) initLines)
-  where
-    splitLines = splitOn "\n" prefix
-
-    initLines = Data.List.NonEmpty.init splitLines
-    lastLine  = Data.List.NonEmpty.last splitLines
-
-    Chunks pairs₁ suffix₁ :| chunks = linesLiteral (Chunks pairs₀ suffix₀)
-
-unlinesLiteral :: NonEmpty (Chunks s a) -> Chunks s a
-unlinesLiteral chunks =
-    Data.Foldable.fold (Data.List.NonEmpty.intersperse "\n" chunks)
-
-emptyLine :: Chunks s a -> Bool
-emptyLine (Chunks [] ""  ) = True
-emptyLine (Chunks [] "\r") = True  -- So that `\r\n` is treated as a blank line
-emptyLine  _               = False
-
-leadingSpaces :: Chunks s a -> Text
-leadingSpaces chunks = Data.Text.takeWhile isSpace firstText
-  where
-    isSpace c = c == '\x20' || c == '\x09'
-
-    firstText =
-        case chunks of
-            Chunks                []  suffix -> suffix
-            Chunks ((prefix, _) : _ ) _      -> prefix
-
-dropLiteral :: Int -> Chunks s a -> Chunks s a
-dropLiteral n (Chunks [] suffix) =
-    Chunks [] (Data.Text.drop n suffix)
-dropLiteral n (Chunks ((prefix, interpolation) : rest) suffix) =
-    Chunks ((Data.Text.drop n prefix, interpolation) : rest) suffix
-
-toDoubleQuoted :: Chunks Src a -> Chunks Src a
-toDoubleQuoted literal =
-    unlinesLiteral (fmap (dropLiteral indent) literals)
-  where
-    literals = linesLiteral literal
-
-    sharedPrefix ab ac =
-        case Data.Text.commonPrefixes ab ac of
-            Just (a, _b, _c) -> a
-            Nothing          -> ""
-
-    -- The standard specifies to filter out blank lines for all lines *except*
-    -- for the last line
-    filteredLines = newInit <> pure oldLast
-      where
-        oldInit = Data.List.NonEmpty.init literals
-
-        oldLast = Data.List.NonEmpty.last literals
-
-        newInit = filter (not . emptyLine) oldInit
-
-    longestSharedPrefix =
-        case filteredLines of
-            l : ls ->
-                Data.Foldable.foldl' sharedPrefix (leadingSpaces l) (fmap leadingSpaces ls)
-            [] ->
-                ""
-
-    indent = Data.Text.length longestSharedPrefix
