@@ -1,5 +1,6 @@
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE OverloadedStrings #-}
+
 -- | Parse Dhall tokens. Even though we don't have a tokenizer per-se this
 ---  module is useful for keeping some small parsing utilities.
 module Dhall.Parser.Token (
@@ -9,8 +10,9 @@ module Dhall.Parser.Token (
     bashEnvironmentVariable,
     posixEnvironmentVariable,
     ComponentType(..),
+    text,
+    char,
     file_,
-    labelOnly,
     label,
     anyLabel,
     labels,
@@ -26,7 +28,6 @@ module Dhall.Parser.Token (
     _if,
     _then,
     _else,
-    _letOnly,
     _let,
     _in,
     _as,
@@ -44,6 +45,8 @@ module Dhall.Parser.Token (
     _NaturalToInteger,
     _NaturalShow,
     _NaturalSubtract,
+    _IntegerClamp,
+    _IntegerNegate,
     _IntegerShow,
     _IntegerToDouble,
     _DoubleShow,
@@ -70,7 +73,6 @@ module Dhall.Parser.Token (
     _Kind,
     _Sort,
     _Location,
-    _equalOnly,
     _equal,
     _or,
     _plus,
@@ -91,7 +93,6 @@ module Dhall.Parser.Token (
     _comma,
     _openParens,
     _closeParens,
-    _colonOnly,
     _colon,
     _at,
     _equivalent,
@@ -103,21 +104,24 @@ module Dhall.Parser.Token (
     _lambda,
     _forall,
     _arrow,
+    _doubleColon,
     ) where
 
 import           Dhall.Parser.Combinators
 
 import Control.Applicative (Alternative(..), optional)
-import Data.Functor (void)
+import Data.Bits ((.&.))
+import Data.Functor (void, ($>))
 import Data.Semigroup (Semigroup(..))
 import Data.Text (Text)
-import Dhall.Core
+import Dhall.Syntax
 import Dhall.Set (Set)
 import Prelude hiding (const, pi)
 import Text.Parser.Combinators (choice, try, (<?>))
 
 import qualified Control.Monad
 import qualified Data.Char                  as Char
+import qualified Data.Foldable
 import qualified Data.HashSet
 import qualified Data.List.NonEmpty
 import qualified Data.Text
@@ -131,17 +135,26 @@ import qualified Text.Parser.Combinators
 import Numeric.Natural (Natural)
 import Prelude hiding (const, pi)
 
-import qualified Text.Parser.Token
-
-validCodepoint :: Char -> Bool
+-- | Returns `True` if the given `Int` is a valid Unicode codepoint
+validCodepoint :: Int -> Bool
 validCodepoint c =
-    not (category == Char.Surrogate || category == Char.NotAssigned)
+    not (category == Char.Surrogate 
+      || c .&. 0xFFFE == 0xFFFE 
+      || c .&. 0xFFFF == 0xFFFF)
   where
-    category = Char.generalCategory c
+    category = Char.generalCategory (Char.chr c)
 
+{-| Parse 0 or more whitespace characters (including comments)
+
+    This corresponds to the @whsp@ rule in the official grammar
+-}
 whitespace :: Parser ()
 whitespace = Text.Parser.Combinators.skipMany whitespaceChunk
 
+{-| Parse 1 or more whitespace characters (including comments)
+
+    This corresponds to the @whsp1@ rule in the official grammar
+-}
 nonemptyWhitespace :: Parser ()
 nonemptyWhitespace = Text.Parser.Combinators.skipSome whitespaceChunk
 
@@ -154,6 +167,10 @@ digit c = '\x30' <= c && c <= '\x39'
 alphaNum :: Char -> Bool
 alphaNum c = alpha c || digit c
 
+{-| Parse a hex digit (uppercase or lowercase)
+
+    This corresponds to the @HEXDIG@ rule in the official grammar
+-}
 hexdig :: Char -> Bool
 hexdig c =
         ('0' <= c && c <= '9')
@@ -162,45 +179,89 @@ hexdig c =
 
 signPrefix :: Num a => Parser (a -> a)
 signPrefix = (do
-    let positive = fmap (\_ -> id    ) (Text.Parser.Char.char '+')
-    let negative = fmap (\_ -> negate) (Text.Parser.Char.char '-')
+    let positive = fmap (\_ -> id    ) (char '+')
+    let negative = fmap (\_ -> negate) (char '-')
     positive <|> negative ) <?> "sign"
 
+{-| Parse a `Double` literal
+
+    This corresponds to the @double-literal@ rule from the official grammar
+-}
 doubleLiteral :: Parser Double
 doubleLiteral = (do
     sign <- signPrefix <|> pure id
-    a <- Text.Parser.Token.double
-    return (sign a) ) <?> "double literal"
+    a <- Text.Megaparsec.Char.Lexer.float
+    return (sign a) ) <?> "literal"
 
+{-| Parse a signed @Infinity@
+
+    This corresponds to the @minus-infinity-literal@ and @plus-infinity-literal@
+    rules from the official grammar
+-}
 doubleInfinity :: Parser Double
 doubleInfinity = (do
-    let negative = fmap (\_ -> negate) (Text.Parser.Char.char '-')
+    let negative = fmap (\_ -> negate) (char '-')
     sign <- negative <|> pure id
-    a <- Text.Parser.Char.text "Infinity" >> whitespace >> return (1.0/0.0)
-    return (sign a) ) <?> "double infinity"
+    a <- text "Infinity" >> return (1.0/0.0)
+    return (sign a) ) <?> "literal"
 
+{-| Parse an `Integer` literal
+
+    This corresponds to the @integer-literal@ rule from the official grammar
+-}
 integerLiteral :: Parser Integer
 integerLiteral = (do
     sign <- signPrefix
-    a <- Text.Megaparsec.Char.Lexer.decimal
-    whitespace
-    return (sign a) ) <?> "integer literal"
+    a    <- naturalLiteral
+    return (sign (fromIntegral a)) ) <?> "literal"
 
+{-| Parse a `Natural` literal 
+
+    This corresponds to the @natural-literal@ rule from the official grammar
+-}
 naturalLiteral :: Parser Natural
 naturalLiteral = (do
-    a <- Text.Megaparsec.Char.Lexer.decimal
-    whitespace
-    return a ) <?> "natural literal"
+    a <-    try (char '0' >> char 'x' >> Text.Megaparsec.Char.Lexer.hexadecimal)
+        <|> decimal
+        <|> (char '0' $> 0)
+    return a ) <?> "literal"
+  where
+    decimal = do
+        n <- headDigit
+        ns <- many tailDigit
+        return (mkNum (n:ns))
+      where
+        headDigit = decimalDigit nonZeroDigit <?> "non-zero digit"
+          where
+            nonZeroDigit c = '1' <= c && c <= '9'
 
+        tailDigit = decimalDigit digit <?> "digit"
+
+        decimalDigit predicate = do
+            c <- Text.Parser.Char.satisfy predicate
+            return (fromIntegral (Char.ord c - Char.ord '0'))
+
+        mkNum = Data.Foldable.foldl' step 0
+          where
+            step acc x = acc * 10 + x
+
+
+{-| Parse an identifier (i.e. a variable or built-in)
+
+    Variables can have an optional index to disambiguate shadowed variables
+
+    This corresponds to the @identifier@ rule from the official grammar
+-}
 identifier :: Parser Var
 identifier = do
     x <- label
 
-    let indexed = do
-            _ <- Text.Parser.Char.char '@'
-            n <- Text.Megaparsec.Char.Lexer.decimal
+    let indexed = try $ do
             whitespace
-            return n
+            _at
+            whitespace
+            n <- naturalLiteral
+            return (fromIntegral n)
 
     n <- indexed <|> pure 0
     return (V x n)
@@ -209,13 +270,14 @@ whitespaceChunk :: Parser ()
 whitespaceChunk =
     choice
         [ void (Dhall.Parser.Combinators.takeWhile1 predicate)
-        , void (Text.Parser.Char.text "\r\n")
+        , void (Text.Parser.Char.text "\r\n" <?> "newline")
         , lineComment
         , blockComment
         ] <?> "whitespace"
   where
     predicate c = c == ' ' || c == '\t' || c == '\n'
 
+-- | Parse a hexademical number and convert to the corresponding `Int`
 hexNumber :: Parser Int
 hexNumber = choice [ hexDigit, hexUpper, hexLower ]
   where
@@ -239,7 +301,7 @@ hexNumber = choice [ hexDigit, hexUpper, hexLower ]
 
 lineComment :: Parser ()
 lineComment = do
-    _ <- Text.Parser.Char.text "--"
+    _ <- text "--"
 
     let predicate c = ('\x20' <= c && c <= '\x10FFFF') || c == '\t'
 
@@ -250,12 +312,13 @@ lineComment = do
     return ()
   where
     endOfLine =
-            void (Text.Parser.Char.char '\n'  )
+        (   void (Text.Parser.Char.char '\n'  )
         <|> void (Text.Parser.Char.text "\r\n")
+        ) <?> "newline"
 
 blockComment :: Parser ()
 blockComment = do
-    _ <- Text.Parser.Char.text "{-"
+    _ <- text "{-"
     blockCommentContinue
 
 blockCommentChunk :: Parser ()
@@ -278,12 +341,12 @@ blockCommentChunk =
       where
         predicate c = '\x20' <= c && c <= '\x10FFFF' || c == '\n' || c == '\t'
 
-    endOfLine = void (Text.Parser.Char.text "\r\n")
+    endOfLine = void (Text.Parser.Char.text "\r\n" <?> "newline")
 
 blockCommentContinue :: Parser ()
 blockCommentContinue = endOfComment <|> continue
   where
-    endOfComment = void (Text.Parser.Char.text "-}")
+    endOfComment = void (text "-}")
 
     continue = do
         blockCommentChunk
@@ -293,28 +356,38 @@ simpleLabel :: Bool -> Parser Text
 simpleLabel allowReserved = try (do
     c    <- Text.Parser.Char.satisfy headCharacter
     rest <- Dhall.Parser.Combinators.takeWhile tailCharacter
-    let text = Data.Text.cons c rest
-    Control.Monad.guard (allowReserved || not (Data.HashSet.member text reservedIdentifiers))
-    return text )
+    let t = Data.Text.cons c rest
+    Control.Monad.guard (allowReserved || not (Data.HashSet.member t reservedIdentifiers))
+    return t )
   where
-    headCharacter c = alpha c || c == '_'
 
-    tailCharacter c = alphaNum c || c == '_' || c == '-' || c == '/'
+headCharacter :: Char -> Bool
+headCharacter c = alpha c || c == '_'
+
+tailCharacter :: Char -> Bool
+tailCharacter c = alphaNum c || c == '_' || c == '-' || c == '/'
 
 backtickLabel :: Parser Text
 backtickLabel = do
-    _ <- Text.Parser.Char.char '`'
+    _ <- char '`'
     t <- takeWhile1 predicate
-    _ <- Text.Parser.Char.char '`'
+    _ <- char '`'
     return t
   where
     predicate c =
             '\x20' <= c && c <= '\x5F'
         ||  '\x61' <= c && c <= '\x7E'
 
+{-| Parse a braced sequence of comma-separated labels
+
+    For example, this is used to parse the record projection syntax
+
+    This corresponds to the @labels@ rule in the official grammar
+-}
 labels :: Parser (Set Text)
 labels = do
     _openBrace
+    whitespace
     xs <- nonEmptyLabels <|> emptyLabels
     _closeBrace
     return xs
@@ -323,25 +396,33 @@ labels = do
 
     nonEmptyLabels = do
         x  <- anyLabel
-        xs <- many (do _ <- _comma; anyLabel)
+        whitespace
+        xs <- many (do _comma; whitespace; l <- anyLabel; whitespace; return l)
         noDuplicates (x : xs)
 
+{-| Parse a label (e.g. a variable\/field\/alternative name)
 
-labelOnly :: Parser Text
-labelOnly = backtickLabel <|> simpleLabel False <?> "label"
+    Rejects labels that match built-in names (e.g. @Natural/even@)
 
+    This corresponds to the @nonreserved-label@ rule in the official grammar
+-}
 label :: Parser Text
-label = (do
-    t <- backtickLabel <|> simpleLabel False
-    whitespace
-    return t ) <?> "label"
+label = backtickLabel <|> simpleLabel False <?> "label"
 
+{-| Same as `label` except that built-in names are allowed
+
+    This corresponds to the @any-label@ rule in the official grammar
+-}
 anyLabel :: Parser Text
 anyLabel = (do
     t <- backtickLabel <|> simpleLabel True
-    whitespace
     return t ) <?> "any label"
 
+{-| Parse a valid Bash environment variable name
+
+    This corresponds to the @bash-environment-variable@ rule in the official
+    grammar
+-}
 bashEnvironmentVariable :: Parser Text
 bashEnvironmentVariable = satisfy predicate0 <> star (satisfy predicate1)
   where
@@ -349,6 +430,12 @@ bashEnvironmentVariable = satisfy predicate0 <> star (satisfy predicate1)
 
     predicate1 c = alphaNum c || c == '_'
 
+{-| Parse a valid POSIX environment variable name, which permits a wider range
+    of characters than a Bash environment variable name
+
+    This corresponds to the @posix-environment-variable@ rule in the official
+    grammar
+-}
 posixEnvironmentVariable :: Parser Text
 posixEnvironmentVariable = plus posixEnvironmentVariableCharacter
 
@@ -357,7 +444,7 @@ posixEnvironmentVariableCharacter =
     escapeCharacter <|> satisfy predicate1
   where
     escapeCharacter = do
-        _ <- Text.Parser.Char.char '\\'
+        _ <- char '\\'
 
         c <- Text.Parser.Char.satisfy (`elem` ("\"\\abfnrtv" :: String))
 
@@ -385,8 +472,12 @@ quotedPathCharacter c =
     ||  ('\x23' <= c && c <= '\x2E')
     ||  ('\x30' <= c && c <= '\x10FFFF')
 
+{-| The `pathComponent` function uses this type to distinguish whether to parse
+    a URL path component or a file path component
+-}
 data ComponentType = URLComponent | FileComponent
 
+-- | Parse a path component
 pathComponent :: ComponentType -> Parser Text
 pathComponent componentType = do
     _ <- "/" :: Parser Text
@@ -394,23 +485,24 @@ pathComponent componentType = do
     let pathData =
             case componentType of
                 FileComponent -> do
-                    Text.Megaparsec.takeWhile1P Nothing Dhall.Core.pathCharacter
+                    Text.Megaparsec.takeWhile1P Nothing Dhall.Syntax.pathCharacter
                 URLComponent -> do
                     star pchar
 
     let quotedPathData = do
-            _    <- Text.Parser.Char.char '"'
-            text <- Text.Megaparsec.takeWhile1P Nothing quotedPathCharacter
-            _    <- Text.Parser.Char.char '"'
+            _ <- char '"'
+            t <- Text.Megaparsec.takeWhile1P Nothing quotedPathCharacter
+            _ <- char '"'
 
             case componentType of
               FileComponent -> do
-                return text
+                return t
               URLComponent -> do
-                return (URI.Encode.encodeText text)
+                return (URI.Encode.encodeText t)
 
     quotedPathData <|> pathData
 
+-- | Parse a `File`
 file_ :: ComponentType -> Parser File
 file_ componentType = do
     let emptyPath =
@@ -431,6 +523,10 @@ scheme_ =
     *>  ((("s" :: Parser Text) *> pure HTTPS) <|> pure HTTP)
     <*  ("://" :: Parser Text)
 
+{-| Parse an HTTP(S) URL without trailing whitespace
+
+    This corresponds to the @http-raw@ rule in the official grammar
+-}
 httpRaw :: Parser URL
 httpRaw = do
     scheme    <- scheme_
@@ -575,278 +671,514 @@ unreserved :: Char -> Bool
 unreserved c =
     alphaNum c || c == '-' || c == '.' || c == '_' || c == '~'
 
-reserved :: Data.Text.Text -> Parser ()
-reserved x = do _ <- Text.Parser.Char.text x; whitespace
+{-| A variation on `Text.Parser.Char.text` that doesn't quote the expected
+    in error messages
+-}
+text :: Data.Text.Text -> Parser Text
+text t = Text.Parser.Char.text t <?> Data.Text.unpack t
+{-# INLINE text #-}
 
-reservedCharOnly :: Char -> Parser ()
-reservedCharOnly c = do _ <- Text.Parser.Char.char c; return ()
+{-| A variation on `Text.Parser.Char.char` that doesn't quote the expected
+    token in error messages
+-}
+char :: Char -> Parser Char
+char c = Text.Parser.Char.char c <?> [ c ]
+{-# INLINE char #-}
+
+reserved :: Data.Text.Text -> Parser ()
+reserved x = void (text x)
 
 reservedChar :: Char -> Parser ()
-reservedChar c = do _ <- Text.Parser.Char.char c; whitespace
+reservedChar c = void (char c)
 
-keywordOnly :: Data.Text.Text -> Parser ()
-keywordOnly x = try (do _ <- Text.Parser.Char.text x; return ())
+builtin :: Data.Text.Text -> Parser ()
+builtin x = reserved x <?> "built-in"
+{-# INLINE builtin #-}
+
+operator :: Data.Text.Text -> Parser ()
+operator x = reserved x <?> "operator"
+{-# INLINE operator #-}
+
+operatorChar :: Char -> Parser ()
+operatorChar x = reservedChar x <?> "operator"
+{-# INLINE operatorChar #-}
 
 keyword :: Data.Text.Text -> Parser ()
-keyword x = try (do _ <- Text.Parser.Char.text x; nonemptyWhitespace)
+keyword x = try (void (text x)) <?> "keyword"
 
+{-| Parse the @if@ keyword
+
+    This corresponds to the @if@ rule from the official grammar
+-}
 _if :: Parser ()
 _if = keyword "if"
 
+{-| Parse the @then@ keyword
+
+    This corresponds to the @then@ rule from the official grammar
+-}
 _then :: Parser ()
 _then = keyword "then"
 
+{-| Parse the @else@ keyword
+
+    This corresponds to the @else@ rule from the official grammar
+-}
 _else :: Parser ()
 _else = keyword "else"
 
-_letOnly :: Parser ()
-_letOnly = keywordOnly "let"
+{-| Parse the @let@ keyword
 
+    This corresponds to the @let@ rule from the official grammar
+-}
 _let :: Parser ()
 _let = keyword "let"
 
+{-| Parse the @in@ keyword
+
+    This corresponds to the @in@ rule from the official grammar
+-}
 _in :: Parser ()
 _in = keyword "in"
 
+{-| Parse the @as@ keyword
+
+    This corresponds to the @as@ rule from the official grammar
+-}
 _as :: Parser ()
 _as = keyword "as"
 
+{-| Parse the @using@ keyword
+
+    This corresponds to the @using@ rule from the official grammar
+-}
 _using :: Parser ()
 _using = keyword "using"
 
+{-| Parse the @merge@ keyword
+
+    This corresponds to the @merge@ rule from the official grammar
+-}
 _merge :: Parser ()
 _merge = keyword "merge"
 
+{-| Parse the @toMap@ keyword
+
+    This corresponds to the @toMap@ rule from the official grammar
+-}
 _toMap :: Parser ()
 _toMap = keyword "toMap"
 
+{-| Parse the @assert@ keyword
+
+    This corresponds to the @assert@ rule from the official grammar
+-}
 _assert :: Parser ()
 _assert = keyword "assert"
 
+{-| Parse the @Some@ built-in
+
+    This corresponds to the @Some@ rule from the official grammar
+-}
 _Some :: Parser ()
 _Some = keyword "Some"
 
+{-| Parse the @None@ built-in
+
+    This corresponds to the @None@ rule from the official grammar
+-}
 _None :: Parser ()
-_None = reserved "None"
+_None = builtin "None"
 
+{-| Parse the @Natural/fold@ built-in
+
+    This corresponds to the @Natural-fold@ rule from the official grammar
+-}
 _NaturalFold :: Parser ()
-_NaturalFold = reserved "Natural/fold"
+_NaturalFold = builtin "Natural/fold"
 
+{-| Parse the @Natural/build@ built-in
+
+    This corresponds to the @Natural-build@ rule from the official grammar
+-}
 _NaturalBuild :: Parser ()
-_NaturalBuild = reserved "Natural/build"
+_NaturalBuild = builtin "Natural/build"
 
+{-| Parse the @Natural/isZero@ built-in
+
+    This corresponds to the @Natural-isZero@ rule from the official grammar
+-}
 _NaturalIsZero :: Parser ()
-_NaturalIsZero = reserved "Natural/isZero"
+_NaturalIsZero = builtin "Natural/isZero"
 
+{-| Parse the @Natural/even@ built-in
+
+    This corresponds to the @Natural-even@ rule from the official grammar
+-}
 _NaturalEven :: Parser ()
-_NaturalEven = reserved "Natural/even"
+_NaturalEven = builtin "Natural/even"
 
+{-| Parse the @Natural/odd@ built-in
+
+    This corresponds to the @Natural-odd@ rule from the official grammar
+-}
 _NaturalOdd :: Parser ()
-_NaturalOdd = reserved "Natural/odd"
+_NaturalOdd = builtin "Natural/odd"
 
+{-| Parse the @Natural/toInteger@ built-in
+
+    This corresponds to the @Natural-toInteger@ rule from the official grammar
+-}
 _NaturalToInteger :: Parser ()
-_NaturalToInteger = reserved "Natural/toInteger"
+_NaturalToInteger = builtin "Natural/toInteger"
 
+{-| Parse the @Natural/show@ built-in
+
+    This corresponds to the @Natural-show@ rule from the official grammar
+-}
 _NaturalShow :: Parser ()
-_NaturalShow = reserved "Natural/show"
+_NaturalShow = builtin "Natural/show"
 
+{-| Parse the @Natural/subtract@ built-in
+
+    This corresponds to the @Natural-subtract@ rule from the official grammar
+-}
 _NaturalSubtract :: Parser ()
-_NaturalSubtract = reserved "Natural/subtract"
+_NaturalSubtract = builtin "Natural/subtract"
 
+{-| Parse the @Integer/clamp@ built-in
+
+    This corresponds to the @Integer-clamp@ rule from the official grammar
+-}
+_IntegerClamp :: Parser ()
+_IntegerClamp = builtin "Integer/clamp"
+
+{-| Parse the @Integer/negate@ built-in
+
+    This corresponds to the @Integer-negate@ rule from the official grammar
+-}
+_IntegerNegate :: Parser ()
+_IntegerNegate = builtin "Integer/negate"
+
+{-| Parse the @Integer/show@ built-in
+
+    This corresponds to the @Integer-show@ rule from the official grammar
+-}
 _IntegerShow :: Parser ()
-_IntegerShow = reserved "Integer/show"
+_IntegerShow = builtin "Integer/show"
 
+{-| Parse the @Integer/toDouble@ built-in
+
+    This corresponds to the @Integer-toDouble@ rule from the official grammar
+-}
 _IntegerToDouble :: Parser ()
-_IntegerToDouble = reserved "Integer/toDouble"
+_IntegerToDouble = builtin "Integer/toDouble"
 
+{-| Parse the @Double/show@ built-in
+
+    This corresponds to the @Double-show@ rule from the official grammar
+-}
 _DoubleShow :: Parser ()
-_DoubleShow = reserved "Double/show"
+_DoubleShow = builtin "Double/show"
 
+{-| Parse the @List/build@ built-in
+
+    This corresponds to the @List-build@ rule from the official grammar
+-}
 _ListBuild :: Parser ()
-_ListBuild = reserved "List/build"
+_ListBuild = builtin "List/build"
 
+{-| Parse the @List/fold@ built-in
+
+    This corresponds to the @List-fold@ rule from the official grammar
+-}
 _ListFold :: Parser ()
-_ListFold = reserved "List/fold"
+_ListFold = builtin "List/fold"
 
+{-| Parse the @List/length@ built-in
+
+    This corresponds to the @List-length@ rule from the official grammar
+-}
 _ListLength :: Parser ()
-_ListLength = reserved "List/length"
+_ListLength = builtin "List/length"
 
+{-| Parse the @List/head@ built-in
+
+    This corresponds to the @List-head@ rule from the official grammar
+-}
 _ListHead :: Parser ()
-_ListHead = reserved "List/head"
+_ListHead = builtin "List/head"
 
+{-| Parse the @List/last@ built-in
+
+    This corresponds to the @List-last@ rule from the official grammar
+-}
 _ListLast :: Parser ()
-_ListLast = reserved "List/last"
+_ListLast = builtin "List/last"
 
+{-| Parse the @List/indexed@ built-in
+
+    This corresponds to the @List-indexed@ rule from the official grammar
+-}
 _ListIndexed :: Parser ()
-_ListIndexed = reserved "List/indexed"
+_ListIndexed = builtin "List/indexed"
 
+{-| Parse the @List/reverse@ built-in
+
+    This corresponds to the @List-reverse@ rule from the official grammar
+-}
 _ListReverse :: Parser ()
-_ListReverse = reserved "List/reverse"
+_ListReverse = builtin "List/reverse"
 
+{-| Parse the @Optional/fold@ built-in
+
+    This corresponds to the @Optional-fold@ rule from the official grammar
+-}
 _OptionalFold :: Parser ()
-_OptionalFold = reserved "Optional/fold"
+_OptionalFold = builtin "Optional/fold"
 
+{-| Parse the @Optional/build@ built-in
+
+    This corresponds to the @Optional-build@ rule from the official grammar
+-}
 _OptionalBuild :: Parser ()
-_OptionalBuild = reserved "Optional/build"
+_OptionalBuild = builtin "Optional/build"
 
+{-| Parse the @Bool@ built-in
+
+    This corresponds to the @Bool@ rule from the official grammar
+-}
 _Bool :: Parser ()
-_Bool = reserved "Bool"
+_Bool = builtin "Bool"
 
+{-| Parse the @Optional@ built-in
+
+    This corresponds to the @Optional@ rule from the official grammar
+-}
 _Optional :: Parser ()
-_Optional = reserved "Optional"
+_Optional = builtin "Optional"
 
+{-| Parse the @Natural@ built-in
+
+    This corresponds to the @Natural@ rule from the official grammar
+-}
 _Natural :: Parser ()
-_Natural = reserved "Natural"
+_Natural = builtin "Natural"
 
+{-| Parse the @Integer@ built-in
+
+    This corresponds to the @Integer@ rule from the official grammar
+-}
 _Integer :: Parser ()
-_Integer = reserved "Integer"
+_Integer = builtin "Integer"
 
+{-| Parse the @Double@ built-in
+
+    This corresponds to the @Double@ rule from the official grammar
+-}
 _Double :: Parser ()
-_Double = reserved "Double"
+_Double = builtin "Double"
 
+{-| Parse the @Text@ built-in
+
+    This corresponds to the @Text@ rule from the official grammar
+-}
 _Text :: Parser ()
-_Text = reserved "Text"
+_Text = builtin "Text"
 
+{-| Parse the @Text/show@ built-in
+
+    This corresponds to the @Text-show@ rule from the official grammar
+-}
 _TextShow :: Parser ()
-_TextShow = reserved "Text/show"
+_TextShow = builtin "Text/show"
 
+{-| Parse the @List@ built-in
+
+    This corresponds to the @List@ rule from the official grammar
+-}
 _List :: Parser ()
-_List = reserved "List"
+_List = builtin "List"
 
+{-| Parse the @True@ built-in
+
+    This corresponds to the @True@ rule from the official grammar
+-}
 _True :: Parser ()
-_True = reserved "True"
+_True = builtin "True"
 
+{-| Parse the @False@ built-in
+
+    This corresponds to the @False@ rule from the official grammar
+-}
 _False :: Parser ()
-_False = reserved "False"
+_False = builtin "False"
 
+{-| Parse a @NaN@ literal
+
+    This corresponds to the @NaN@ rule from the official grammar
+-}
 _NaN :: Parser ()
-_NaN = reserved "NaN"
+_NaN = builtin "NaN"
 
+{-| Parse the @Type@ built-in
+
+    This corresponds to the @Type@ rule from the official grammar
+-}
 _Type :: Parser ()
-_Type = reserved "Type"
+_Type = builtin "Type"
 
+{-| Parse the @Kind@ built-in
+
+    This corresponds to the @Kind@ rule from the official grammar
+-}
 _Kind :: Parser ()
-_Kind = reserved "Kind"
+_Kind = builtin "Kind"
 
+{-| Parse the @Sort@ built-in
+
+    This corresponds to the @Sort@ rule from the official grammar
+-}
 _Sort :: Parser ()
-_Sort = reserved "Sort"
+_Sort = builtin "Sort"
 
+{-| Parse the @Location@ keyword
+
+    This corresponds to the @Location@ rule from the official grammar
+-}
 _Location :: Parser ()
-_Location = reserved "Location"
-
-_equalOnly :: Parser ()
-_equalOnly = reservedCharOnly '='
+_Location = builtin "Location"
 
+-- | Parse the @=@ symbol
 _equal :: Parser ()
 _equal = reservedChar '='
 
+-- | Parse the @||@ symbol
 _or :: Parser ()
-_or = reserved "||"
+_or = operator "||"
 
+-- | Parse the @+@ symbol
 _plus :: Parser ()
-_plus = reservedChar '+'
+_plus = operatorChar '+'
 
+-- | Parse the @++@ symbol
 _textAppend :: Parser ()
-_textAppend = reserved "++"
+_textAppend = operator "++"
 
+-- | Parse the @#@ symbol
 _listAppend :: Parser ()
-_listAppend = reservedChar '#'
+_listAppend = operatorChar '#'
 
+-- | Parse the @&&@ symbol
 _and :: Parser ()
-_and = reserved "&&"
+_and = operator "&&"
 
+-- | Parse the @*@ symbol
 _times :: Parser ()
-_times = reservedChar '*'
+_times = operatorChar '*'
 
+-- | Parse the @==@ symbol
 _doubleEqual :: Parser ()
-_doubleEqual = reserved "=="
+_doubleEqual = operator "=="
 
+-- | Parse the @!=@ symbol
 _notEqual :: Parser ()
-_notEqual = reserved "!="
+_notEqual = operator "!="
 
+-- | Parse the @.@ symbol
 _dot :: Parser ()
-_dot = reservedChar '.'
+_dot = operatorChar '.'
 
+-- | Parse the @{@ symbol
 _openBrace :: Parser ()
 _openBrace = reservedChar '{'
 
+-- | Parse the @}@ symbol
 _closeBrace :: Parser ()
 _closeBrace = reservedChar '}'
 
+-- | Parse the @[@] symbol
 _openBracket :: Parser ()
 _openBracket = reservedChar '['
 
+-- | Parse the @]@ symbol
 _closeBracket :: Parser ()
 _closeBracket = reservedChar ']'
 
+-- | Parse the @<@ symbol
 _openAngle :: Parser ()
 _openAngle = reservedChar '<'
 
+-- | Parse the @>@ symbol
 _closeAngle :: Parser ()
 _closeAngle = reservedChar '>'
 
+-- | Parse the @|@ symbol
 _bar :: Parser ()
 _bar = reservedChar '|'
 
+-- | Parse the @,@ symbol
 _comma :: Parser ()
-_comma = reservedChar ','
+_comma = reservedChar ',' <?> "\',\'"
 
+-- | Parse the @(@ symbol
 _openParens :: Parser ()
 _openParens = reservedChar '('
 
+-- | Parse the @)@ symbol
 _closeParens :: Parser ()
 _closeParens = reservedChar ')'
 
-_colonOnly :: Parser ()
-_colonOnly = reservedCharOnly ':'
-
+-- | Parse the @:@ symbol
 _colon :: Parser ()
 _colon = reservedChar ':'
 
+-- | Parse the @\@@ symbol
 _at :: Parser ()
-_at = reservedChar '@'
+_at = reservedChar '@' <?> "\"@\""
 
+-- | Parse the equivalence symbol (@===@ or @≡@)
 _equivalent :: Parser ()
-_equivalent = do
-    void (Text.Parser.Char.char '≡' <?> "\"≡\"") <|> void (Text.Parser.Char.text "===")
-    whitespace
+_equivalent = (void (char '≡' <?> "\"≡\"") <|> void (text "===")) <?> "operator"
 
+-- | Parse the @missing@ keyword
 _missing :: Parser ()
-_missing = reserved "missing"
+_missing =
+        keyword "missing"
+    *>  Text.Megaparsec.notFollowedBy (Text.Parser.Char.satisfy tailCharacter)
 
+-- | Parse the @?@ symbol
 _importAlt :: Parser ()
-_importAlt = reservedChar '?'
+_importAlt = operatorChar '?'
 
+-- | Parse the record combine operator (@/\\@ or @∧@)
 _combine :: Parser ()
-_combine = do
-    void (Text.Parser.Char.char '∧' <?> "\"∧\"") <|> void (Text.Parser.Char.text "/\\")
-    whitespace
+_combine = (void (char '∧' <?> "\"∧\"") <|> void (text "/\\")) <?> "operator"
 
+-- | Parse the record type combine operator (@//\\\\@ or @⩓@)
 _combineTypes :: Parser ()
-_combineTypes = do
-    void (Text.Parser.Char.char '⩓' <?> "\"⩓\"") <|> void (Text.Parser.Char.text "//\\\\")
-    whitespace
+_combineTypes = (void (char '⩓' <?> "\"⩓\"") <|> void (text "//\\\\")) <?> "operator"
 
+-- | Parse the record \"prefer\" operator (@//@ or @⫽@)
 _prefer :: Parser ()
-_prefer = do
-    void (Text.Parser.Char.char '⫽' <?> "\"⫽\"") <|> void (Text.Parser.Char.text "//")
-    whitespace
+_prefer = (void (char '⫽' <?> "\"⫽\"") <|> void (text "//")) <?> "operator"
 
+-- | Parse a lambda (@\\@ or @λ@)
 _lambda :: Parser ()
-_lambda = do
-    _ <- Text.Parser.Char.satisfy predicate
-    whitespace
+_lambda = void (Text.Parser.Char.satisfy predicate) <?> "\\"
   where
     predicate 'λ'  = True
     predicate '\\' = True
     predicate _    = False
 
+-- | Parse a forall (@forall@ or @∀@)
 _forall :: Parser ()
-_forall = do
-    void (Text.Parser.Char.char '∀' <?> "\"∀\"") <|> void (Text.Parser.Char.text "forall")
-    whitespace
+_forall = (void (char '∀' <?> "\"∀\"") <|> void (text "forall")) <?> "forall"
 
+-- | Parse a right arrow (@->@ or @→@)
 _arrow :: Parser ()
-_arrow = do
-    void (Text.Parser.Char.char '→' <?> "\"→\"") <|> void (Text.Parser.Char.text "->")
-    whitespace
+_arrow = (void (char '→' <?> "\"→\"") <|> void (text "->")) <?> "->"
+
+-- | Parse a double colon (@::@)
+_doubleColon :: Parser ()
+_doubleColon = operator "::"
