@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP               #-}
+{-# LANGUAGE LambdaCase        #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards   #-}
 
@@ -18,10 +18,13 @@ module Dhall.Pretty.Internal (
     , prettyCharacterSet
 
     , prettyVar
-    , pretty
-    , escapeText
+    , pretty_
+    , escapeText_
+    , escapeEnvironmentVariable
+    , prettyEnvironmentVariable
 
     , prettyConst
+    , escapeLabel
     , prettyLabel
     , prettyAnyLabel
     , prettyLabels
@@ -31,6 +34,8 @@ module Dhall.Pretty.Internal (
     , prettyDouble
     , prettyToStrictText
     , prettyToString
+    , layout
+    , layoutOpts
 
     , docToStrictText
 
@@ -58,19 +63,15 @@ module Dhall.Pretty.Internal (
     , rparen
     ) where
 
-import Dhall.Core
-
-#if MIN_VERSION_base(4,8,0)
-#else
-import Control.Applicative (Applicative(..), (<$>))
-#endif
 import Data.Foldable
+import Data.List.NonEmpty (NonEmpty)
 import Data.Monoid ((<>))
 import Data.Text (Text)
 import Data.Text.Prettyprint.Doc (Doc, Pretty, space)
 import Dhall.Map (Map)
 import Dhall.Set (Set)
 import Dhall.Src (Src(..))
+import Dhall.Syntax
 import Numeric.Natural (Natural)
 import Prelude hiding (succ)
 import qualified Data.Text.Prettyprint.Doc.Render.Terminal as Terminal
@@ -78,12 +79,13 @@ import qualified Data.Text.Prettyprint.Doc.Render.Terminal as Terminal
 import qualified Data.Char
 import qualified Data.HashSet
 import qualified Data.List
+import qualified Data.List.NonEmpty                      as NonEmpty
 import qualified Data.Set
 import qualified Data.Text                               as Text
 import qualified Data.Text.Prettyprint.Doc               as Pretty
 import qualified Data.Text.Prettyprint.Doc.Render.Text   as Pretty
 import qualified Data.Text.Prettyprint.Doc.Render.String as Pretty
-import qualified Dhall.Map
+import qualified Dhall.Map                               as Map
 import qualified Dhall.Set
 
 {-| Annotation type used to tag elements in a pretty-printed document for
@@ -96,6 +98,7 @@ data Ann
   | Literal     -- ^ Literals such as integers and strings
   | Builtin     -- ^ Builtin types and values
   | Operator    -- ^ Operators
+  deriving Show
 
 {-| Convert annotations to their corresponding color for syntax highlighting
     purposes
@@ -108,7 +111,8 @@ annToAnsiStyle Literal  = Terminal.colorDull Terminal.Magenta
 annToAnsiStyle Builtin  = Terminal.underlined
 annToAnsiStyle Operator = Terminal.bold <> Terminal.colorDull Terminal.Green
 
-data CharacterSet = ASCII | Unicode
+-- | This type determines whether to render code as `ASCII` or `Unicode`
+data CharacterSet = ASCII | Unicode deriving Show
 
 -- | Pretty print an expression
 prettyExpr :: Pretty a => Expr s a -> Doc Ann
@@ -138,29 +142,28 @@ isWhitespace c =
 
     >>> let unusedSourcePos = Text.Megaparsec.SourcePos "" (Text.Megaparsec.mkPos 1) (Text.Megaparsec.mkPos 1)
     >>> let nonEmptySrc = Src unusedSourcePos unusedSourcePos "-- Documentation for x\n"
-    >>> "let" <> renderSrc (Just nonEmptySrc) " " <> "x = 1 in x"
+    >>> "let" <> " " <> renderSrc id (Just nonEmptySrc) <> "x = 1 in x"
     let -- Documentation for x
         x = 1 in x
     >>> let emptySrc = Src unusedSourcePos unusedSourcePos "      "
-    >>> "let" <> renderSrc (Just emptySrc) " " <> "x = 1 in x"
+    >>> "let" <> " " <> renderSrc id (Just emptySrc) <> "x = 1 in x"
     let x = 1 in x
-    >>> "let" <> renderSrc Nothing " " <> "x = 1 in x"
+    >>> "let" <> " " <> renderSrc id Nothing <> "x = 1 in x"
     let x = 1 in x
 -}
 renderSrc
-    :: Maybe Src
+    :: (Text -> Text)
+    -- ^ Used to preprocess the comment string (e.g. to strip whitespace)
+    -> Maybe Src
     -- ^ Source span to render (if present)
     -> Doc Ann
-    -- ^ Used as the prefix (when the source span contains a comment) and as a
-    --   fallback (when the source span is absent or comment-free)
-    -> Doc Ann
-renderSrc (Just (Src {..})) prefix
+renderSrc strip (Just (Src {..}))
     | not (Text.all isWhitespace srcText) =
-        prefix <> Pretty.align (Pretty.concatWith f newLines <> suffix)
+        Pretty.align (Pretty.concatWith f newLines <> suffix)
   where
     horizontalSpace c = c == ' ' || c == '\t'
 
-    strippedText = Text.dropAround horizontalSpace srcText
+    strippedText = strip srcText
 
     suffix =
         if Text.null strippedText
@@ -195,8 +198,8 @@ renderSrc (Just (Src {..})) prefix
                 in  Pretty.pretty l0 : map perLine (l1 : ls)
 
     f x y = x <> Pretty.hardline <> y
-renderSrc _ prefix =
-    prefix
+renderSrc _ _ =
+    mempty
 
 -- Annotation helpers
 keyword, syntax, label, literal, builtin, operator :: Doc Ann -> Doc Ann
@@ -207,7 +210,7 @@ literal  = Pretty.annotate Literal
 builtin  = Pretty.annotate Builtin
 operator = Pretty.annotate Operator
 
-comma, lbracket, rbracket, langle, rangle, lbrace, rbrace, lparen, rparen, pipe, backtick, dollar, colon, equals, dot :: Doc Ann
+comma, lbracket, rbracket, langle, rangle, lbrace, rbrace, lparen, rparen, pipe, dollar, colon, equals, dot :: Doc Ann
 comma    = syntax Pretty.comma
 lbracket = syntax Pretty.lbracket
 rbracket = syntax Pretty.rbracket
@@ -218,7 +221,6 @@ rbrace   = syntax Pretty.rbrace
 lparen   = syntax Pretty.lparen
 rparen   = syntax Pretty.rparen
 pipe     = syntax Pretty.pipe
-backtick = syntax "`"
 dollar   = syntax "$"
 colon    = syntax ":"
 equals   = syntax "="
@@ -236,6 +238,9 @@ rarrow :: CharacterSet -> Doc Ann
 rarrow Unicode = syntax "→"
 rarrow ASCII   = syntax "->"
 
+doubleColon :: Doc Ann
+doubleColon = syntax "::"
+
 -- | Pretty-print a list
 list :: [Doc Ann] -> Doc Ann
 list   [] = lbracket <> rbracket
@@ -275,6 +280,32 @@ braces docs =
         rbrace
         docs
 
+hangingBraces :: Int -> [(Doc Ann, Doc Ann)] -> Doc Ann
+hangingBraces _ [] =
+    lbrace <> rbrace
+hangingBraces n docs =
+    Pretty.group
+        (Pretty.flatAlt
+            (  lbrace
+            <> Pretty.hardline
+            <> Pretty.indent n
+               ( mconcat (zipWith combineLong (repeat separator) docsLong)
+               <> rbrace
+               )
+            )
+            (mconcat (zipWith (<>) (beginShort : repeat separator) docsShort) <> space <> rbrace)
+        )
+  where
+    separator = comma <> space
+
+    docsShort = fmap fst docs
+
+    docsLong = fmap snd docs
+
+    beginShort = lbrace <> space
+
+    combineLong x y = x <> y <> Pretty.hardline
+
 -- | Pretty-print anonymous functions and function types
 arrows :: CharacterSet -> [(Doc Ann, Doc Ann)] -> Doc Ann
 arrows ASCII =
@@ -380,21 +411,26 @@ alpha c = ('\x41' <= c && c <= '\x5A') || ('\x61' <= c && c <= '\x7A')
 digit :: Char -> Bool
 digit c = '\x30' <= c && c <= '\x39'
 
+alphaNum :: Char -> Bool
+alphaNum c = alpha c || digit c
+
 headCharacter :: Char -> Bool
 headCharacter c = alpha c || c == '_'
 
 tailCharacter :: Char -> Bool
-tailCharacter c = alpha c || digit c || c == '_' || c == '-' || c == '/'
+tailCharacter c = alphaNum c || c == '_' || c == '-' || c == '/'
+
+-- | Escape a label if it is not valid when unquoted
+escapeLabel :: Bool -> Text -> Text
+escapeLabel allowReserved l =
+    case Text.uncons l of
+        Just (h, t)
+            | headCharacter h && Text.all tailCharacter t && (allowReserved || not (Data.HashSet.member l reservedIdentifiers))
+                -> l
+        _       -> "`" <> l <> "`"
 
 prettyLabelShared :: Bool -> Text -> Doc Ann
-prettyLabelShared allowReserved a = label doc
-    where
-        doc =
-            case Text.uncons a of
-                Just (h, t)
-                    | headCharacter h && Text.all tailCharacter t && (allowReserved || not (Data.HashSet.member a reservedIdentifiers))
-                        -> Pretty.pretty a
-                _       -> backtick <> Pretty.pretty a <> backtick
+prettyLabelShared b l = label (Pretty.pretty (escapeLabel b l))
 
 prettyLabel :: Text -> Doc Ann
 prettyLabel = prettyLabelShared False
@@ -402,6 +438,10 @@ prettyLabel = prettyLabelShared False
 prettyAnyLabel :: Text -> Doc Ann
 prettyAnyLabel = prettyLabelShared True
 
+prettyAnyLabels :: NonEmpty Text -> Doc Ann
+prettyAnyLabels =
+    mconcat . Pretty.punctuate dot . fmap prettyAnyLabel . toList
+
 prettyLabels :: Set Text -> Doc Ann
 prettyLabels a
     | Data.Set.null (Dhall.Set.toSet a) =
@@ -430,6 +470,32 @@ prettyVar :: Var -> Doc Ann
 prettyVar (V x 0) = label (Pretty.unAnnotate (prettyLabel x))
 prettyVar (V x n) = label (Pretty.unAnnotate (prettyLabel x <> "@" <> prettyInt n))
 
+prettyEnvironmentVariable :: Text -> Doc ann
+prettyEnvironmentVariable t = Pretty.pretty (escapeEnvironmentVariable t)
+
+-- | Escape an environment variable if not a valid Bash environment variable
+escapeEnvironmentVariable :: Text -> Text
+escapeEnvironmentVariable t
+  | validBashEnvVar t = t
+  | otherwise         = "\"" <> escapeText_ t <> "\""
+  where
+    validBashEnvVar v = case Text.uncons v of
+        Nothing      -> False
+        Just (c, v') ->
+                (alpha c || c == '_')
+            &&  Text.all (\c' -> alphaNum c' || c' == '_') v'
+
+{-  There is a close correspondence between the pretty-printers in 'prettyCharacterSet'
+    and the sub-parsers in 'Dhall.Parser.Expression.parsers'.  Most pretty-printers are
+    named after the corresponding parser and the relationship between pretty-printers
+    exactly matches the relationship between parsers.  This leads to the nice emergent
+    property of automatically getting all the parentheses and precedences right.
+
+    This approach has one major disadvantage: you can get an infinite loop if
+    you add a new constructor to the syntax tree without adding a matching
+    case the corresponding builder.
+-}
+
 {-| Pretty-print an 'Expr' using the given 'CharacterSet'.
 
 'prettyCharacterSet' largely ignores 'Note's. 'Note's do however matter for
@@ -519,37 +585,40 @@ prettyCharacterSet characterSet expression =
       where
         MultiLet as b = multiLet a0 b0
 
+        stripSpaces = Text.dropAround (\c -> c == ' ' || c == '\t')
+
+        -- Strip a single newline character. Needed to ensure idempotency in
+        -- cases where we add hard line breaks.
+        stripNewline t =
+            case Text.uncons t' of
+                Just ('\n', t'') -> stripSpaces t''
+                _ -> t'
+          where t' = stripSpaces t
+
         docA (Binding src0 c src1 Nothing src2 e) =
             Pretty.group (Pretty.flatAlt long short)
           where
             long =  keyword "let" <> space
                 <>  Pretty.align
-                    (   renderSrc src0 mempty
-                    <>  prettyLabel c <> renderSrc src1 space
-                    <>  equals <> renderSrc src2 Pretty.hardline
+                    (   renderSrc stripSpaces src0
+                    <>  prettyLabel c <> space <> renderSrc stripSpaces src1
+                    <>  equals <> Pretty.hardline <> renderSrc stripNewline src2
                     <>  "  " <> prettyExpression e
                     )
 
-            short = keyword "let" <> renderSrc src0 space
-                <>  prettyLabel c <> renderSrc src1 space
-                <>  equals <> renderSrc src2 space
+            short = keyword "let" <> space <> renderSrc stripSpaces src0
+                <>  prettyLabel c <> space <> renderSrc stripSpaces src1
+                <>  equals <> space <> renderSrc stripSpaces src2
                 <>  prettyExpression e
         docA (Binding src0 c src1 (Just (src3, d)) src2 e) =
-            Pretty.group (Pretty.flatAlt long short)
-          where
-            long = keyword "let" <> space
-                <>  Pretty.align
-                    (   renderSrc src0 mempty
-                    <>  prettyLabel c <> renderSrc src1 Pretty.hardline
-                    <>  colon <> renderSrc src3 space <> prettyExpression d <> Pretty.hardline <> equals <> renderSrc src2 space
-                    <>  prettyExpression e
-                    )
-
-            short = keyword "let" <> renderSrc src0 space
-                <>  prettyLabel c <> renderSrc src1 space
-                <>  colon <> renderSrc src3 space
-                <>  prettyExpression d <> space <> equals <> renderSrc src2 space
+                keyword "let" <> space
+            <>  Pretty.align
+                (   renderSrc stripSpaces src0
+                <>  prettyLabel c <> Pretty.hardline <> renderSrc stripNewline src1
+                <>  colon <> space <> renderSrc stripSpaces src3 <> prettyExpression d <> Pretty.hardline
+                <>  equals <> space <> renderSrc stripSpaces src2
                 <>  prettyExpression e
+                )
 
         docB =
             ( keyword "in" <> " " <> prettyExpression b
@@ -616,22 +685,6 @@ prettyCharacterSet characterSet expression =
             <>  prettyImportExpression b
             <>  space <> colon <> space
             <>  prettyApplicationExpression c
-    prettyAnnotatedExpression (Merge a b Nothing) =
-        Pretty.group (Pretty.flatAlt long short)
-      where
-        long =
-            Pretty.align
-                (   keyword "merge"
-                <>  Pretty.hardline
-                <>  Pretty.indent 2 (prettyImportExpression a)
-                <>  Pretty.hardline
-                <>  Pretty.indent 2 (prettyImportExpression b)
-                )
-
-        short = keyword "merge" <> space
-            <>  prettyImportExpression a
-            <>  " "
-            <>  prettyImportExpression b
     prettyAnnotatedExpression (ToMap a (Just b)) =
         Pretty.group (Pretty.flatAlt long short)
       where
@@ -649,18 +702,6 @@ prettyCharacterSet characterSet expression =
             <>  prettyImportExpression a
             <>  space <> colon <> space
             <>  prettyApplicationExpression b
-    prettyAnnotatedExpression (ToMap a Nothing) =
-        Pretty.group (Pretty.flatAlt long short)
-      where
-        long =
-            Pretty.align
-                (   keyword "toMap"
-                <>  Pretty.hardline
-                <>  Pretty.indent 2 (prettyImportExpression a)
-                )
-
-        short = keyword "toMap" <> space
-            <>  prettyImportExpression a
     prettyAnnotatedExpression a0@(Annot _ _) =
         enclose'
             ""
@@ -770,12 +811,12 @@ prettyCharacterSet characterSet expression =
        prettyCombineExpression a0
 
     prettyCombineExpression :: Pretty a => Expr Src a -> Doc Ann
-    prettyCombineExpression a0@(Combine _ _) =
+    prettyCombineExpression a0@(Combine _ _ _) =
         prettyOperator (combine characterSet) (docs a0)
       where
-        docs (Combine a b) = prettyPreferExpression b : docs a
-        docs (Note    _ b) = docs b
-        docs            b  = [ prettyPreferExpression b ]
+        docs (Combine _ a b) = prettyPreferExpression b : docs a
+        docs (Note      _ b) = docs b
+        docs              b  = [ prettyPreferExpression b ]
     prettyCombineExpression (Note _ a) =
         prettyCombineExpression a
     prettyCombineExpression a0 =
@@ -854,18 +895,23 @@ prettyCharacterSet characterSet expression =
         prettyApplicationExpression a0
 
     prettyApplicationExpression :: Pretty a => Expr Src a -> Doc Ann
-    prettyApplicationExpression a0 = case a0 of
-        App _ _  -> result
-        Some _   -> result
-        Note _ b -> prettyApplicationExpression b
-        _        -> prettyImportExpression a0
+    prettyApplicationExpression = go []
       where
-        result = enclose' "" "" " " "" (reverse (docs a0))
+        go args = \case
+            App a b           -> go (b : args) a
+            Some a            -> app (builtin "Some") (a : args)
+            Merge a b Nothing -> app (keyword "merge") (a : b : args)
+            ToMap a Nothing   -> app (keyword "toMap") (a : args)
+            Note _ b          -> go args b
+            e | null args     -> prettyImportExpression e -- just a performance optimization
+              | otherwise     -> app (prettyImportExpression e) args
 
-        docs (App  a b) = ( prettyImportExpression b, Pretty.indent 2 (prettyImportExpression b) ) : docs a
-        docs (Some   a) = map duplicate [ prettyImportExpression a , builtin "Some" ]
-        docs (Note _ b) = docs b
-        docs         b  = map duplicate [ prettyImportExpression b ]
+        app f args =
+            enclose'
+                "" "" " " ""
+                ( duplicate f
+                : map (fmap (Pretty.indent 2) . duplicate . prettyImportExpression) args
+                )
 
     prettyImportExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyImportExpression (Embed a) =
@@ -873,6 +919,24 @@ prettyCharacterSet characterSet expression =
     prettyImportExpression (Note _ a) =
         prettyImportExpression a
     prettyImportExpression a0 =
+        prettyCompletionExpression a0
+
+    prettyCompletionExpression :: Pretty a => Expr Src a -> Doc Ann
+    prettyCompletionExpression (RecordCompletion a b) =
+        case shallowDenote b of
+            RecordLit kvs ->
+                Pretty.align
+                    (   prettySelectorExpression a
+                    <>  doubleColon
+                    <>  prettyCompletionLit 0 kvs
+                    )
+            _ ->    prettySelectorExpression a
+                <>  doubleColon
+                <>  prettySelectorExpression b
+
+    prettyCompletionExpression (Note _ a) =
+        prettyCompletionExpression a
+    prettyCompletionExpression a0 =
         prettySelectorExpression a0
 
     prettySelectorExpression :: Pretty a => Expr Src a -> Doc Ann
@@ -918,6 +982,10 @@ prettyCharacterSet characterSet expression =
         builtin "Natural/subtract"
     prettyPrimitiveExpression Integer =
         builtin "Integer"
+    prettyPrimitiveExpression IntegerClamp =
+        builtin "Integer/clamp"
+    prettyPrimitiveExpression IntegerNegate =
+        builtin "Integer/negate"
     prettyPrimitiveExpression IntegerShow =
         builtin "Integer/show"
     prettyPrimitiveExpression IntegerToDouble =
@@ -963,7 +1031,7 @@ prettyCharacterSet characterSet expression =
         | otherwise = prettyNumber a
     prettyPrimitiveExpression (NaturalLit a) =
         prettyNatural a
-    prettyPrimitiveExpression (DoubleLit a) =
+    prettyPrimitiveExpression (DoubleLit (DhallDouble a)) =
         prettyDouble a
     prettyPrimitiveExpression (TextLit a) =
         prettyChunks a
@@ -986,73 +1054,142 @@ prettyCharacterSet characterSet expression =
 
         short = lparen <> prettyExpression a <> rparen
 
-    prettyKeyValue :: Pretty a => Doc Ann -> (Text, Expr Src a) -> (Doc Ann, Doc Ann)
-    prettyKeyValue separator (key, val) =
+    prettyKeyValue
+        :: Pretty a => (k -> Doc Ann) -> Doc Ann -> (k, Expr Src a) -> (Doc Ann, Doc Ann)
+    prettyKeyValue prettyKey separator (key, val) =
         duplicate (Pretty.group (Pretty.flatAlt long short))
       where
-        short = prettyAnyLabel key
+        completion _T r =
+                " "
+            <>  prettySelectorExpression _T
+            <>  doubleColon
+            <>  case shallowDenote r of
+                    RecordLit kvs ->
+                        prettyCompletionLit 2 kvs
+                    _ ->
+                        prettySelectorExpression r
+
+        short = prettyKey key
             <>  " "
             <>  separator
             <>  " "
             <>  prettyExpression val
 
-        long =  prettyAnyLabel key
+        long =
+                prettyKey key
             <>  " "
             <>  separator
-            <>  Pretty.hardline
-            <>  "    "
-            <>  prettyExpression val
+            <>  case shallowDenote val of
+                    Some val' ->
+                            " " <> builtin "Some"
+                        <>  case shallowDenote val' of
+                                RecordCompletion _T r ->
+                                    completion _T r
+                                ListLit _ xs
+                                    | not (null xs) ->
+                                            Pretty.hardline
+                                        <>  "  "
+                                        <>  prettyExpression val'
+                                _ ->    Pretty.hardline
+                                    <>  "    "
+                                    <>  prettyImportExpression val'
+                    ToMap val' Nothing ->
+                            " " <> keyword "toMap"
+                        <>  case shallowDenote val' of
+                                RecordCompletion _T r ->
+                                    completion _T r
+                                _ ->    Pretty.hardline
+                                    <>  "    "
+                                    <>  prettyImportExpression val'
+                    RecordCompletion _T r ->
+                        completion _T r
+                    ListLit _ xs
+                        | not (null xs) ->
+                                Pretty.hardline
+                            <>  "  "
+                            <>  prettyExpression val
+                    _ -> 
+                            Pretty.hardline
+                        <>  "    "
+                        <>  prettyExpression val
 
     prettyRecord :: Pretty a => Map Text (Expr Src a) -> Doc Ann
     prettyRecord =
-        braces . map (prettyKeyValue colon) . Dhall.Map.toList
+        braces . map (prettyKeyValue prettyAnyLabel colon) . Map.toList
 
     prettyRecordLit :: Pretty a => Map Text (Expr Src a) -> Doc Ann
     prettyRecordLit a
         | Data.Foldable.null a =
             lbrace <> equals <> rbrace
-        | otherwise
-            = braces (map (prettyKeyValue equals) (Dhall.Map.toList a))
+        | otherwise =
+            braces (map prettyRecordEntry (Map.toList consolidated))
+      where
+        consolidated = consolidateRecordLiteral a
 
-    prettyAlternative (key, Just val) = prettyKeyValue colon (key, val)
+        prettyRecordEntry = prettyKeyValue prettyAnyLabels equals
+
+    prettyCompletionLit
+        :: Pretty a => Int -> Map Text (Expr Src a) -> Doc Ann
+    prettyCompletionLit n a
+        | Data.Foldable.null a =
+            lbrace <> equals <> rbrace
+        | otherwise =
+            hangingBraces n (map prettyRecordEntry (Map.toList consolidated))
+      where
+        consolidated = consolidateRecordLiteral a
+
+        prettyRecordEntry = prettyKeyValue prettyAnyLabels equals
+
+    prettyAlternative (key, Just val) = prettyKeyValue prettyAnyLabel colon (key, val)
     prettyAlternative (key, Nothing ) = duplicate (prettyAnyLabel key)
 
     prettyUnion :: Pretty a => Map Text (Maybe (Expr Src a)) -> Doc Ann
     prettyUnion =
-        angles . map prettyAlternative . Dhall.Map.toList
+        angles . map prettyAlternative . Map.toList
 
     prettyChunks :: Pretty a => Chunks Src a -> Doc Ann
-    prettyChunks (Chunks a b) =
-        if any (\(builder, _) -> hasNewLine builder) a || hasNewLine b
-        then Pretty.flatAlt long short
-        else short
+    prettyChunks chunks@(Chunks a b)
+        | anyText (== '\n') =
+            if not (null a) || anyText (/= '\n')
+            then long
+            else Pretty.group (Pretty.flatAlt long short)
+        | otherwise =
+            short
       where
         long =
             Pretty.align
-            (   literal ("''" <> Pretty.hardline)
+            (   literal "''" <> Pretty.hardline
             <>  Pretty.align
-                (foldMap prettyMultilineChunk a <> prettyMultilineBuilder b)
+                (foldMap prettyMultilineChunk a' <> prettyMultilineText b')
             <>  literal "''"
             )
+          where
+            Chunks a' b' = multilineChunks chunks
 
         short =
             literal "\"" <> foldMap prettyChunk a <> literal (prettyText b <> "\"")
 
-        hasNewLine = Text.any (== '\n')
+        anyText predicate = any (\(text, _) -> Text.any predicate text) a || Text.any predicate b
 
         prettyMultilineChunk (c, d) =
-                prettyMultilineBuilder c
+                prettyMultilineText c
             <>  dollar
             <>  lbrace
             <>  prettyExpression d
             <>  rbrace
 
-        prettyMultilineBuilder builder = literal (mconcat docs)
+        prettyMultilineText text = mconcat docs
           where
-            lazyLines = Text.splitOn "\n" (escapeSingleQuotedText builder)
+            lines_ = Text.splitOn "\n" (escapeSingleQuotedText text)
+
+            -- Annotate only non-empty lines so trailing whitespace can be
+            -- removed on empty ones.
+            prettyLine line =
+                (if Text.null line then id else literal)
+                    (Pretty.pretty line)
 
             docs =
-                Data.List.intersperse Pretty.hardline (fmap Pretty.pretty lazyLines)
+                Data.List.intersperse Pretty.hardline (map prettyLine lines_)
 
         prettyChunk (c, d) =
                 prettyText c
@@ -1060,22 +1197,139 @@ prettyCharacterSet characterSet expression =
             <>  prettyExpression d
             <>  syntax rbrace
 
-        prettyText t = literal (Pretty.pretty (escapeText t))
+        prettyText t = literal (Pretty.pretty (escapeText_ t))
+
+
+-- | Prepare 'Chunks' for multi-line formatting by escaping problematic
+-- character sequences via string interpolations
+--
+-- >>> multilineChunks (Chunks [] "\n \tx")
+-- Chunks [("\n",TextLit (Chunks [] " \t"))] "x"
+-- >>> multilineChunks (Chunks [] "\n\NUL\b\f\t")
+-- Chunks [("\n",TextLit (Chunks [] "\NUL\b\f"))] "\t"
+multilineChunks :: Chunks s a -> Chunks s a
+multilineChunks =
+     escapeTrailingSingleQuote
+   . escapeControlCharacters
+   . escapeSharedWhitespacePrefix
+
+-- | Escape any leading whitespace shared by all lines
+--
+-- This ensures that significant shared leading whitespace is not stripped
+--
+-- >>> escapeSharedWhitespacePrefix (Chunks [] "\n \tx")
+-- Chunks [("\n",TextLit (Chunks [] " \t"))] "x"
+-- >>> escapeSharedWhitespacePrefix (Chunks [("\n",Var (V "x" 0))] " ")
+-- Chunks [("\n",Var (V "x" 0))] " "
+-- >>> escapeSharedWhitespacePrefix (Chunks [("\n ",Var (V "x" 0))] "")
+-- Chunks [("\n",TextLit (Chunks [] " ")),("",Var (V "x" 0))] ""
+-- >>> escapeSharedWhitespacePrefix (Chunks [("\n ",Var (V "x" 0))] "\n")
+-- Chunks [("\n ",Var (V "x" 0))] "\n"
+-- >>> escapeSharedWhitespacePrefix (Chunks [] " ")
+-- Chunks [("",TextLit (Chunks [] " "))] ""
+escapeSharedWhitespacePrefix :: Chunks s a -> Chunks s a
+escapeSharedWhitespacePrefix literal_ = unlinesLiteral literals₁
+  where
+    literals₀ = linesLiteral literal_
+
+    sharedPrefix = longestSharedWhitespacePrefix literals₀
+
+    stripPrefix = Text.drop (Text.length sharedPrefix)
+
+    escapeSharedPrefix (Chunks [] prefix₀)
+        | Text.isPrefixOf sharedPrefix prefix₀ =
+            Chunks [ ("", TextLit (Chunks [] sharedPrefix)) ] prefix₁
+      where
+        prefix₁ = stripPrefix prefix₀
+    escapeSharedPrefix (Chunks ((prefix₀, y) : xys) z)
+        | Text.isPrefixOf sharedPrefix prefix₀ =
+            Chunks (("", TextLit (Chunks [] sharedPrefix)) : (prefix₁, y) : xys) z
+      where
+        prefix₁ = stripPrefix prefix₀
+    escapeSharedPrefix line = line
+
+    literals₁
+        | not (Text.null sharedPrefix) = fmap escapeSharedPrefix literals₀
+        | otherwise = literals₀
+
+-- | Escape control characters by moving them into string interpolations
+--
+-- >>> escapeControlCharacters (Chunks [] "\n\NUL\b\f\t")
+-- Chunks [("\n",TextLit (Chunks [] "\NUL\b\f"))] "\t"
+escapeControlCharacters :: Chunks s a -> Chunks s a
+escapeControlCharacters (Chunks as0 b0) = Chunks as1 b1
+  where
+    as1 = foldr f (map toChunk bs) as0
+
+    (bs, b1) = splitOnPredicate predicate b0
+
+    f (t0, e) chunks = map toChunk ts1 ++ (t1, e) : chunks
+      where
+        (ts1, t1) = splitOnPredicate predicate t0
+
+    predicate c = Data.Char.isControl c && c /= ' ' && c /= '\t' && c /= '\n'
+
+    toChunk (t0, t1) = (t0, TextLit (Chunks [] t1))
+
+-- | Split `Text` on a predicate, preserving all parts of the original string.
+--
+-- >>> splitOnPredicate (== 'x') ""
+-- ([],"")
+-- >>> splitOnPredicate (== 'x') " xx "
+-- ([(" ","xx")]," ")
+-- >>> splitOnPredicate (== 'x') "xx"
+-- ([("","xx")],"")
+--
+-- prop> \(Fun _ p) s -> let {t = Text.pack s; (as, b) = splitOnPredicate p t} in foldMap (uncurry (<>)) as <> b == t
+splitOnPredicate :: (Char -> Bool) -> Text -> ([(Text, Text)], Text)
+splitOnPredicate p t = case Text.break p t of
+    (a, "") -> ([], a)
+    (a, b)  -> case Text.span p b of
+        (c, d) -> case splitOnPredicate p d of
+            (e, f) -> ((a, c) : e, f)
+
+-- | Escape a trailing single quote by moving it into a string interpolation
+--
+-- Otherwise the multiline-string would end with @'''@, which would be parsed
+-- as an escaped @''@.
+--
+-- >>> escapeTrailingSingleQuote (Chunks [] "\n'")
+-- Chunks [("\n",TextLit (Chunks [] "'"))] ""
+escapeTrailingSingleQuote :: Chunks s a -> Chunks s a
+escapeTrailingSingleQuote chunks@(Chunks as b) =
+    case Text.unsnoc b of
+        Just (b', '\'') -> Chunks (as ++ [(b', TextLit (Chunks [] "'"))]) ""
+        _               -> chunks
 
 -- | Pretty-print a value
-pretty :: Pretty a => a -> Text
-pretty = Pretty.renderStrict . Pretty.layoutPretty options . Pretty.pretty
+pretty_ :: Pretty a => a -> Text
+pretty_ = prettyToStrictText
+
+{- This utility function converts
+   `{ x = { y = { z = 1 } } }` to `{ x.y.z. = 1 }`
+-}
+consolidateRecordLiteral
+    :: Map Text (Expr s a) -> Map (NonEmpty Text) (Expr s a)
+consolidateRecordLiteral = Map.fromList . fmap adapt . Map.toList
   where
-   options = Pretty.LayoutOptions { Pretty.layoutPageWidth = Pretty.Unbounded }
+    adapt :: (Text, Expr s a) -> (NonEmpty Text, Expr s a)
+    adapt (key, expression) =
+        case shallowDenote expression of
+            RecordLit m ->
+                case fmap adapt (Map.toList m) of
+                    [ (keys, expression') ] ->
+                        (NonEmpty.cons key keys, expression')
+                    _ ->
+                        (pure key, RecordLit m)
+            _ ->
+                (pure key, expression)
 
 -- | Escape a `Text` literal using Dhall's escaping rules for single-quoted
 --   @Text@
 escapeSingleQuotedText :: Text -> Text
-escapeSingleQuotedText inputBuilder = outputBuilder
+escapeSingleQuotedText inputText = outputText
   where
-    outputText = substitute "${" "''${" (substitute "''" "'''" inputBuilder)
-
-    outputBuilder = outputText
+    outputText = substitute "${" "''${" (substitute "''" "'''" inputText)
 
     substitute before after = Text.intercalate after . Text.splitOn before
 
@@ -1083,8 +1337,8 @@ escapeSingleQuotedText inputBuilder = outputBuilder
 
     Note that the result does not include surrounding quotes
 -}
-escapeText :: Text -> Text
-escapeText text = Text.concatMap adapt text
+escapeText_ :: Text -> Text
+escapeText_ text = Text.concatMap adapt text
   where
     adapt c
         | '\x20' <= c && c <= '\x21'     = Text.singleton c
@@ -1116,14 +1370,29 @@ escapeText text = Text.concatMap adapt text
 
 prettyToString :: Pretty a => a -> String
 prettyToString =
-    Pretty.renderString . Pretty.layoutPretty options . Pretty.pretty
-  where
-   options = Pretty.LayoutOptions { Pretty.layoutPageWidth = Pretty.Unbounded }
+    Pretty.renderString . layout . Pretty.pretty
 
 docToStrictText :: Doc ann -> Text.Text
-docToStrictText = Pretty.renderStrict . Pretty.layoutPretty options
-  where
-   options = Pretty.LayoutOptions { Pretty.layoutPageWidth = Pretty.Unbounded }
+docToStrictText = Pretty.renderStrict . layout
 
 prettyToStrictText :: Pretty a => a -> Text.Text
 prettyToStrictText = docToStrictText . Pretty.pretty
+
+-- | Layout using 'layoutOpts'
+--
+-- Tries hard to fit the document into 80 columns.
+--
+-- This also removes trailing space characters (@' '@) /unless/
+-- they are enclosed in an annotation.
+layout :: Pretty.Doc ann -> Pretty.SimpleDocStream ann
+layout = Pretty.removeTrailingWhitespace . Pretty.layoutSmart layoutOpts
+
+-- | Default layout options
+layoutOpts :: Pretty.LayoutOptions
+layoutOpts =
+    Pretty.defaultLayoutOptions
+        { Pretty.layoutPageWidth = Pretty.AvailablePerLine 80 1.0 }
+
+{- $setup
+>>> import Test.QuickCheck (Fun(..))
+-}
