@@ -22,8 +22,10 @@ import Data.List ( isPrefixOf, nub )
 import Data.Maybe ( mapMaybe )
 import Data.Semigroup ((<>))
 import Data.Text ( Text )
+import Data.Void (Void)
 import Dhall.Context (Context)
 import Dhall.Import (hashExpressionToCode)
+import Dhall.Parser (Parser(..))
 import Dhall.Src (Src)
 import Dhall.Pretty (CharacterSet(..))
 import System.Console.Haskeline (Interrupt(..))
@@ -42,18 +44,21 @@ import qualified Dhall
 import qualified Dhall.Context
 import qualified Dhall.Core
 import qualified Dhall.Core as Dhall ( Var(V), Expr, normalize )
+import qualified Dhall.Parser.Token                      as Parser.Token
 import qualified Dhall.Pretty
+import qualified Dhall.Pretty.Internal
 import qualified Dhall.Core as Expr ( Expr(..) )
-import qualified Dhall.Import as Dhall
-import qualified Dhall.Map as Map
-import qualified Dhall.Parser as Dhall
-import qualified Dhall.TypeCheck as Dhall
-import qualified Dhall.Version as Meta
+import qualified Dhall.Import                            as Dhall
+import qualified Dhall.Map                               as Map
+import qualified Dhall.Parser                            as Dhall
+import qualified Dhall.TypeCheck                         as Dhall
+import qualified Dhall.Version                           as Meta
 import qualified System.Console.ANSI
-import qualified System.Console.Haskeline.Completion as Haskeline
+import qualified System.Console.Haskeline.Completion     as Haskeline
 import qualified System.Console.Haskeline.MonadException as Haskeline
-import qualified System.Console.Repline as Repline
+import qualified System.Console.Repline                  as Repline
 import qualified System.IO
+import qualified Text.Megaparsec                         as Megaparsec
 
 type Repl = Repline.HaskelineT (State.StateT Env IO)
 
@@ -100,8 +105,8 @@ emptyEnv =
 
 
 data Binding = Binding
-  { bindingExpr :: Dhall.Expr Dhall.Src Dhall.X
-  , bindingType :: Dhall.Expr Dhall.Src Dhall.X
+  { bindingExpr :: Dhall.Expr Dhall.Src Void
+  , bindingType :: Dhall.Expr Dhall.Src Void
   }
 
 
@@ -116,10 +121,10 @@ envToContext Env{ envBindings, envIt } =
 
 
 parseAndLoad
-  :: MonadIO m => String -> m ( Dhall.Expr Dhall.Src Dhall.X )
+  :: MonadIO m => String -> m ( Dhall.Expr Dhall.Src Void) 
 parseAndLoad src = do
   parsed <-
-    case Dhall.exprFromText "(stdin)" ( Text.pack src ) of
+    case Dhall.exprFromText "(stdin)" (Text.pack src <> "\n") of
       Left e ->
         liftIO ( throwIO e )
 
@@ -158,16 +163,13 @@ typeOf srcs = do
   exprType <-
     typeCheck loaded
 
-  exprType' <-
-    normalize exprType
-
-  output exprType'
+  output exprType
 
 
 applyContext
     :: Context Binding
-    -> Dhall.Expr Dhall.Src Dhall.X
-    -> Dhall.Expr Dhall.Src Dhall.X
+    -> Dhall.Expr Dhall.Src Void
+    -> Dhall.Expr Dhall.Src Void
 applyContext context expression =
     Dhall.Core.wrapInLets bindings expression
   where
@@ -180,7 +182,7 @@ applyContext context expression =
 
 normalize
   :: MonadState Env m
-  => Dhall.Expr Dhall.Src Dhall.X -> m ( Dhall.Expr t Dhall.X )
+  => Dhall.Expr Dhall.Src Void -> m ( Dhall.Expr t Void )
 normalize e = do
   env <- get
 
@@ -189,7 +191,7 @@ normalize e = do
 
 typeCheck
   :: ( MonadIO m, MonadState Env m )
-  => Dhall.Expr Dhall.Src Dhall.X -> m ( Dhall.Expr Dhall.Src Dhall.X )
+  => Dhall.Expr Dhall.Src Void -> m ( Dhall.Expr Dhall.Src Void )
 typeCheck expression = do
   env <- get
 
@@ -214,23 +216,19 @@ separateEqual (str : strs)
 
 addBinding :: ( MonadFail m, MonadIO m, MonadState Env m ) => [String] -> m ()
 addBinding (k : "=" : srcs) = do
-  let
-    varName =
-      Text.pack k
+  varName <- case Megaparsec.parse (unParser Parser.Token.label) "(stdin)" (Text.pack k) of
+      Left   _      -> Fail.fail "Invalid variable name"
+      Right varName -> return varName
 
-  loaded <-
-    parseAndLoad ( unwords srcs )
+  loaded <- parseAndLoad ( unwords srcs )
 
-  t <-
-    typeCheck loaded
+  t <- typeCheck loaded
 
-  expr <-
-    normalize loaded
+  expr <- normalize loaded
 
   modify
     ( \e ->
-        e
-          { envBindings =
+        e { envBindings =
               Dhall.Context.insert
                 varName
                 Binding { bindingType = t, bindingExpr = expr }
@@ -353,7 +351,11 @@ saveBinding [file] = do
       handler handle =
           State.evalStateT
             (forM_ bindings $ \(name, expr) -> do
-              liftIO (System.IO.hPutStr handle $ ":let " <> Text.unpack name <> " = ")
+              let doc = Dhall.Pretty.Internal.prettyLabel name
+
+              let label = Dhall.Pretty.Internal.docToStrictText doc
+
+              liftIO (System.IO.hPutStr handle $ ":let " <> Text.unpack label <> " = ")
               outputWithoutSpacing expr)
             (env { outputHandle = Just handle })
 
@@ -546,7 +548,7 @@ completeFunc reversedPrev word
     listCompletion = map simpleCompletion . filter (word `isPrefixOf`)
 
     algebraicComplete
-        :: [Text.Text] -> Dhall.Expr Dhall.Src Dhall.X -> [Text.Text]
+        :: [Text.Text] -> Dhall.Expr Dhall.Src Void -> [Text.Text]
     algebraicComplete subFields expr =
       let keys = fmap ("." <>) . Map.keys
 
@@ -610,9 +612,7 @@ outputWithoutSpacing expr = do
   case outputHandle of
     Nothing     -> pure ()
     Just handle -> do
-      let stream =
-              Pretty.layoutSmart Dhall.Pretty.layoutOpts
-                  (Dhall.Pretty.prettyCharacterSet characterSet expr)
+      let stream = Dhall.Pretty.layout (Dhall.Pretty.prettyCharacterSet characterSet expr)
 
       supportsANSI <- liftIO (System.Console.ANSI.hSupportsANSI handle)
       let ansiStream =
