@@ -10,6 +10,7 @@
 module Dhall.Set (
       Set(..)
     , toList
+    , toAscList
     , toSet
     , toSeq
     , fromList
@@ -20,6 +21,7 @@ module Dhall.Set (
     , sort
     , isSorted
     , null
+    , size
     ) where
 
 import Prelude hiding (null)
@@ -34,10 +36,15 @@ import Language.Haskell.TH.Syntax (Lift)
 import qualified Data.Set
 import qualified Data.Sequence
 import qualified Data.Foldable
+import qualified Language.Haskell.TH.Syntax as Syntax
 
--- Invariant: In @Set set seq@, @toAscList set == sort (toList seq)@.
+{-| This is a variation on @"Data.Set".`Data.Set.Set`@ that remembers the
+    original order of elements.  This ensures that ordering is not lost when
+    formatting Dhall code
+-}
 data Set a = Set (Data.Set.Set a) (Seq a)
     deriving (Generic, Show, Data, NFData)
+-- Invariant: In @Set set seq@, @toAscList set == sort (toList seq)@.
 
 instance Eq a => Eq (Set a) where
     (Set _ x) == (Set _ y) = x == y
@@ -47,36 +54,57 @@ instance Ord a => Ord (Set a) where
     compare (Set _ x) (Set _ y) = compare x y
     {-# INLINABLE compare #-}
 
-instance (Data a, Lift a, Ord a) => Lift (Set a)
+instance (Data a, Lift a, Ord a) => Lift (Set a) where
+    lift = Syntax.liftData
 
 instance Foldable Set where
     foldMap f (Set _ x) = foldMap f x
     {-# INLINABLE foldMap #-}
 
+    toList = Dhall.Set.toList
+    {-# INLINABLE toList #-}
+
+    null = Dhall.Set.null
+    {-# INLINABLE null #-}
+
+    length = Dhall.Set.size
+    {-# INLINABLE length #-}
+
+-- | Convert to an unordered @"Data.Set".`Data.Set.Set`@
 toSet :: Set a -> Data.Set.Set a
 toSet (Set s _) = s
 
+-- | Convert to an ordered `Seq`
 toSeq :: Set a -> Seq a
 toSeq (Set _ xs) = xs
 
+-- | Convert a `Set` to a list, preserving the original order of the elements
 toList :: Set a -> [a]
-toList = Data.Foldable.toList
+toList (Set _ xs) = Data.Foldable.toList xs
 
+-- | Convert a `Set` to a list of ascending elements
+toAscList :: Set a -> [a]
+toAscList (Set s _) = Data.Set.toAscList s
+
+-- | Convert a list to a `Set`, remembering the element order
+fromList :: Ord a => [a] -> Set a
+fromList = foldl' (flip append) empty
 -- O(n log n) time complexity, O(n) space complexity.
 -- Implementing it this way is a little silly, but is faster than (nub xs).
 -- n.b. toList . fromList = id, only if the list elements are unique
-fromList :: Ord a => [a] -> Set a
-fromList = foldl' (flip append) empty
 
+-- | Convert a @"Data.Set".`Data.Set.Set`@ to a sorted `Set`
 fromSet :: Data.Set.Set a -> Set a
 fromSet s = Set s (Data.Sequence.fromList (Data.Set.elems s))
 
--- O(log n) time complexity.
+-- | Append an element to the end of a `Set`
 append :: Ord a => a -> Set a -> Set a
 append x os@(Set s xs)
     | Data.Set.member x s = os
     | otherwise = Set (Data.Set.insert x s) (xs |> x)
+-- O(log n) time complexity.
 
+-- | The empty `Set`
 empty :: Set a
 empty = Set Data.Set.empty Data.Sequence.empty
 
@@ -92,7 +120,7 @@ difference os (Set s _) =
 True
 -}
 sort :: Ord a => Set a -> Set a
-sort (Set s xs) = Set s (Data.Sequence.sort xs)
+sort (Set s _) = Set s (Data.Sequence.fromList (Data.Set.toList s))
 
 {-|
 >>> isSorted (fromList [2, 1])
@@ -111,3 +139,10 @@ True
 -}
 null :: Set a -> Bool
 null (Set s _) = Data.Set.null s
+
+{-|
+>>> size (fromList [1])
+1
+-}
+size :: Set a -> Int
+size (Set s _) = Data.Set.size s
