@@ -22,7 +22,7 @@ import Data.Sequence (Seq)
 import Data.Scientific (Scientific)
 import Data.Text (Text)
 import Data.Vector (Vector)
-import Dhall (Inject, Interpret)
+import Dhall (ToDhall, FromDhall)
 import Dhall.Core (Expr(..))
 import GHC.Generics (Generic)
 import Numeric.Natural (Natural)
@@ -40,14 +40,14 @@ data ExprF expr
    = LitF Natural
    | AddF expr expr
    | MulF expr expr
-   deriving (Eq, Functor, Generic, Interpret, Show)
+   deriving (Eq, Functor, Generic, FromDhall, Show)
 
 tests :: TestTree
 tests =
     testGroup "Input"
      [ shouldShowDetailedTypeError
      , shouldHandleUnionLiteral
-     , shouldHaveWorkingRecursiveInterpret
+     , shouldHaveWorkingRecursiveFromDhall
      , shouldHaveWorkingGenericAuto
      , shouldHandleUnionsCorrectly
      , shouldTreatAConstructorStoringUnitAsEmptyAlternative
@@ -57,8 +57,8 @@ tests =
 
 data MyType = MyType { foo :: String , bar :: Natural }
 
-wrongDhallType :: Dhall.Type MyType
-wrongDhallType = Dhall.Type { .. }
+wrongDhallType :: Dhall.Decoder MyType
+wrongDhallType = Dhall.Decoder { .. }
   where expected =
           Dhall.Core.Record
             ( Dhall.Map.fromList
@@ -74,12 +74,12 @@ shouldShowDetailedTypeError = testCase "detailed TypeError" $ do
         try ( Dhall.input wrongDhallType "{  bar = 0, foo = \"foo\" }")
 
   let expectedMsg =
-        "\ESC[1;31mError\ESC[0m: Invalid Dhall.Type                                                  \n\
+        "\ESC[1;31mError\ESC[0m: Invalid Dhall.Decoder                                               \n\
         \                                                                                \n\
-        \Every Type must provide an extract function that succeeds if an expression      \n\
-        \matches the expected type.  You provided a Type that disobeys this contract     \n\
+        \Every Decoder must provide an extract function that succeeds if an expression   \n\
+        \matches the expected type.  You provided a Decoder that disobeys this contract  \n\
         \                                                                                \n\
-        \The Type provided has the expected dhall type:                                  \n\
+        \The Decoder provided has the expected dhall type:                               \n\
         \                                                                                \n\
         \↳ { bar : Natural, foo : Text }\n\
         \                                                                                \n\
@@ -92,12 +92,12 @@ shouldShowDetailedTypeError = testCase "detailed TypeError" $ do
 
   case inputEx of
     Left ex -> assertEqual assertMsg expectedMsg (show ex)
-    Right _ -> fail "The extraction using a wrong type succeded"
+    Right _ -> fail "The extraction using a wrong type succeeded"
 
 -- https://github.com/dhall-lang/dhall-haskell/issues/915
 shouldHandleUnionLiteral :: TestTree
 shouldHandleUnionLiteral = testCase "Marshal union literals" $ do
-    let example :: Dhall.Type Bool
+    let example :: Dhall.Decoder Bool
         example = Dhall.union (Dhall.constructor "Test" Dhall.bool)
 
     _ <- Dhall.input example "< Test : Bool >.Test True"
@@ -106,18 +106,18 @@ shouldHandleUnionLiteral = testCase "Marshal union literals" $ do
 
 shouldTreatAConstructorStoringUnitAsEmptyAlternative :: TestTree
 shouldTreatAConstructorStoringUnitAsEmptyAlternative = testCase "Handle unit constructors" $ do
-    let exampleType :: Dhall.Type ()
+    let exampleType :: Dhall.Decoder ()
         exampleType = Dhall.union (Dhall.constructor "A" Dhall.unit)
 
     () <- Dhall.input exampleType "< A >.A"
 
-    let exampleInputType :: Dhall.InputType ()
-        exampleInputType = Dhall.inputUnion (Dhall.inputConstructor "A")
+    let exampleEncoder :: Dhall.Encoder ()
+        exampleEncoder = Dhall.unionEncoder (Dhall.encodeConstructor "A")
 
-    Dhall.embed exampleInputType () @=? Field (Union (Dhall.Map.singleton "A" Nothing)) "A"
+    Dhall.embed exampleEncoder () @=? Field (Union (Dhall.Map.singleton "A" Nothing)) "A"
 
-shouldHaveWorkingRecursiveInterpret :: TestTree
-shouldHaveWorkingRecursiveInterpret = testGroup "recursive Interpret instance"
+shouldHaveWorkingRecursiveFromDhall :: TestTree
+shouldHaveWorkingRecursiveFromDhall = testGroup "recursive FromDhall instance"
     [ testCase "works for a recursive expression" $ do
         actual <- Dhall.input Dhall.auto "./tests/recursive/expr0.dhall"
 
@@ -156,10 +156,10 @@ shouldHaveWorkingGenericAuto = testGroup "genericAuto"
   ]
 
 data NonEmptyUnion = N0 Bool | N1 Natural | N2 Text
-    deriving (Eq, Generic, Inject, Interpret, Show)
+    deriving (Eq, Generic, ToDhall, FromDhall, Show)
 
 data Enum = E0 | E1 | E2
-    deriving (Eq, Generic, Inject, Interpret, Show)
+    deriving (Eq, Generic, ToDhall, FromDhall, Show)
 
 data Records
     = R0 {}
@@ -167,24 +167,22 @@ data Records
     | R2 { x :: Double }
     | R3 { a :: (), b :: () }
     | R4 { x :: Double, y :: Double }
-    deriving (Eq, Generic, Inject, Interpret, Show)
+    deriving (Eq, Generic, ToDhall, FromDhall, Show)
 
 data Products = P0 | P1 () | P2 Double | P3 () () | P4 Double Double
-    deriving (Eq, Generic, Inject, Interpret, Show)
-
-deriving instance Interpret ()
+    deriving (Eq, Generic, ToDhall, FromDhall, Show)
 
 shouldHandleUnionsCorrectly :: TestTree
 shouldHandleUnionsCorrectly =
   testGroup "Handle union literals"
     [ "λ(x : < N0 : { _1 : Bool } | N1 : { _1 : Natural } | N2 : { _1 : Text } >) → x"
-        `shouldPassThrough` [ N0 True, N1 5, N2 "ABC" ]
+        `shouldPassThroughWrapped` [ N0 True, N1 5, N2 "ABC" ]
     , "λ(x : < E0 | E1 | E2 >) → x"
-        `shouldPassThrough` [ E0, E1, E2 ]
+        `shouldPassThroughWrapped` [ E0, E1, E2 ]
     , "λ(x : < R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >) → x"
-        `shouldPassThrough` [ R0 {}, R1 { a = () }, R2 { x = 1.0 }, R3 { a = (), b = () }, R4 { x = 1.0, y = 2.0 } ]
+        `shouldPassThroughWrapped` [ R0 {}, R1 { a = () }, R2 { x = 1.0 }, R3 { a = (), b = () }, R4 { x = 1.0, y = 2.0 } ]
     , "λ(x : < P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >) → x"
-        `shouldPassThrough` [ P0 , P1 (), P2 1.0, P3 () (), P4 1.0 2.0 ]
+        `shouldPassThroughWrapped` [ P0 , P1 (), P2 1.0, P3 () (), P4 1.0 2.0 ]
 
     , "λ(x : < N0 : Bool | N1 : Natural | N2 : Text >) → x"
         `shouldPassThroughSmart` [ N0 True, N1 5, N2 "ABC" ]
@@ -194,12 +192,12 @@ shouldHandleUnionsCorrectly =
         `shouldPassThroughSmart` [ P0 , P1 (), P2 1.0, P3 () (), P4 1.0 2.0 ]
 
     , "(< N0 : { _1 : Bool } | N1 : { _1 : Natural } | N2 : { _1 : Text } >).N0 { _1 = True }"
-        `shouldMarshalInto` N0 True
+        `shouldMarshalIntoWrapped` N0 True
     , "(< N0 : { _1 : Bool } | N1 : { _1 : Natural } | N2 : { _1 : Text } >).N1 { _1 = 5 }"
-        `shouldMarshalInto` N1 5
+        `shouldMarshalIntoWrapped` N1 5
     , "(< N0 : { _1 : Bool } | N1 : { _1 : Natural } | N2 : { _1 : Text } >).N2 { _1 = \"ABC\" }"
 
-        `shouldMarshalInto` N2 "ABC"
+        `shouldMarshalIntoWrapped` N2 "ABC"
 
     , "(< N0 : Bool | N1 : Natural | N2 : Text >).N0 True"
         `shouldMarshalIntoSmart` N0 True
@@ -208,20 +206,20 @@ shouldHandleUnionsCorrectly =
     , "(< N0 : Bool | N1 : Natural | N2 : Text >).N2 \"ABC\""
         `shouldMarshalIntoSmart` N2 "ABC"
 
-    , "(< E0 | E1 | E2>).E0" `shouldMarshalInto` E0
-    , "(< E0 | E1 | E2>).E1" `shouldMarshalInto` E1
-    , "(< E0 | E1 | E2>).E2" `shouldMarshalInto` E2
+    , "(< E0 | E1 | E2>).E0" `shouldMarshalIntoWrapped` E0
+    , "(< E0 | E1 | E2>).E1" `shouldMarshalIntoWrapped` E1
+    , "(< E0 | E1 | E2>).E2" `shouldMarshalIntoWrapped` E2
 
     , "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R0"
-        `shouldMarshalInto` R0
+        `shouldMarshalIntoWrapped` R0
     , "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R1 { a = {=} }"
-        `shouldMarshalInto` R1 { a = () }
+        `shouldMarshalIntoWrapped` R1 { a = () }
     , "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R2 { x = 1.0 }"
-        `shouldMarshalInto` R2 { x = 1.0 }
+        `shouldMarshalIntoWrapped` R2 { x = 1.0 }
     , "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R3 { a = {=}, b = {=} }"
-        `shouldMarshalInto` R3 { a = (), b = () }
+        `shouldMarshalIntoWrapped` R3 { a = (), b = () }
     , "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R4 { x = 1.0, y = 2.0 }"
-        `shouldMarshalInto` R4 { x = 1.0, y = 2.0 }
+        `shouldMarshalIntoWrapped` R4 { x = 1.0, y = 2.0 }
 
     , "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R0"
         `shouldMarshalIntoSmart` R0
@@ -235,15 +233,15 @@ shouldHandleUnionsCorrectly =
         `shouldMarshalIntoSmart` R4 { x = 1.0, y = 2.0 }
 
     , "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P0"
-        `shouldMarshalInto` P0
+        `shouldMarshalIntoWrapped` P0
     , "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P1 { _1 = {=} }"
-        `shouldMarshalInto` P1 ()
+        `shouldMarshalIntoWrapped` P1 ()
     , "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P2 { _1 = 1.0 }"
-        `shouldMarshalInto` P2 1.0
+        `shouldMarshalIntoWrapped` P2 1.0
     , "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P3 { _1 = {=}, _2 = {=} }"
-        `shouldMarshalInto` P3 () ()
+        `shouldMarshalIntoWrapped` P3 () ()
     , "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P4 { _1 = 1.0, _2 = 2.0 }"
-        `shouldMarshalInto` P4 1.0 2.0
+        `shouldMarshalIntoWrapped` P4 1.0 2.0
 
     , "< P0 | P1 | P2 : Double | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P0"
         `shouldMarshalIntoSmart` P0
@@ -257,60 +255,64 @@ shouldHandleUnionsCorrectly =
         `shouldMarshalIntoSmart` P4 1.0 2.0
 
     , N0 True
-        `shouldInjectInto`
+        `shouldEmbedAsWrapped`
         "(< N0 : { _1 : Bool } | N1 : { _1 : Natural } | N2 : { _1 : Text } >).N0 { _1 = True }"
     , N1 5
-        `shouldInjectInto`
+        `shouldEmbedAsWrapped`
         "(< N0 : { _1 : Bool } | N1 : { _1 : Natural } | N2 : { _1 : Text } >).N1 { _1 = 5 }"
     , N2 "ABC"
-        `shouldInjectInto`
+        `shouldEmbedAsWrapped`
         "(< N0 : { _1 : Bool } | N1 : { _1 : Natural } | N2 : { _1 : Text } >).N2 { _1 = \"ABC\" }"
 
     , N0 True
-        `shouldInjectIntoSmart`
+        `shouldEmbedAsSmart`
         "(< N0 : Bool | N1 : Natural | N2 : Text >).N0 True"
     , N1 5
-        `shouldInjectIntoSmart`
+        `shouldEmbedAsSmart`
         "(< N0 : Bool | N1 : Natural | N2 : Text >).N1 5"
     , N2 "ABC"
-        `shouldInjectIntoSmart`
+        `shouldEmbedAsSmart`
         "(< N0 : Bool | N1 : Natural | N2 : Text >).N2 \"ABC\""
 
-    , E0 `shouldInjectInto` "< E0 | E1 | E2 >.E0"
-    , E1 `shouldInjectInto` "< E0 | E1 | E2 >.E1"
-    , E2 `shouldInjectInto` "< E0 | E1 | E2 >.E2"
+    , E0 `shouldEmbedAsWrapped` "< E0 | E1 | E2 >.E0"
+    , E1 `shouldEmbedAsWrapped` "< E0 | E1 | E2 >.E1"
+    , E2 `shouldEmbedAsWrapped` "< E0 | E1 | E2 >.E2"
 
-    , R0 `shouldInjectInto` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R0"
-    , R1 { a = () } `shouldInjectInto` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R1 { a = {=} }"
-    , R2 { x = 1.0 } `shouldInjectInto` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R2 { x = 1.0}"
-    , R3 { a = (), b = () } `shouldInjectInto` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R3 { a = {=}, b = {=} }"
-    , R4 { x = 1.0, y = 2.0 } `shouldInjectInto` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R4 { x = 1.0, y = 2.0 }"
+    , R0 `shouldEmbedAsWrapped` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R0"
+    , R1 { a = () } `shouldEmbedAsWrapped` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R1 { a = {=} }"
+    , R2 { x = 1.0 } `shouldEmbedAsWrapped` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R2 { x = 1.0}"
+    , R3 { a = (), b = () } `shouldEmbedAsWrapped` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R3 { a = {=}, b = {=} }"
+    , R4 { x = 1.0, y = 2.0 } `shouldEmbedAsWrapped` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R4 { x = 1.0, y = 2.0 }"
 
-    , R0 `shouldInjectIntoSmart` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R0"
-    , R1 { a = () } `shouldInjectIntoSmart` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R1 { a = {=} }"
-    , R2 { x = 1.0 } `shouldInjectIntoSmart` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R2 { x = 1.0}"
-    , R3 { a = (), b = () } `shouldInjectIntoSmart` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R3 { a = {=}, b = {=} }"
-    , R4 { x = 1.0, y = 2.0 } `shouldInjectIntoSmart` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R4 { x = 1.0, y = 2.0 }"
+    , R0 `shouldEmbedAsSmart` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R0"
+    , R1 { a = () } `shouldEmbedAsSmart` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R1 { a = {=} }"
+    , R2 { x = 1.0 } `shouldEmbedAsSmart` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R2 { x = 1.0}"
+    , R3 { a = (), b = () } `shouldEmbedAsSmart` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R3 { a = {=}, b = {=} }"
+    , R4 { x = 1.0, y = 2.0 } `shouldEmbedAsSmart` "< R0 | R1 : { a : {} } | R2 : { x : Double } | R3 : { a : {}, b : {} } | R4 : { x : Double, y : Double } >.R4 { x = 1.0, y = 2.0 }"
 
-    , P0 `shouldInjectInto` "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P0"
-    , P1 () `shouldInjectInto` "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P1 { _1 = {=} }"
-    , P2 1.0 `shouldInjectInto` "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P2 { _1 = 1.0 }"
-    , P3 () () `shouldInjectInto` "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P3 { _1 = {=}, _2 = {=} }"
-    , P4 1.0 2.0 `shouldInjectInto` "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P4 { _1 = 1.0, _2 = 2.0 }"
+    , P0 `shouldEmbedAsWrapped` "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P0"
+    , P1 () `shouldEmbedAsWrapped` "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P1 { _1 = {=} }"
+    , P2 1.0 `shouldEmbedAsWrapped` "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P2 { _1 = 1.0 }"
+    , P3 () () `shouldEmbedAsWrapped` "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P3 { _1 = {=}, _2 = {=} }"
+    , P4 1.0 2.0 `shouldEmbedAsWrapped` "< P0 | P1 : { _1 : {} } | P2 : { _1 : Double } | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P4 { _1 = 1.0, _2 = 2.0 }"
 
-    , P0 `shouldInjectIntoSmart` "< P0 | P1 | P2 : Double | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P0"
-    , P1 () `shouldInjectIntoSmart` "< P0 | P1 | P2 : Double | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P1"
-    , P2 1.0 `shouldInjectIntoSmart` "< P0 | P1 | P2 : Double | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P2 1.0"
-    , P3 () () `shouldInjectIntoSmart` "< P0 | P1 | P2 : Double | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P3 { _1 = {=}, _2 = {=} }"
-    , P4 1.0 2.0 `shouldInjectIntoSmart` "< P0 | P1 | P2 : Double | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P4 { _1 = 1.0, _2 = 2.0 }"
+    , P0 `shouldEmbedAsSmart` "< P0 | P1 | P2 : Double | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P0"
+    , P1 () `shouldEmbedAsSmart` "< P0 | P1 | P2 : Double | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P1"
+    , P2 1.0 `shouldEmbedAsSmart` "< P0 | P1 | P2 : Double | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P2 1.0"
+    , P3 () () `shouldEmbedAsSmart` "< P0 | P1 | P2 : Double | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P3 { _1 = {=}, _2 = {=} }"
+    , P4 1.0 2.0 `shouldEmbedAsSmart` "< P0 | P1 | P2 : Double | P3 : { _1 : {}, _2 : {} } | P4 : { _1 : Double, _2 : Double } >.P4 { _1 = 1.0, _2 = 2.0 }"
     ]
   where
     smartOptions =
         Dhall.defaultInterpretOptions
             { Dhall.singletonConstructors = Dhall.Smart }
 
-    code `shouldPassThrough` values = testCase "Pass through" $ do
-        f <- Dhall.input Dhall.auto code
+    wrappedOptions =
+        Dhall.defaultInterpretOptions
+            { Dhall.singletonConstructors = Dhall.Wrapped }
+
+    code `shouldPassThroughWrapped` values = testCase "Pass through" $ do
+        f <- Dhall.input (Dhall.autoWith wrappedOptions) code
 
         values @=? map f values
 
@@ -319,8 +321,8 @@ shouldHandleUnionsCorrectly =
 
         values @=? map f values
 
-    code `shouldMarshalInto` expectedValue = testCase "Marshal" $ do
-        actualValue <- Dhall.input Dhall.auto code
+    code `shouldMarshalIntoWrapped` expectedValue = testCase "Marshal" $ do
+        actualValue <- Dhall.input (Dhall.autoWith wrappedOptions) code
 
         expectedValue @=? actualValue
 
@@ -329,14 +331,14 @@ shouldHandleUnionsCorrectly =
 
         expectedValue @=? actualValue
 
-    value `shouldInjectInto` expectedCode = testCase "Inject" $ do
+    value `shouldEmbedAsWrapped` expectedCode = testCase "ToDhall" $ do
         parsedExpression <- Dhall.Core.throws (Dhall.Parser.exprFromText "(test)" expectedCode)
 
         resolvedExpression <- Dhall.Import.assertNoImports parsedExpression
 
-        Dhall.Core.denote resolvedExpression @=? Dhall.embed Dhall.inject value
+        Dhall.Core.denote resolvedExpression @=? Dhall.embed (Dhall.injectWith wrappedOptions) value
 
-    value `shouldInjectIntoSmart` expectedCode = testCase "Inject" $ do
+    value `shouldEmbedAsSmart` expectedCode = testCase "ToDhall" $ do
         parsedExpression <- Dhall.Core.throws (Dhall.Parser.exprFromText "(test)" expectedCode)
 
         resolvedExpression <- Dhall.Import.assertNoImports parsedExpression
@@ -366,7 +368,7 @@ shouldConvertDhallToHaskellCorrectly =
         , "{ _1 = True, _2 = {=} }" `correspondsTo` (True, ())
         ]
   where
-    correspondsTo :: (Eq a, Interpret a, Show a) => Text -> a -> TestTree
+    correspondsTo :: (Eq a, FromDhall a, Show a) => Text -> a -> TestTree
     dhallCode `correspondsTo` expectedHaskellValue =
       testCase "Marshall Dhall code to Haskell" $ do
           actualHaskellValue <- Dhall.input Dhall.auto dhallCode
@@ -396,7 +398,7 @@ shouldConvertHaskellToDhallCorrectly =
         , "{ _1 = True, _2 = {=} }" `correspondsTo` (True, ())
         ]
   where
-    correspondsTo :: Inject a => Text -> a -> TestTree
+    correspondsTo :: ToDhall a => Text -> a -> TestTree
     expectedDhallCode `correspondsTo` haskellValue =
         testCase "Marshall Haskell to Dhall code" $ do
             let actualDhallCode =
