@@ -4,8 +4,8 @@ module Dhall.Test.Normalization where
 
 import Data.Monoid ((<>))
 import Data.Text (Text)
+import Data.Void (Void)
 import Dhall.Core (Expr(..), Var(..), throws)
-import Dhall.TypeCheck (X)
 import Prelude hiding (FilePath)
 import Test.Tasty (TestTree)
 import Turtle (FilePath, (</>))
@@ -86,7 +86,7 @@ nestedReduction :: TestTree
 nestedReduction = Tasty.HUnit.testCase "doubleReduction" $ do
     minType        <- Context.insert "min"        <$> Test.Util.code "Natural → Natural → Natural"
     fiveorlessType <- Context.insert "fiveorless" <$> Test.Util.code "Natural → Natural"
-    wurbleType     <- Context.insert "wurble"     <$> Test.Util.code "Natural → Integer"
+    wurbleType     <- Context.insert "wurble"     <$> Test.Util.code "Natural → Natural"
 
     let tyCtx = minType . fiveorlessType . wurbleType $ Context.empty
 
@@ -126,7 +126,7 @@ alphaNormalizationTest prefix = do
 
         expectedResolved <- Import.assertNoImports expectedExpr
 
-        let expectedNormalized = Core.denote expectedResolved :: Expr X X
+        let expectedNormalized = Core.denote expectedResolved :: Expr Void Void
 
         let message =
                 "The normalized expression did not match the expected output"
@@ -138,11 +138,11 @@ alphaNormalizationTest prefix = do
 -}
 unitTest :: Text -> TestTree
 unitTest prefix = do
-    let skip = []
+    let expectedFailures = []
 
     let prefixString = Text.unpack prefix
 
-    Test.Util.testCase prefix skip $ do
+    Test.Util.testCase prefix expectedFailures $ do
         let actualPath   = prefixString <> "A.dhall"
         let expectedPath = prefixString <> "B.dhall"
 
@@ -155,7 +155,7 @@ unitTest prefix = do
 
         let actualNormalized =
                 Core.alphaNormalize
-                    (Core.normalize actualResolved :: Expr X X)
+                    (Core.normalize actualResolved :: Expr Void Void)
 
         expectedExpr <- throws (Parser.exprFromText mempty expectedCode)
 
@@ -183,7 +183,7 @@ betaNormalizationTest prefix =
 
         let actualNormalized =
                 Core.alphaNormalize
-                    (Core.normalize actualResolved :: Expr X X)
+                    (Core.normalize actualResolved :: Expr Void Void)
 
         expectedExpr <- throws (Parser.exprFromText mempty expectedCode)
 
