@@ -22,9 +22,10 @@ import qualified System.Timeout
 import qualified Test.Tasty
 import qualified Test.Tasty.HUnit
 
+import Data.Void (Void)
 import Dhall.Import (Imported, MissingImports(..))
 import Dhall.Parser (Src, SourcedException(..))
-import Dhall.TypeCheck (TypeError, X)
+import Dhall.TypeCheck (TypeError)
 import Test.Tasty (TestTree)
 import Test.Tasty.HUnit ((@?=))
 
@@ -40,6 +41,9 @@ tests =
         , issue253
         , issue1131a
         , issue1131b
+        , issue1341
+        , issue1584
+        , issue1646
         , parsing0
         , typeChecking0
         , typeChecking1
@@ -49,11 +53,11 @@ tests =
         ]
 
 data Foo = Foo Integer Bool | Bar Bool Bool Bool | Baz Integer Integer
-    deriving (Show, Dhall.Generic, Dhall.Interpret, Dhall.Inject)
+    deriving (Show, Dhall.Generic, Dhall.FromDhall, Dhall.ToDhall)
 
 unnamedFields :: TestTree
 unnamedFields = Test.Tasty.HUnit.testCase "Unnamed Fields" (do
-    let ty = Dhall.auto :: Dhall.Type Foo
+    let ty = Dhall.auto :: Dhall.Decoder Foo
     Test.Tasty.HUnit.assertEqual "Good type" (Dhall.expected ty)
         (Dhall.Core.Union
             (Dhall.Map.fromList
@@ -67,14 +71,14 @@ unnamedFields = Test.Tasty.HUnit.testCase "Unnamed Fields" (do
             )
         )
 
-    let inj = Dhall.inject :: Dhall.InputType Foo
-    Test.Tasty.HUnit.assertEqual "Good Inject" (Dhall.declared inj) (Dhall.expected ty)
+    let inj = Dhall.inject :: Dhall.Encoder Foo
+    Test.Tasty.HUnit.assertEqual "Good ToDhall" (Dhall.declared inj) (Dhall.expected ty)
 
-    let tu_ty = Dhall.auto :: Dhall.Type (Integer, Bool)
+    let tu_ty = Dhall.auto :: Dhall.Decoder (Integer, Bool)
     Test.Tasty.HUnit.assertEqual "Auto Tuple" (Dhall.expected tu_ty) (Dhall.Core.Record (
             Dhall.Map.fromList [ ("_1",Dhall.Core.Integer),("_2",Dhall.Core.Bool) ]))
 
-    let tu_in = Dhall.inject :: Dhall.InputType (Integer, Bool)
+    let tu_in = Dhall.inject :: Dhall.Encoder (Integer, Bool)
     Test.Tasty.HUnit.assertEqual "Inj. Tuple" (Dhall.declared tu_in) (Dhall.expected tu_ty)
 
     return () )
@@ -88,14 +92,14 @@ issue96 = Test.Tasty.HUnit.testCase "Issue #96" (do
 issue126 :: TestTree
 issue126 = Test.Tasty.HUnit.testCase "Issue #126" (do
     e <- Util.code "''\nfoo\nbar\n''"
-    Util.normalize' e @?= "\"foo\\nbar\\n\"" )
+    Util.normalize' e @?= "''\nfoo\nbar\n''" )
 
 issue151 :: TestTree
 issue151 = Test.Tasty.HUnit.testCase "Issue #151" (do
     let shouldNotTypeCheck text = do
             let handler :: SourcedException MissingImports -> IO Bool
                 handler (SourcedException _ (MissingImports [e])) =
-                    case Control.Exception.fromException e :: Maybe (Imported (TypeError Src X)) of
+                    case Control.Exception.fromException e :: Maybe (Imported (TypeError Src Void)) of
                         Just _ -> return True
                         Nothing -> return False
                 handler _ = do
@@ -145,7 +149,7 @@ issue216 = Test.Tasty.HUnit.testCase "Issue #216" (do
     source <- Data.Text.IO.readFile "./tests/regression/issue216b.dhall"
     e <- Util.code source
     let doc       = Data.Text.Prettyprint.Doc.pretty e
-    let docStream = Data.Text.Prettyprint.Doc.layoutSmart Dhall.Pretty.layoutOpts doc
+    let docStream = Dhall.Pretty.layout doc
     let text0 = Data.Text.Prettyprint.Doc.Render.Text.renderLazy docStream
 
     text1 <- Data.Text.Lazy.IO.readFile "./tests/regression/issue216b.dhall"
@@ -171,12 +175,33 @@ issue1131b :: TestTree
 issue1131b = Test.Tasty.HUnit.testCase "Issue #1131 b"
     (Util.assertDoesntTypeCheck "toMap {=} : List {mapKey : Text, mapValue : 0}")
 
+issue1341 :: TestTree
+issue1341 = Test.Tasty.HUnit.testCase "Issue #1341" (do
+    let nan    = Dhall.Core.DoubleLit (Dhall.Core.DhallDouble (0/0)) :: Dhall.Core.Expr () ()
+    let actual = Dhall.Core.V "x" 0 `Dhall.Core.freeIn` nan
+    let msg    = "NaN shouldn't contain any free variables"
+    Test.Tasty.HUnit.assertEqual msg False actual)
+
+issue1584 :: TestTree
+issue1584 = Test.Tasty.HUnit.testCase "Issue #1584" (do
+    -- This test ensures that we can parse variables with keyword prefixes
+    -- (e.g. `ifX`)
+    _ <- Util.code "./tests/regression/issue1584.dhall"
+    return () )
+
+issue1646 :: TestTree
+issue1646 = Test.Tasty.HUnit.testCase "Issue #1646" (do
+    -- This test ensures that the parser doesn't eagerly consume trailing
+    -- whitespace after a `Double`
+    _ <- Util.code "./tests/regression/issue1646.dhall"
+    return () )
+
 parsing0 :: TestTree
 parsing0 = Test.Tasty.HUnit.testCase "Parsing regression #0" (do
     -- Verify that parsing should not fail
     --
     -- In 267093f8cddf1c2f909f2d997c31fd0a7cb2440a I broke the parser when left
-    -- factoring the grammer by failing to handle the source tested by this
+    -- factoring the grammar by failing to handle the source tested by this
     -- regression test.  The root of the problem was that the parser was trying
     -- to parse `List ./Node` as `Field List "/Node"` instead of
     -- `App List (Embed (Path (File Homeless "./Node") Code))`.  The latter is
