@@ -44,14 +44,9 @@ getTests = do
 
 successTest :: Text -> TestTree
 successTest prefix = do
-    let skip = [ -- We correctly infer the expected type @NaN ≡ NaN@ here,
-                 -- but the comparison between the inferred and the expected type
-                 -- fails due to `Expr`'s 'Eq' instance, which inherits the
-                 -- @NaN /= NaN@ inequality from 'Double'.
-                 typeInferenceDirectory </> "success/unit/AssertNaN"
-               ]
+    let expectedFailures = []
 
-    Test.Util.testCase prefix skip $ do
+    Test.Util.testCase prefix expectedFailures $ do
         let prefixFP = Text.unpack prefix
 
         actualCode <- Text.IO.readFile (prefixFP <> "A.dhall")
@@ -85,17 +80,18 @@ successTest prefix = do
 
 failureTest :: Text -> TestTree
 failureTest prefix = do
-    let skip = [ typeInferenceDirectory </> "failure/unit/MergeEmptyNeedsDirectAnnotation1"
+    let expectedFailures =
+               [ typeInferenceDirectory </> "failure/unit/MergeEmptyNeedsDirectAnnotation1"
 
                -- Duplicate fields are incorrectly caught during parsing:
                -- https://github.com/dhall-lang/dhall-haskell/issues/772
-               , typeInferenceDirectory </> "failure/unit/RecordLitDuplicateFields"
+               , typeInferenceDirectory </> "failure/unit/RecordProjectionDuplicateFields"
                , typeInferenceDirectory </> "failure/unit/RecordTypeDuplicateFields"
                , typeInferenceDirectory </> "failure/unit/UnionTypeDuplicateVariants1"
                , typeInferenceDirectory </> "failure/unit/UnionTypeDuplicateVariants2"
                ]
 
-    Test.Util.testCase prefix skip $ do
+    Test.Util.testCase prefix expectedFailures $ do
         let prefixFP = Text.unpack prefix
 
         code <- Text.IO.readFile (prefixFP <> ".dhall")
@@ -104,16 +100,8 @@ failureTest prefix = do
             Left _ -> Tasty.HUnit.assertFailure (prefixFP <> " should have parsed")
             Right e -> return e
 
-        tryResolvedExpr <-
-            Exception.try
-               (Test.Util.loadRelativeTo
-                   (FilePath.takeDirectory prefixFP)
-                   Import.IgnoreSemanticCache
-                   expression)
+        resolved <- Import.assertNoImports expression
 
-        case tryResolvedExpr of
-            Left (SomeException _) -> return ()
-
-            Right resolved -> case TypeCheck.typeOf resolved of
-                Left  _ -> return ()
-                Right _ -> Tasty.HUnit.assertFailure (prefixFP <> " should not have type-checked")
+        case TypeCheck.typeOf resolved of
+            Left  _ -> return ()
+            Right _ -> Tasty.HUnit.assertFailure (prefixFP <> " should not have type-checked")
